/*
	Copyright Jeroen Vreeken (pe1rxq@amsat.org), 2008, 2013
	Copyright Stichting C.A. Muller Radioastronomiestation, 2008, 2013

	This program is free software: you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation, either version 3 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program.  If not, see <http://www.gnu.org/licenses/>.

 */

#include <stdlib.h>
#include <unistd.h>
#include <stdio.h>
#include <string.h>
#include <pthread.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/select.h>
#include <sys/ioctl.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <signal.h>
#include <errno.h>
#include <semaphore.h>
#include <limits.h>
#include <stdarg.h>
#include <stdbool.h>

#include <utils/tcp_connect.h>

#include <log/log.h>


static int listen_port = 0;
static int server_port = 0;
static char *server_host = NULL;
static char *client_name = NULL;


struct log_client {
	int used;
	int fd;
};

#define LOG_MAX_CLIENTS	10
#define LOG_MAX_QUEUE	2048
#define LOG_MAX_LEN	400

struct log_msg {
	int used;
	time_t t;
	int type;
	char msg[LOG_MAX_LEN];
};

static struct log_client clients[LOG_MAX_CLIENTS];

static struct log_msg msgs[LOG_MAX_QUEUE];
static int msgs_wr = 0;
static int msgs_rd = 0;
static int lost_counter = 0;

static sem_t queue_wait;

static enum log_type log_level_console;
static enum log_type log_level_remote;

static pthread_mutex_t log_mutex;

void log_send(enum log_type type, char *fmt, ...)
{
	int val;
	va_list ap;

	if (log_level_console < type && log_level_remote < type)
		return;
	
  	va_start(ap, fmt); 

	pthread_mutex_lock(&log_mutex);

  	if (msgs[msgs_wr].used == 0) {
		msgs[msgs_wr].t = time(NULL);
		msgs[msgs_wr].type = type;
		vsnprintf(msgs[msgs_wr].msg, LOG_MAX_LEN, fmt, ap);
		msgs[msgs_wr].used = 1;
		msgs_wr++;
		msgs_wr %= LOG_MAX_QUEUE;
	} else {
		lost_counter++;
	}

	/* If we lost some in the past check if we can log now... */
  	if (lost_counter && msgs[msgs_wr].used == 0) {
		msgs[msgs_wr].t = time(NULL);
		msgs[msgs_wr].type = LOG_T_WARNING;
		snprintf(msgs[msgs_wr].msg, LOG_MAX_LEN, 
		    "Lost %d messages",  lost_counter);
		msgs[msgs_wr].used = 1;
		msgs_wr++;
		msgs_wr %= LOG_MAX_QUEUE;
		lost_counter = 0;
	}
	pthread_mutex_unlock(&log_mutex);
	
	va_end(ap);

	sem_getvalue(&queue_wait, &val);
	if (!val)
		sem_post(&queue_wait);
}

static void timestamp(struct tm *tm, char *time)
{
	sprintf(time, "%04d-%02d-%02dT%02d:%02d:%02dZ",
	    tm->tm_year + 1900, tm->tm_mon + 1, tm->tm_mday,
	    tm->tm_hour, tm->tm_min, tm->tm_sec);
}

static char *logtype(enum log_type type)
{
	switch(type) {
		case LOG_T_ERROR:
			return "  ERROR: ";
		case LOG_T_WARNING:
			return "WARNING: ";
		case LOG_T_INFO:
			return "   INFO: ";
		case LOG_T_DEBUG:
			return "  DEBUG: ";
		default:
			return "UNKNOWN: ";
	}
}


static void *log_server(void *arg)
{
	int fd_accept;
	struct sockaddr_in sockaddr;
	
	signal(SIGPIPE, SIG_IGN);
	
	fd_accept = socket(AF_INET, SOCK_STREAM, 0);
	if (fd_accept < 0) {
		perror("socket() failed");
		return NULL;
	}
	
	sockaddr.sin_family = AF_INET;
	sockaddr.sin_addr.s_addr = htonl(INADDR_ANY);
	sockaddr.sin_port = htons(listen_port);
	
	setsockopt(fd_accept, SOL_SOCKET, SO_REUSEADDR, &(int){1}, sizeof(int));

	if (bind(fd_accept,
	    (struct sockaddr *)&sockaddr, sizeof(sockaddr)) < 0) {
		perror("bind() failed");
		close(fd_accept);
		return NULL;
	}
	
	if (listen(fd_accept, 4) < 0) {
		perror("listen() failed");
		close(fd_accept);
		return NULL;
	}
	
	while (1) {
		int client;
		struct sockaddr_in sockaddr_client;
		socklen_t len;
		struct timeval tv;
		fd_set fdset_r;
		bool newclient;
		
		if (!msgs[msgs_rd].used) {
			sem_wait(&queue_wait);
		}
		
		do {
			FD_ZERO(&fdset_r);
			FD_SET(fd_accept, &fdset_r);
			tv.tv_sec = 0;
			tv.tv_usec = 0;
		
			select(fd_accept + 1, &fdset_r, NULL, NULL, &tv);
			newclient = false;
		
			if (FD_ISSET(fd_accept, &fdset_r)) {
				len = sizeof(sockaddr_client);
				client = accept(fd_accept,
				    (struct sockaddr *)&sockaddr_client, &len);
				newclient = true;
			
				if (client >= 0) {
					int i;
				
					ioctl(client, FIONBIO, &(int){1}, sizeof(int));
					
					for (i = 0; i < LOG_MAX_CLIENTS; i++) {
						if (!clients[i].used) {
							clients[i].used = 1;
							clients[i].fd = client;
							break;
						}
					}
					if (i == LOG_MAX_CLIENTS) {
						close(client);
					}
				}
			}
		} while (newclient);
				
		if (msgs[msgs_rd].used) {
			int i;
			char time[100];
			char *header;
			struct tm gmt;
			
			gmtime_r(&msgs[msgs_rd].t, &gmt);
			timestamp(&gmt, time);

			header = logtype(msgs[msgs_rd].type);
			
			if (msgs[msgs_rd].type <= log_level_remote) {
				for (i = 0; i < LOG_MAX_CLIENTS; i++) {
					int r;
				
					if (clients[i].used == 0)
						continue;
				
					write(clients[i].fd, time, strlen(time));
					write(clients[i].fd, " ", 1);
					write(clients[i].fd, header, strlen(header));

					write(clients[i].fd, msgs[msgs_rd].msg,
					    strlen(msgs[msgs_rd].msg));
					
					r = write(clients[i].fd, "\n", 1);
					fsync(clients[i].fd);
					
					if (r <= 0) {
						close(clients[i].fd);
						clients[i].fd = -1;
						clients[i].used = 0;
					}
				}
			}
			
			if (msgs[msgs_rd].type <= log_level_console)
				printf("%s %s%s\n", time, header, msgs[msgs_rd].msg);
			
			msgs[msgs_rd].used = 0;
			msgs_rd++;
			msgs_rd %= LOG_MAX_QUEUE;
		}
	}

	return NULL;
}

int log_server_start(int port, enum log_type console_level,
    enum log_type remote_level)
{
	pthread_t thread_id;
	pthread_attr_t attr;
	int i;

	log_level_console = console_level;
	log_level_remote = remote_level;

	listen_port = port;

	for (i = 0; i < LOG_MAX_QUEUE; i++) {
		msgs[i].used = 0;
	}
	
	for (i = 0; i < LOG_MAX_CLIENTS; i++) {
		clients[i].fd = -1;
		clients[i].used = 0;
	}

	sem_init(&queue_wait, 0, 1);

	printf("Starting log server on port %d\n", port);

	pthread_attr_init(&attr);
	pthread_attr_setstacksize(&attr, PTHREAD_STACK_MIN * 2);

	pthread_create(&thread_id, &attr, log_server, NULL);

	return 0;
}


static void *log_client(void *arg)
{
	int fd = -1;
	
	signal(SIGPIPE, SIG_IGN);
	
	while (1) {
		if (!msgs[msgs_rd].used) {
			sem_wait(&queue_wait);
		}
						
		if (msgs[msgs_rd].used) {
			char time[100];
			char *header;
			struct tm gmt;
			
			gmtime_r(&msgs[msgs_rd].t, &gmt);
			timestamp(&gmt, time);

			header = logtype(msgs[msgs_rd].type);
			
			if (fd < 0) {
				fd = tcp_connect(server_host, server_port);
			}
			
			if (msgs[msgs_rd].type <= log_level_remote && fd >= 0) {
				int r;
				
				write(fd, time, strlen(time));
				write(fd, " ", 1);
				write(fd, header, strlen(header));
					
				if (client_name) {
					write(fd, client_name, strlen(client_name));
					write(fd, ": ", 2);
				}
					
				write(fd, msgs[msgs_rd].msg,
				    strlen(msgs[msgs_rd].msg));
					
				r = write(fd, "\n", 1);
				fsync(fd);
					
				if (r <= 0) {
					close(fd);
					fd = -1;
				}
			}
			
			if (msgs[msgs_rd].type <= log_level_console)
				printf("%s %s%s\n", time, header, msgs[msgs_rd].msg);
			
			msgs[msgs_rd].used = 0;
			msgs_rd++;
			msgs_rd %= LOG_MAX_QUEUE;
		}
	}

	return NULL;
}

int log_client_start(char *host, int port, enum log_type console_level,
    enum log_type remote_level, char *name)
{
	pthread_t thread_id;
	pthread_attr_t attr;
	int i;

	log_level_console = console_level;
	log_level_remote = remote_level;

	server_port = port;
	if (server_host)
		free(server_host);
	server_host = strdup(host);
	if (client_name)
		free(client_name);
	client_name = strdup(name);
	
	for (i = 0; i < LOG_MAX_QUEUE; i++) {
		msgs[i].used = 0;
	}
	
	for (i = 0; i < LOG_MAX_CLIENTS; i++) {
		clients[i].fd = -1;
		clients[i].used = 0;
	}

	sem_init(&queue_wait, 0, 1);

	printf("Starting log client %s:%d for %s\n", host, port, name);

	pthread_attr_init(&attr);
	pthread_attr_setstacksize(&attr, PTHREAD_STACK_MIN * 2);

	pthread_create(&thread_id, &attr, log_client, NULL);

	return 0;
}

void log_server_flush(void)
{
	sleep(1);
	do {
		sleep(1);
	} while (msgs[msgs_rd].used);
}

void log_string(char *str, size_t len, enum log_type type, char *fmt, ...)
{
	va_list ap;
	time_t t = time(NULL);
	int plen;
	struct tm gmt;

  	va_start(ap, fmt); 

	gmtime_r(&t, &gmt);
	timestamp(&gmt, str);

	len -= strlen(str);
	str += strlen(str);

	plen = snprintf(str, len, " %s", logtype(type));
	len -= plen;
	str += plen;
	
	vsnprintf(str, len, fmt, ap);
	
	va_end(ap);
}
