/*
	trace support

	Copyright Jeroen Vreeken (pe1rxq@amsat.org), 2008, 2011
	Copyright Stichting C.A. Muller Radioastronomiestation, 2008, 2011

	This program is free software: you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation, either version 3 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program.  If not, see <http://www.gnu.org/licenses/>.

*/

#ifndef _INCLUDE_TRACE_H_
#define _INCLUDE_TRACE_H_

#include <time.h>
#include <stdbool.h>
#include <sys/select.h>
#include <stdint.h>


enum trace_ptype {
	TRACE_PTYPE_TIMESTAMP = 0,
	TRACE_PTYPE_LIST = 1,
	TRACE_PTYPE_NAME = 2,
	TRACE_PTYPE_INTERVAL = 3,
	TRACE_PTYPE_DATA = 4,
	TRACE_PTYPE_VALUE_TYPE = 5,
};

enum trace_value_type {
	TRACE_VALUE_TYPE_FLOAT,	/* 4 bytes */
	TRACE_VALUE_TYPE_BOOL,	/* 1 byte */
	TRACE_VALUE_TYPE_UINT8,
	TRACE_VALUE_TYPE_UINT16,
	TRACE_VALUE_TYPE_UINT32,
	TRACE_VALUE_TYPE_SINT8,
	TRACE_VALUE_TYPE_SINT16,
	TRACE_VALUE_TYPE_SINT32,
};

char *enum_trace_value_type2str(enum trace_value_type type);

enum trace_interval_type {
	TRACE_INTERVAL_TYPE_INTERVAL = 1,
	TRACE_INTERVAL_TYPE_CHANGED = 2,
	TRACE_INTERVAL_TYPE_CHANGED_INTERVAL = 3,
};

char *enum_trace_interval_type2str(enum trace_interval_type type);

enum trace_state {
	TRACE_STATE_DISCONNECTED,
	TRACE_STATE_CONNECTED,
	TRACE_STATE_READY,
	TRACE_STATE_RECEIVING,
};

char *enum_trace_state2str(enum trace_state state);

struct trace_value {
	union {
		float f;
		bool b;
		uint8_t u8;
		uint16_t u16;
		uint32_t u32;
		int8_t s8;
		int16_t s16;
		int32_t s32;
	} value;
	struct timespec t;
};

struct trace {
	int fd;
	char *host;
	int port;
	
	unsigned char *rx_buffer;
	size_t rx_len;
	size_t rx_buf_size;
	
	size_t packet_size;
	
	enum trace_value_type type;
	struct timespec timestamp;
	struct timespec interval;
	enum trace_interval_type interval_type;
	struct timespec interval_req;
	enum trace_interval_type interval_type_req;
	
	char *name;

	bool interval_set;
	bool name_set;
	bool list_received;
	bool recover;
	bool type_set;
	bool value_set;
	time_t recovertime;
	
	struct trace_value value;

	void (*handler_interval)(struct trace *, 
	    struct timespec *interval, enum trace_interval_type type);
	void (*handler_timestamp)(struct trace *, 
	    struct timespec *timestamp);
	void (*handler_list_entry)(struct trace *, 
	    char *name, enum trace_value_type type, char *unit);
	void (*handler_type)(struct trace *, 
	    enum trace_value_type type);
	void (*handler_value)(struct trace *,
	    struct trace_value *value);
	void (*handler_name)(struct trace *, char *name);
	void (*handler_close)(struct trace *);
	
	void *private;
};

struct trace_pkt {
	enum trace_ptype type;
	unsigned char *data;
	size_t len;
	size_t buflen;
};


struct trace *trace_alloc(void);
struct trace *trace_open(char *host, int port);
void trace_initialize(struct trace *trace, size_t buffer);
void trace_initialize_fd(struct trace *trace, int fd);
int trace_interval_set(struct trace *trace,
    struct timespec *interval, enum trace_interval_type type);
int trace_name_set(struct trace *trace, char *name);
void trace_autorecover(struct trace *trace, bool value);
void trace_close(struct trace *trace);
void trace_free(struct trace *trace);
int trace_fd_get(struct trace *trace);
bool trace_fd_read(struct trace *trace);

struct trace_pkt *trace_packet_get(struct trace *trace);
void trace_packet_put(struct trace_pkt *pkt);

void trace_packet_initialize(struct trace_pkt *pkt);

int trace_packet_write(struct trace *trace, struct trace_pkt *pkt);

struct trace_pkt *trace_packet_new(void);
int trace_packet_resize(struct trace_pkt *pkt, size_t len);

int trace_packet_list_add(struct trace_pkt *list_pkt, 
    char *name, enum trace_value_type type, char *unit);

int trace_packet_interval_set(struct trace_pkt *pkt, 
    struct timespec *interval, enum trace_interval_type type);
int trace_packet_timestamp_set(struct trace_pkt *pkt, 
    struct timespec *timestamp);
int trace_packet_name_set(struct trace_pkt *pkt, char *name);
int trace_packet_type_set(struct trace_pkt *pkt, enum trace_value_type type);
int trace_packet_value_add(struct trace_pkt *pkt, struct trace_value *value, enum trace_value_type type);

bool trace_fd_set(struct trace *trace, fd_set *set, int *high);
int trace_handle(struct trace *trace, fd_set *set);

enum trace_state trace_state_get(struct trace *trace);

#endif /* _INCLUDE_TRACE_H_ */
