/*
	dt_model.c
	Dwingeloo Telescope correction model.
	
	Copyright Jeroen Vreeken (pe1rxq@amsat.org), 2011
	Copyright Stichting C.A. Muller Radioastronomiestation, 2011

	This program is free software: you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation, either version 3 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program.  If not, see <http://www.gnu.org/licenses/>.

 */

#include <unistd.h>
#include <math.h>
#include <sys/mman.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <string.h>
#include <stdio.h>

#include "dt_model.h"

/*
	Pointing model made by Paul Boven
	Version 2010-08 taken from:
	http://www.xs4all.nl/~epboven/Camras/pointing.html

Het pointing model:
 
d-az = a0 + c1 / cos(el) - c2 * tan(el) - za * sin(az - aa) * tan(el)
d-el = e0 + b * cos(el) - r / sin(el) - za * cos(az - aa)
 
az, el: azimut, elevatie
 
a0: noord offset
c1: horizontale scheefstand beam
c2: hoogteverschil elevatie lagers
 
e0: elevatie offset
b: doorbuiging
r: refractie
 
za: scheefstand rail
aa: kompasrichting scheefstand

*/

static struct dt_model_params dt_model_default;

struct dt_model_params *params = &dt_model_default;


double dt_model_azimuth_delta(double az, double el)
{
	double delta_az;
	double c1_div_cos_el = params->c1 / cos(el);
	
	if (isnan(c1_div_cos_el))
		c1_div_cos_el = 0.0;
	
	delta_az = 
	    params->a0 
	    + params->c1 / cos(el) 
	    - params->c2 * tan(el) 
	    - params->za * sin(az - params->aa) * tan(el);

	return delta_az;
}

double dt_model_azimuth_delta_rev(double az, double el)
{
	/* No reverse model available, (ab)use the forward model */
	return -dt_model_azimuth_delta(az, el);
}

double dt_model_elevation_delta(double az, double el)
{
	double delta_el;

	delta_el =
	     params->e0 
	     + params->b * cos(el) 
	     - params->za * cos(az - params->aa);
	
	return delta_el;
}

double dt_model_elevation_delta_rev(double az, double el)
{
	/* No reverse model available, (ab)use the forward model */
	return -dt_model_elevation_delta(az, el);
}

int dt_model_init(void)
{
	int fd;
	
	fd = shm_open("/dt_model", 
	    O_RDWR | O_CREAT,
	    S_IRUSR | S_IWUSR | S_IRGRP | S_IROTH);

	if (fd < 0)
		return -1;
	
	if((ftruncate(fd, sizeof(struct dt_model_params))) == -1){
		printf("ftruncate failure\n");
		return -1;
	}
	
	params = mmap(NULL, sizeof(struct dt_model_params), 
	    PROT_READ,
	    MAP_SHARED, fd, 0);

	return params != NULL;
}

int dt_model_init_server(void)
{
	int fd;
	
	fd = shm_open("/dt_model", 
	    O_RDWR | O_CREAT,
	    S_IRUSR | S_IWUSR | S_IRGRP | S_IROTH);

	if (fd < 0)
		return -1;
	
	if((ftruncate(fd, sizeof(struct dt_model_params))) == -1){
		printf("ftruncate failure\n");
		return -1;
	}
	
	params = mmap(NULL, sizeof(struct dt_model_params), 
	    PROT_READ | PROT_WRITE,
	    MAP_SHARED, fd, 0);

	return params != NULL;
}

int dt_model_params_set(struct dt_model_params *newparams)
{
	memcpy(params, newparams, sizeof(struct dt_model_params));
	
	return 0;
}

int dt_model_params_get(struct dt_model_params *getparams)
{
	memcpy(getparams, params, sizeof(struct dt_model_params));
	
	return 0;
}

