/*
	Console interface to the DT model

	Copyright Jeroen Vreeken (pe1rxq@amsat.org), 2013
	Copyright Stichting C.A. Muller Radioastronomiestation, 2013

	This program is free software: you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation, either version 3 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program.  If not, see <http://www.gnu.org/licenses/>.

*/

#include <time.h>
#include <sys/time.h>
#include <unistd.h>
#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <string.h>
#include <fcntl.h>
#include <termios.h>
#include <signal.h>
#include <ctype.h>
#include <inttypes.h>
#include <arpa/inet.h>
#include <sys/select.h>
#include <sys/ioctl.h>
#include <errno.h>
#include <stdbool.h>
#include <glib.h>

#include <utils/dt_model.h>

#include <dt_port_numbers.h>
#include <utils/dt_host.h>
#include <utils/command_server.h>
#include <utils/status_server.h>
#include <log/log.h>


int stat_port = CONSOLE_DT_MODEL_STAT_PORT;
int cmd_port = CONSOLE_DT_MODEL_CMD_PORT;


#define OUTPUT_RATE 3


/************************************/

static int load_ini(void)
{
	GKeyFile *keyfile;
	gboolean loaded = false;
	gchar *keyfilename[] = { "/etc/dt/dt_model.ini", "dt_model.ini" };
	int ret = 0;
	int i;
	char *active_name = "none";
	struct dt_model_params params = {};
	
	keyfile = g_key_file_new();
	
	for (i = 0; i < sizeof(keyfilename)/sizeof(keyfilename[0]) && !loaded; i++) {
		loaded = g_key_file_load_from_file(
		    keyfile,
		    keyfilename[i],
		    G_KEY_FILE_KEEP_COMMENTS,
		    NULL);
	}
	
	if (loaded) {
		gsize groups_nr;
		gchar **groups;
		
		groups = g_key_file_get_groups(keyfile, &groups_nr);

		printf("Number of groups in ini file: %d\n", (int)groups_nr);
		
		for (i = 0; i < groups_nr; i++) {
			if (!strcmp(groups[i], "global")) {
				if (g_key_file_has_key(keyfile, groups[i], "active", NULL)) {
					active_name =
					    g_key_file_get_string(keyfile, groups[i], "active", NULL);
					printf("Active params: '%s'\n", active_name);
				}
			}
		}
		for (i = 0; i < groups_nr; i++) {
			if (strcmp(groups[i], "global")) {
				if (g_key_file_has_key(keyfile, groups[i], "a0", NULL)) {
					params.a0 =
					    g_key_file_get_double(keyfile, groups[i], "a0", NULL);
				}
				if (g_key_file_has_key(keyfile, groups[i], "c1", NULL)) {
					params.c1 =
					    g_key_file_get_double(keyfile, groups[i], "c1", NULL);
				}
				if (g_key_file_has_key(keyfile, groups[i], "c2", NULL)) {
					params.c2 =
					    g_key_file_get_double(keyfile, groups[i], "c2", NULL);
				}
				if (g_key_file_has_key(keyfile, groups[i], "e0", NULL)) {
					params.e0 =
					    g_key_file_get_double(keyfile, groups[i], "e0", NULL);
				}
				if (g_key_file_has_key(keyfile, groups[i], "b", NULL)) {
					params.b =
					    g_key_file_get_double(keyfile, groups[i], "b", NULL);
				}
				if (g_key_file_has_key(keyfile, groups[i], "za", NULL)) {
					params.za =
					    g_key_file_get_double(keyfile, groups[i], "za", NULL);
				}
				if (g_key_file_has_key(keyfile, groups[i], "aa", NULL)) {
					params.aa =
					    g_key_file_get_double(keyfile, groups[i], "aa", NULL);
				}
				
				strncpy(params.name, groups[i], DT_MODEL_NAME_SIZE);
				params.name[DT_MODEL_NAME_SIZE-1] = 0;
				
				if (!strcmp(groups[i], active_name)) {
					printf("Set params: %d\n", 
					    dt_model_params_set(&params));
				}
			}
		}
	} else {
		log_send(LOG_T_ERROR, "ini file load failed\n");
		ret = -1;
	}

	g_key_file_free(keyfile);
	
	return ret;
}



/************************************/


static int handle_cmd(char *name, char *val)
{
	printf("name %s value %s\n", name, val);
	if (!strcmp(name, "reload")) {
		log_send(LOG_T_WARNING, "Reloading ini file");
		load_ini();
	}

	return 0;
}


void output(struct status_server *stat_srv)
{
	static time_t last = 0;
	time_t now;
	char statline[800];
	char *statlinepos;
	struct dt_model_params params;
	
	now = time(NULL);
	if (now <= last)
		return;
	
	last = now;
	
	statlinepos = statline;

	dt_model_params_get(&params);

	sprintf(statlinepos, 
	    "name=%s,a0=%f,c1=%f,c2=%f,e0=%f,b=%f,za=%f,aa=%f\n",
	    params.name, params.a0, params.c1, params.c2, params.e0,
	    params.b, params.za, params.aa);

	status_server_send(stat_srv, statline);
}


int main(int argc, char **argv)
{
	struct command_server *cmd_srv;
	struct status_server *stat_srv;
	
	dt_model_init_server();
	
	load_ini();

	signal(SIGPIPE, SIG_IGN);

	cmd_srv = command_server_create(cmd_port, 0, 100);
	if (!cmd_srv) {
		printf("Could not open listen port for commands\n");
	} else {
		command_server_handler_set(cmd_srv, handle_cmd);
	}

	stat_srv = status_server_create(stat_port, 0, 100);
	if (!cmd_srv) {
		printf("Could not open listen port for status\n");
	}


	log_client_start(dt_host_console(), CONSOLE_LOG_PORT_IN, 
	    LOG_T_DEBUG, LOG_T_INFO, "console/dt_model");

	while (1)
	{
		int high = 0;
		fd_set fdset_rx;
		struct timeval tv;


		FD_ZERO(&fdset_rx);
	
		command_server_fdset_add(cmd_srv, &fdset_rx, &high);
		status_server_fdset_add(stat_srv, &fdset_rx, &high);
		
		tv.tv_sec = OUTPUT_RATE;
		tv.tv_usec = 0;

		select(high, &fdset_rx, NULL, NULL, &tv);

		command_server_fdset_handle(cmd_srv, &fdset_rx);
		status_server_fdset_handle(stat_srv, &fdset_rx);
		
		output(stat_srv);
	}
}
