/*
	Command generator for J2000.0 positions

	Copyright Jeroen Vreeken (pe1rxq@amsat.org), 2008
	Copyright Stichting C.A. Muller Radioastronomiestation, 2008

	This program is free software: you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation, either version 3 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program.  If not, see <http://www.gnu.org/licenses/>.

*/

#include <string.h>
#include <libindi/indidevapi.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <math.h>
#include <fcntl.h>
#include <sys/ioctl.h>

#include <utils/tcp_connect.h>

#define mydev		"Dwingeloo Radio Telescope"
#define BASIC_GROUP	"Main Control"
#define POLLMS		1000
#define RAVALUE		eqWN[0].value
#define DECVALUE	eqWN[1].value

static char *tracker_host = "localhost";
static int stat_port = 11030;
static int cmd_port = 11031;
static int stat_fd = -1;
static int cmd_fd = -1;

static char stat_buffer[100];
static size_t stat_buffer_pos = 0;
static float stat_ra = 0.0;
static float stat_dec = 0.0;

static int allow_set = 0;

void update_stat(void)
{
	int ret;
	
	if (stat_fd < 0) {
		stat_fd = tcp_connect(tracker_host, stat_port);
		if (stat_fd < 0)
			return;
		ioctl(stat_fd, FIONBIO, &(int){ 1 });
	}
	
	do {
		ret = read(stat_fd, stat_buffer + stat_buffer_pos, 1);
		if (ret > 0) {
			stat_buffer_pos++;
			if (stat_buffer[stat_buffer_pos-1] == '\n') {
				char *p;
				char *ra, *dec;
				char *rah, *ram,  *ras;
				char *decd, *decm, *decs;
				stat_buffer[stat_buffer_pos-1] = 0;
				ra = strtok_r(stat_buffer, " ", &p);
				dec = strtok_r(NULL, " ", &p);
				ra = strtok_r(NULL, " ", &p);
				dec = strtok_r(NULL, " ", &p);
				if (!ra)
					ra = "0:0:0";
				if (!dec)
					dec = "0:0:0";
				rah = strtok_r(ra, "hm:", &p);
				ram = strtok_r(NULL, "hm:", &p);
				ras = strtok_r(NULL, "hm:", &p);
				decd = strtok_r(dec, "dm:", &p);
				decm = strtok_r(NULL, "dm:", &p);
				decs = strtok_r(NULL, "dm:", &p);
				if (!rah || !ram || !ras || !decd || !decm || !decs) {
					rah = "0";
					ram = "0";
					ras = "0";
					decd = "0";
					decm = "0";
					decs = "0";
				}
				fprintf(stderr, "%s:%s:%s %s:%s:%s\n", 
				    rah, ram, ras, 
				    decd, decm, decs);
				stat_ra = 
				    atof(rah) + 
				    copysign(atof(ram)/60.0, atof(rah)) +
				    copysign(atof(ras)/3600.0, atof(rah));
				stat_dec =
				    atof(decd) +
				    copysign(atof(decm)/60.0, atof(decd)) +
				    copysign(atof(decs)/3600.0, atof(decd));
				stat_buffer_pos = 0;
			}
		}
	} while (ret == 1);
	if (ret == 0) {
		close(stat_fd);
		stat_fd = -1;
	}
}

static void ISPoll(void *);
static void ISInit(void);
static void connectTelescope(void);

static ISwitch PowerS[] = {
	{"CONNECT" , "Connect" , ISS_OFF, 0, 0},
	{"DISCONNECT", "Disconnect", ISS_ON, 0, 0}
};

ISwitchVectorProperty PowerSP = {
	mydev, 
	"CONNECTION" , 
	"Connection", 
	BASIC_GROUP, 
	IP_RW, ISR_1OFMANY, 
	0, 
	IPS_IDLE, 
	PowerS, 
	NARRAY(PowerS), 
	"", 
	0
};

/* equatorial position */
INumber eqWN[] = {
    {"RA",  "RA  H:M:S", "%10.6m",  0., 24., 0., 0., 0, 0, 0},
    {"DEC", "Dec D:M:S", "%10.6m", -90., 90., 0., 0., 0, 0, 0},
};

INumberVectorProperty eqWNP = {
	mydev, 
	"EQUATORIAL_COORD", 
	"Equatorial J2000", 
	BASIC_GROUP, 
	IP_RW, 
	120, 
	IPS_IDLE,
	eqWN, 
	NARRAY(eqWN), 
	"", 
	0
};

static ISwitch OnCoordSetS[] = {
	{"TRACK", "Track", ISS_ON, 0, 0}, 
};

static ISwitchVectorProperty OnCoordSetSP = { 
	mydev, 
	"ON_COORD_SET", 
	"On Set", 
	BASIC_GROUP, 
	IP_RW, 
	ISR_1OFMANY, 
	0, 
	IPS_IDLE, 
	OnCoordSetS, 
	NARRAY(OnCoordSetS), 
	"", 
	0
};

void ISInit(void)
{
	static int isInit=0;

	if (isInit)
		return;

	isInit = 1;

	IEAddTimer (POLLMS, ISPoll, NULL);
}

void ISGetProperties (const char *dev)
{
	ISInit(); 

	dev=dev;

	IDDefSwitch(&PowerSP, NULL);
	IDDefSwitch(&OnCoordSetSP, NULL);
	IDDefNumber(&eqWNP, NULL);
}

void ISNewSwitch (const char *dev, const char *name, ISState *states, char *names[], int n)
{
	ISInit();

	dev = dev;

	if (!strcmp(name, PowerSP.name))
	{
		IUResetSwitch(&PowerSP);
		IUUpdateSwitch(&PowerSP, states, names, n);
		connectTelescope();
		return;
	}
	if (!strcmp(name, OnCoordSetSP.name))
	{
		IUResetSwitch(&OnCoordSetSP);
		IUUpdateSwitch(&OnCoordSetSP, states, names, n);
		OnCoordSetSP.s = IPS_OK;
		IDSetSwitch(&OnCoordSetSP, "DT tracking.");
		return;
	}
}

void ISNewText (const char *dev, const char *name, char *texts[], char *names[], int n)
{
	dev=dev; names=names; n=n;
}

void ISNewNumber (const char *dev, const char *name, double values[], char *names[], int n)
{
	double ra, dec;
	int i, nset;
	
	dev=dev;
	name=name;
	values=values;
	names=names;
	n=n;
	
	for (nset = i = 0; i < n; i++) {
		INumber *eqp = IUFindNumber(&eqWNP, names[i]);
		if (eqp == &eqWN[0]) {
			ra = values[i];
			nset++;
		} else if (eqp == &eqWN[1]) {
			dec = values[i];
			nset++;
		}
	}
	
	if (allow_set && nset == 2) {
		int rah;
		int ram;
		float ras;
		int decd;
		int decm;
		float decs;
		fprintf(stderr, "new  RA: %f DEC: %f\n", ra, dec);
		
		rah = ra;
		ra -= rah;
		ra *= 60.0;
		ram = ra;
		ra -= ram;
		ras = ra * 60.0;
		decd = dec;
		dec -= decd;
		dec *= 60.0;
		decm = dec;
		dec -= decm;
		decs = dec * 60.0;
		if (decm < 0)
			decm = -decm;
		if (decs < 0)
			decs = -decs;
		if (cmd_fd < 0) {
			cmd_fd = tcp_connect(tracker_host, cmd_port);
		}
		if (cmd_fd >= 0) {
			char command[100];
			int ret;
			sprintf(command, "%dh%dm%f %dd%dm%f\n",
			    rah, ram, ras, decd, decm, decs);
			fprintf(stderr, command);
			ret = write(cmd_fd, command, strlen(command));
			if (ret <= 0) {
				close(cmd_fd);
				cmd_fd = -1;
			}
		}
	}
}

void ISNewBLOB (const char *dev, const char *name, int sizes[], int blobsizes[], char *blobs[], char *formats[], char *names[], int n) 
{
	INDI_UNUSED(dev);
	INDI_UNUSED(name);
	INDI_UNUSED(sizes);
	INDI_UNUSED(blobsizes);
	INDI_UNUSED(blobs);
	INDI_UNUSED(formats);
	INDI_UNUSED(names);
	INDI_UNUSED(n);
}
void ISSnoopDevice (XMLEle *root) 
{
	INDI_UNUSED(root);
}

void ISPoll (void *p)
{
	p=p;

	update_stat();

	if (PowerS[0].s == ISS_ON)
	{
		switch (eqWNP.s)
		{
			case IPS_IDLE:
			case IPS_OK:
			case IPS_BUSY:
				RAVALUE = stat_ra;
				DECVALUE = stat_dec;
				fprintf(stderr, "poll RA: %f DEC: %f\n", RAVALUE, DECVALUE);
				IDSetNumber(&eqWNP, NULL);
				break;
	        
			case IPS_ALERT:
			break;
		}
	}

	IEAddTimer(POLLMS, ISPoll, NULL);
}

static void connectTelescope(void)
{
	switch (PowerS[0].s) {
		case ISS_ON:
			PowerSP.s = IPS_OK;
			IDSetSwitch(&PowerSP, "DT is online.");
			break;

		case ISS_OFF:
			IUResetSwitch(&PowerSP);
			eqWNP.s = PowerSP.s = IPS_IDLE;
			IDSetSwitch(&PowerSP, "DT is offline.");
			IDSetNumber(&eqWNP, NULL);
			break;
	}
}

