/*
	console_joystick.c
	
	Copyright Jeroen Vreeken (pe1rxq@amsat.org), 2007, 2008
	Copyright Stichting C.A. Muller Radioastronomiestation, 2007, 2008

	This program is free software: you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation, either version 3 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program.  If not, see <http://www.gnu.org/licenses/>.

 */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <string.h>
#include <math.h>
#include <sys/select.h>
#include <time.h>
#include <errno.h>
#include <signal.h>
#include <fcntl.h>
#include <linux/joystick.h>

#include <utils/config.h>

#include <command/command.h>

char *x_command_host = "localhost";
int x_command_port = 11000;
char *y_command_host = "localhost";
int y_command_port = 11000;
char *x_command_spg = "Azimuth_Setpoint";
char *y_command_spg = "Elevation_Setpoint";

char *joystick_device = "/dev/input/js0";

struct command *sp_command_x = NULL;
struct command *sp_command_y = NULL;

double x_scale = (0.1 * M_PI * 2.0 / 60.0) / 32768.0;
double y_scale = (0.08 * M_PI * 2.0 / 60.0) / 32768.0;

int update_x(struct command *sp_cmd, int value)
{
	float speed;
	struct command_entry entry;
	
	speed = (float)value;

	/* make the joystick logarithmic */
	speed = speed * speed / 32768.0;
	if (value < 0)
		speed *= -1.0;
	
	/* Scale to rad/s */
	speed *= x_scale;
	
	entry.type = COMMAND_PTYPE_SPEED;
	entry.value.f = speed;
	return command_send(sp_cmd, &entry);
}

int update_y(struct command *sp_cmd, int value)
{
	float speed;
	struct command_entry entry;
	
	speed = (float)value;
	
	/* make the joystick logaritmic */
	speed = speed * speed / 32768.0;
	if (value < 0)
		speed *= -1.0;
	
	/* Scale to rad/s */
	speed *= y_scale;
	
	entry.type = COMMAND_PTYPE_SPEED;
	entry.value.f = speed;
	return command_send(sp_cmd, &entry);
}

int main(int argc, char **argv)
{
	int ret;
	fd_set fdset_rx;
	struct js_event event;
	int fdjoy;
	int optval;
	
	printf("Controller joystick interface starting...\n");
	
	signal(SIGPIPE, SIG_IGN);

	config_add("x_command_host", config_string, &x_command_host);
	config_add("x_command_port", config_int, &x_command_port);
	config_add("x_scale", config_double, &x_scale);
	config_add("x_command_spg", config_string, &x_command_spg);
	config_add("y_command_host", config_string, &x_command_host);
	config_add("y_command_port", config_int, &x_command_port);
	config_add("y_scale", config_double, &y_scale);
	config_add("y_command_spg", config_string, &x_command_spg);

	if (argc >= 2) {
		printf("Reading config file '%s'\n", argv[1]);
		config_read(argv[1]);
	}
	
	sp_command_x = command_open_simple(x_command_host, x_command_port,
	    x_command_spg, "console/joystick", COMMAND_VALUE_TYPE_FLOAT);
	if (!sp_command_x) {
		printf("Setpoint generator %s does not exist\n", x_command_spg);
	}
	sp_command_y = command_open_simple(y_command_host, y_command_port,
	    y_command_spg, "console/joystick", COMMAND_VALUE_TYPE_FLOAT);
	if (!sp_command_y) {
		printf("Setpoint generator %s does not exist\n", x_command_spg);
	}

	fdjoy = open(joystick_device, O_RDONLY);
	if (fdjoy < 0) {
		perror("Could not open joystick device");
		return 1;
	}
	optval = 1;
	ioctl(fdjoy, FIONBIO, &optval);
	
	printf("Starting main loop.\n");

	do {
		FD_ZERO(&fdset_rx);
		
		FD_SET(fdjoy, &fdset_rx);
			
		ret = select(fdjoy + 1, &fdset_rx, NULL, NULL, NULL);
		
		ret = read(fdjoy, &event, sizeof(event));
		if (ret == sizeof(event)) {
			if (event.type & JS_EVENT_AXIS && event.number == 0) {
				if (sp_command_x)
					update_x(sp_command_x, event.value);
			}
			if (event.type & JS_EVENT_AXIS && event.number == 1) {
				if (sp_command_y)
					update_y(sp_command_y, event.value);
			}
		} else if (errno != EAGAIN) {
			break;
		}
	} while (1);

	return 0;
}
