/*
	Command generator for satellite positions

	Copyright Jeroen Vreeken (pe1rxq@amsat.org), 2011, 2014
	Copyright Stichting C.A. Muller Radioastronomiestation, 2011

	This program is free software: you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation, either version 3 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program.  If not, see <http://www.gnu.org/licenses/>.

*/

#include <time.h>
#include <sys/time.h>
#include <unistd.h>
#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <string.h>
#include <fcntl.h>
#include <termios.h>
#include <signal.h>
#include <ctype.h>
#include <inttypes.h>
#include <arpa/inet.h>
#include <sys/select.h>
#include <sys/ioctl.h>
#include <errno.h>
#include <stdbool.h>

#include <libnova/julian_day.h>
#include <libnova/refraction.h>

#include <command/command.h>
#include <utils/weather.h>
#include <utils/dt_model.h>
#include <utils/location.h>
#include <predictlib/predictlib.h>

#include <dt_port_numbers.h>
#include <utils/dt_host.h>
#include <utils/command_server.h>
#include <utils/status_server.h>
#include <log/log.h>

#define TIME_OFFSET 2
#define EL_HORIZON -1.0
#define NEXTPASS_RANGE_T (3600 * 24)
#define COORD_NR 50


char *command_host = "localhost";
int command_port = CONSOLE_COMMAND_PORT;
int stat_port = CONSOLE_SAT_STAT_PORT;
int cmd_port = CONSOLE_SAT_CMD_PORT;
char *az_command_spg = "Azimuth_Setpoint";
char *el_command_spg = "Elevation_Setpoint";

char tle_str[139] =
"1 25544U 98067A   11346.36157800  .00019501  00000-0  25689-3 0  6081"
"2 25544  51.6432 344.3748 0024836 155.4957 344.3191 15.58092544748721";
char *name = NULL;

struct location *location;
struct predict *predict = NULL;

static bool passknown = false;
time_t delay;
static double az_aos = 0;
static time_t t_aos;
static time_t t_los;
struct status_server *stat_srv;

void nextpass(struct predict *predict)
{
	time_t t, tnow = time(NULL);
	double JD;
	struct ln_date date;
	int i;
	double prev_az, az, el;
	bool aos = false, los = false;
	bool cross_west = false, cross_east = false;
	int t_west_1 = 0, t_east_1 = 0, t_west_2 = 0, t_east_2 = 0;
	time_t t_step;
	char statline[2000];
	char *statlinepos;
	
	if (delay > tnow)
		return;
	
	for (i = TIME_OFFSET; i < NEXTPASS_RANGE_T; i++) {
		t = tnow + i;
		ln_get_date_from_timet(&t, &date);
		JD = ln_get_julian_day(&date);

		predict_calc_azel(predict, JD, &az, &el);
		
		az -= 180.0;
		
		if (el >= EL_HORIZON && !aos) {
			log_send(LOG_T_INFO, 
			    "Next pass starts in %d seconds, az=%f", i, az);
			aos = true;
			prev_az = az;
			az_aos = az;
			t_aos = tnow + i;
		}
		if (el < EL_HORIZON && aos) {
			log_send(LOG_T_INFO,
			    "Next pass ends in %d seconds", i);
			los = true;
			t_los = tnow + i;
			break;
		}
		
		if (aos && !los) {
//			printf("%d az: %f\n", (int)t, az);
			/* this is a pass */
			if (az < -90.0 && prev_az > 90.0) {
				/* crossed north */
				prev_az -= 360.0;
			}
			if (az > 90.0 && prev_az < -90.0) {
				/* crossed north */
				prev_az += 360.0;
			}
			if (az <= 90.0 && prev_az > 90.0)
				cross_west = true;
			if (az >= -90.0 && prev_az < -90.0)
				cross_east = true;
			if (cross_east) {
				t_west_2++;
			} else {
				t_west_1++;
			}
			if (cross_west) {
				t_east_2++;
			} else {
				t_east_1++;
			}
		}
	}
	
	if (!aos) {
		log_send(LOG_T_WARNING,
		    "No pass within %d seconds, will calculate again after %d seconds", 
		    NEXTPASS_RANGE_T, NEXTPASS_RANGE_T / 2);
		delay = tnow +  NEXTPASS_RANGE_T / 2;
		if (stat_srv)
			status_server_send(stat_srv, "pass=unknown\n");
		return;
	}
	delay = 0;
	
	log_send(LOG_T_INFO,
	    "cross_west: %d, cross_east: %d, t_w_1: %d, t_w_2: %d, t_e_1: %d, t_e_2: %d",
	    cross_west, cross_east, t_west_1, t_west_2, t_east_1, t_east_2);
	if (t_west_1 >= t_west_2 &&
	    t_west_1 >= t_east_1 &&
	    t_west_1 >= t_east_2) {
		/* if we begin with start az we might go through east late */
		if (az_aos < -90) {
			az_aos += 360;
		}
	} else if (t_west_2 >= t_west_1 &&
	    t_west_2 >= t_east_1 &&
	    t_west_2 >= t_east_2) {
		/* if we begin with start az we go through east in wrong way first */
		az_aos += 360;
	} else if (t_east_1 >= t_west_1 &&
	    t_east_1 >= t_west_2 &&
	    t_east_1 >= t_east_2) {
		/* if we begin with start az we might go through west late */
		if (az_aos > 90) {
			az_aos -= 360;
		}
	} else {
		az_aos -= 360;
	}
	
	if (az_aos > 270.0)
		az_aos = 270.0;
	if (az_aos < -270.0)
		az_aos = -270.0;

	t_step = (t_los - t_aos) / COORD_NR;
	if (!t_step)
		t_step = 1;

	statlinepos = statline;
	statlinepos += sprintf(statlinepos, "pass=");
	
	for (i = t_aos; i < t_los; i += t_step) {
		t = i;
		ln_get_date_from_timet(&t, &date);
		JD = ln_get_julian_day(&date);

		predict_calc_azel(predict, JD, &az, &el);
		az -= 180.0;
		statlinepos += sprintf(statlinepos, "%lld:%d:%d ", 
		    (long long)t, (int)round(az), (int)round(el));
	}
	sprintf(statlinepos, "\n");

	if (stat_srv)
		status_server_send(stat_srv, statline);

	passknown = true;
}

bool set_tle(char *newtle)
{
	double lat, lon, alt;
	char processed_tle[139];
	char *new_name;

	lat = location_get_latitude(location);
	lon = location_get_longitude(location);
	alt = location_get_altitude(location);
	log_send(LOG_T_DEBUG, "Using lat: %f, lon: %f, alt: %f",
	    lat, lon, alt);
	
	if (predict_tle_check(newtle, processed_tle, &new_name)) {
		log_send(LOG_T_ERROR, "Invallid TLE");
		return false;
	}
	
	strncpy(tle_str, processed_tle, 139);
	if (name) {
		free(name);
		name = NULL;
	}
	if (new_name)
		name = new_name;
	else
		name = strdup("unknown");
	
	log_send(LOG_T_INFO, "Switching to new TLE for satellite '%s'",
	    name ? name : "unknown");
	
	if (predict)
		predict_free(predict);
	predict = predict_create(tle_str, lat, lon, alt);
	passknown = false;
	delay = 0;
	nextpass(predict);
	
	return true;
}



bool refraction_enable = false;
bool dt_model_enable = false;
bool switch_enabled = false;
bool switch_cmd = false;

struct compensation_switch {
	char *name;
	bool *value;
};

struct compensation_switch compensation_switches[] = {
	{ "enabled", &switch_enabled },
	{ "refraction", &refraction_enable },
	{ "dt_model", &dt_model_enable },
	{ NULL, NULL }
};

struct weather *weather;

struct lnh_lnlat_posn dwingeloo;
struct ln_lnlat_posn dwingeloo_pos;
struct ln_equ_posn object;
struct ln_equ_posn object_prec;
struct ln_equ_posn object_aber;
struct ln_hms track_hms;
struct ln_dms track_dms;


int handle_cmd(char *name, char *val)
{
	int i;

	log_send(LOG_T_DEBUG, "name %s value %s", name, val);
	for (i = 0; compensation_switches[i].name; i++) {
		if (!strcmp(name, compensation_switches[i].name)) {
			log_send(LOG_T_DEBUG, "found switch %s", name);
			if (val[0] == '1') {
				*compensation_switches[i].value = true;
				if (!strcmp(name, "enabled")) {
					switch_cmd = true;
				}
			} else {
				*compensation_switches[i].value = false;
			}
		}
	}
	if (!strcmp(name, "tle")) {
		if (!set_tle(val))
			switch_enabled = false;
	}

	return 0;
}

void output(struct status_server *stat_srv)
{
	static time_t last = 0;
	time_t now;
	char statline[800];
	int ret, i;
	char *statlinepos;
	
	now = time(NULL);
	if (now <= last)
		return;
	
	last = now;
	
	statlinepos = statline;
	for (i = 0; compensation_switches[i].name; i++) {
		if (i) {
			ret = sprintf(statlinepos, ",");
			statlinepos += ret;
		}
		ret = sprintf(statlinepos,
		    "%s=%d",
		    compensation_switches[i].name,
		    *compensation_switches[i].value);
		statlinepos += ret;
	}
	if (name) {
		ret = sprintf(statlinepos, ",name=%s", name);
		statlinepos += ret;
	}
	sprintf(statlinepos, ",tle=%s\n", tle_str);

	status_server_send(stat_srv, statline);
}


int main(int argc, char **argv)
{
	struct command *sp_command_az = NULL;
	struct command *sp_command_el = NULL;
	struct command_server *cmd_srv;
	time_t lastt = 0;
	double last_az = 0.0;
	
	dt_model_init();

	location = location_create("Dwingeloo");
	if (!location) {
		printf("Could not get location\n");
		return -1;
	}

	
	set_tle(tle_str);

	signal(SIGPIPE, SIG_IGN);

	if (argc == 2) {
		command_host = argv[argc - 1];
	}

	log_client_start(dt_host_console(), CONSOLE_LOG_PORT_IN, 
	    LOG_T_DEBUG, LOG_T_INFO, "console/sattracker");

	sp_command_az = command_open_simple(command_host, command_port,
	    az_command_spg, "console/sattracker",
	    COMMAND_VALUE_TYPE_FLOAT);
	sp_command_el = command_open_simple(command_host, command_port,
	    el_command_spg, "console/sattracker",
	    COMMAND_VALUE_TYPE_FLOAT);
	if (!sp_command_az || !sp_command_el) {
		log_send(LOG_T_ERROR, "Setpoint generator(s) not found");
		exit(-1);
	}

	cmd_srv = command_server_create(cmd_port, 0, 100);
	if (!cmd_srv) {
		log_send(LOG_T_ERROR,
		    "Could not open listen port for commands");
		return -1;
	} else {
		command_server_handler_set(cmd_srv, handle_cmd);
	}

	stat_srv = status_server_create(stat_port, 0, 100);
	if (!stat_srv) {
		log_send(LOG_T_ERROR,
		    "Could not open listen port for status");
		return -1;
	}

	weather = weather_create("Dwingeloo");
	if (!weather) {
		log_send(LOG_T_ERROR, "Could not create weather handle");
		exit(-1);
	}

	while (1)
	{
		time_t t;
		struct ln_date date;
		double JD;
		int high = 0;
		fd_set fdset_rx;
		struct timeval tv;
		double pressure, temperature;
		double alt_adj;
		
		t = time(NULL);
		if (t != lastt) {
			double az, el;
			
			lastt = t;
			t += TIME_OFFSET;
		
			ln_get_date_from_timet(&t, &date);
			JD = ln_get_julian_day(&date);

			predict_calc_azel(predict, JD, &az, &el);

			/* predict has an az range from 0 to 360 with 0 being South.
			   Dwingeloo has an az range from -270 to 270 with 0 being South
			 */
			az -= 180.0;
			
			if (el < EL_HORIZON) {
				if (!passknown)
					nextpass(predict);
				
				/* large step? */
				if (fabs(az_aos - az) > 180.0) {
					log_send(LOG_T_DEBUG,
					    "Preset to aos position %f -> %f", 
					    az, az_aos);
					az += copysign(360.0, az_aos);
				}
			} else {
				/* check if we might have to compensate 360 
				   degrees to prevent large az steps
				 */
				log_send(LOG_T_DEBUG,
				    "fabs(%f - %f): %f", last_az,az,fabs(last_az- az));
				if (az > 90.0 || az < -90.0) {
					/* large step? */
					if (fabs(last_az - az) > 200.0) {
						log_send(LOG_T_DEBUG,
						    "Swapping AZ");
						az += copysign(360.0, last_az);
					}
				}
				passknown = false;
			}
			
			if (t > t_los)
				passknown = false;
						
			if (refraction_enable) {
				/* Get current weather */
				temperature = weather_get_temperature(weather);
				pressure = weather_get_pressure(weather);
				/* Refraction adjustment */
				alt_adj = ln_get_refraction_adj(el, pressure, temperature);
				log_send(LOG_T_DEBUG,
				    "Refraction: %f @ %fC,%fmbar", alt_adj, temperature, pressure);
				el += alt_adj;
			}
			
			if (dt_model_enable) {
				double daz, del;
				
				daz = dt_model_azimuth_delta(az, el);
				del = dt_model_elevation_delta(az, el);
				log_send(LOG_T_DEBUG, "model: %g %g", daz, del);
				az += daz;
				el += del;
			}

			if (!switch_cmd && fabs(last_az - az) > 90.0) {
				if (switch_enabled)
					log_send(LOG_T_WARNING,
					    "Disabling tracker to prevent large azimuth jump");
				switch_enabled = false;
			}

			last_az = az;
			
			/* Temporary limit on elevation */
			if (el < 0.0)
				el = 0.0;
			
			log_send(LOG_T_DEBUG, "Setpoint: %d %f %f %f",
			    (unsigned int)t, JD, az, el);

			if (switch_enabled) {
				struct command_entry entry;
				
				entry.type = COMMAND_PTYPE_SETPOINT_TIME;
				entry.t.tv_sec = t;
				entry.t.tv_nsec = 0;
				entry.value.f = az * 2 * M_PI / 360.0;
				command_send(sp_command_az, &entry);
				entry.value.f = el * 2 * M_PI / 360.0;
				command_send(sp_command_el, &entry);
			} else {
				struct command_entry entry;
				/* not enabled, send idle just in case we were
				   still moving */
				   
				entry.type = COMMAND_PTYPE_SPEED;
				entry.value.f = 0.0;
				command_send(sp_command_az, &entry);
				command_send(sp_command_el, &entry);
			}
			switch_cmd = false;
		}

		FD_ZERO(&fdset_rx);
	
		command_server_fdset_add(cmd_srv, &fdset_rx, &high);
		status_server_fdset_add(stat_srv, &fdset_rx, &high);
		
		tv.tv_sec = 1;
		tv.tv_usec = 0;

		select(high, &fdset_rx, NULL, NULL, &tv);

		command_server_fdset_handle(cmd_srv, &fdset_rx);
		status_server_fdset_handle(stat_srv, &fdset_rx);
		
		output(stat_srv);
	}
}
