#include "sgp4sdp4.h"
#include "predictlib.h"

struct predict {
	sat_t sat;
	double lat;
	double lon;
	double alt;
};

/** \brief SGP4SDP4 driver for doing AOS/LOS calculations.
 *  \param sat Pointer to the satellite data.
 *  \param qth Pointer to the QTH data.
 *  \param t The time for calculation (Julian Date)
 *
 */
void
predict_calc (struct predict *predict, double t)
{
	sat_t *sat = &predict->sat;

    obs_set_t     obs_set;
    geodetic_t    sat_geodetic;
    geodetic_t    obs_geodetic;
    double        age;


    obs_geodetic.lon = predict->lon * de2ra;
    obs_geodetic.lat = predict->lat * de2ra;
    obs_geodetic.alt = predict->alt / 1000.0;
    obs_geodetic.theta = 0;


    sat->jul_utc = t;
    sat->tsince = (sat->jul_utc - sat->jul_epoch) * xmnpda;

    /* call the norad routines according to the deep-space flag */
    if (sat->flags & DEEP_SPACE_EPHEM_FLAG)
        SDP4 (sat, sat->tsince);
    else
        SGP4 (sat, sat->tsince);

    Convert_Sat_State (&sat->pos, &sat->vel);

    /* get the velocity of the satellite */
    Magnitude (&sat->vel);
    sat->velo = sat->vel.w;
    Calculate_Obs (sat->jul_utc, &sat->pos, &sat->vel, &obs_geodetic, &obs_set);
    Calculate_LatLonAlt (sat->jul_utc, &sat->pos, &sat_geodetic);

    while (sat_geodetic.lon < -pi)
        sat_geodetic.lon += twopi;

    while (sat_geodetic.lon > (pi))
        sat_geodetic.lon -= twopi;

    sat->az = Degrees (obs_set.az);
    sat->el = Degrees (obs_set.el);
    sat->range = obs_set.range;
    sat->range_rate = obs_set.range_rate;
    sat->ssplat = Degrees (sat_geodetic.lat);
    sat->ssplon = Degrees (sat_geodetic.lon);
    sat->alt = sat_geodetic.alt;
    sat->ma = Degrees (sat->phase);
    sat->ma *= 256.0/360.0;
    sat->phase = Degrees (sat->phase);

    /* same formulas, but the one from predict is nicer */
    //sat->footprint = 2.0 * xkmper * acos (xkmper/sat->pos.w);
    sat->footprint = 12756.33 * acos (xkmper / (xkmper+sat->alt));
    age = sat->jul_utc - sat->jul_epoch;
    sat->orbit = (long) floor((sat->tle.xno * xmnpda/twopi +
                    age * sat->tle.bstar * ae) * age +
                    sat->tle.xmo/twopi) + sat->tle.revnum - 1;
}

void gtk_sat_data_init_sat (struct predict *predict)
{
    geodetic_t obs_geodetic;
    obs_set_t obs_set;
    geodetic_t sat_geodetic;
    double jul_utc, age;
	sat_t *sat = &predict->sat;

        sat->flags = 0;

        select_ephemeris (sat);

        /* initialise variable fields */
        sat->jul_utc = 0.0;
        sat->tsince = 0.0;
        sat->az = 0.0;
        sat->el = 0.0;
        sat->range = 0.0;
        sat->range_rate = 0.0;
        sat->ra = 0.0;
        sat->dec = 0.0;
        sat->ssplat = 0.0;
        sat->ssplon = 0.0;
        sat->alt = 0.0;
        sat->velo = 0.0;
        sat->ma = 0.0;
        sat->footprint = 0.0;
        sat->phase = 0.0;
        sat->aos = 0.0;
        sat->los = 0.0;

    jul_utc = Julian_Date_of_Epoch (sat->tle.epoch); // => tsince = 0.0
    sat->jul_epoch = jul_utc;

    /* initialise observer location */
    obs_geodetic.lon = 0.0;
    obs_geodetic.lat = 0.0;
    obs_geodetic.alt = 0.0;
    obs_geodetic.theta = 0;


    /* execute computations */
    if (sat->flags & DEEP_SPACE_EPHEM_FLAG)
        SDP4 (sat, 0.0);
    else
        SGP4 (sat, 0.0);

    /* scale position and velocity to km and km/sec */
    Convert_Sat_State (&sat->pos, &sat->vel);

    /* get the velocity of the satellite */
    Magnitude (&sat->vel);
    sat->velo = sat->vel.w;
    Calculate_Obs (jul_utc, &sat->pos, &sat->vel, &obs_geodetic, &obs_set);
    Calculate_LatLonAlt (jul_utc, &sat->pos, &sat_geodetic);

    while (sat_geodetic.lon < -pi)
        sat_geodetic.lon += twopi;
    
    while (sat_geodetic.lon > (pi))
        sat_geodetic.lon -= twopi;

    sat->az = Degrees (obs_set.az);
    sat->el = Degrees (obs_set.el);
    sat->range = obs_set.range;
    sat->range_rate = obs_set.range_rate;
    sat->ssplat = Degrees (sat_geodetic.lat);
    sat->ssplon = Degrees (sat_geodetic.lon);
    sat->alt = sat_geodetic.alt;
    sat->ma = Degrees (sat->phase);
    sat->ma *= 256.0/360.0;
    sat->footprint = 2.0 * xkmper * acos (xkmper/sat->pos.w);
    age = 0.0;
    sat->orbit = (long) floor((sat->tle.xno * xmnpda/twopi +
                               age * sat->tle.bstar * ae) * age +
                              sat->tle.xmo/twopi) + sat->tle.revnum - 1;

    /* orbit type */
//    sat->otype = get_orbit_type (sat);
}

struct predict *predict_create(char *tle, double lat, double lon, double alt)
{
	struct predict *predict;
	char *tle_dup = strdup(tle);
	int i;
	
	predict = calloc(sizeof(struct predict), 1);
	
	for (i = 0; tle[i] != '1' && i < strlen(tle); i++);
	tle += i;
	
        Convert_Satellite_Data (tle, &predict->sat.tle);

	predict->lat = lat;
	predict->lon = lon;
	predict->alt = alt;
	
	gtk_sat_data_init_sat (predict);

	free(tle_dup);
	
	return predict;
}

bool predict_tle_check(char *tle, char processed_tle[139], char **name)
{
	char tmp_tle[strlen(tle)+1];
	int i;
	
	strcpy(tmp_tle, tle);
	for (i = 0; i < strlen(tmp_tle); i++) {
		if (tmp_tle[i] == '\n' ||
		    tmp_tle[i] == '\r') {
			strcpy(tmp_tle + i, tmp_tle + i + 1);
			i--;
		}
	}
	
	for (i = 0; i < (int)strlen(tmp_tle) - 137; i++) {
		if (Good_Elements(tmp_tle + i)) {
			if (processed_tle) {
				memcpy(processed_tle, tmp_tle + i, 138);
				processed_tle[138] = 0;
			}
			if (name && i) {
				*name = malloc(i+1);
				if (*name) {
					memcpy(*name, tmp_tle, i);
					(*name)[i] = 0;
					while (i && (
					    (*name)[i-1] == ' ' ||
					    (*name)[i-1] == '\t')) {
						(*name)[i-1] = 0;
						i--;
					}
				}
			} else {
				*name = NULL;
			}
			return false;
		}
	}
	return true;
}

void predict_free(struct predict *predict)
{
	free(predict);
}

/* calculate for a given julian date */
int predict_calc_azel(struct predict *predict, double t, double *az, double *el)
{
	predict_calc(predict, t);

	*az = predict->sat.az;
	*el = predict->sat.el;
	
	return 0;
}


