/* 
	Controller block for interfacing with a winrotor-usb rotor interface.

	Copyright Jeroen Vreeken (pe1rxq@amsat.org), 2007
	Copyright Stichting C.A. Muller Radioastronomiestation, 2007

	This program is free software: you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation, either version 3 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program.  If not, see <http://www.gnu.org/licenses/>.

*/

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>

#include "controller_block.h"
#include "winrotor-usb.h"

#include <usb.h>

#define VEND_AKMOD	0x0a2c
#define DEV_COMPULAB	0x0002

static struct usb_dev_handle *akmod_compulab_open(void)
{
	struct usb_bus *usb_bus;
	struct usb_device *dev;
	struct usb_dev_handle *handle = NULL;
	
	usb_init();
	usb_find_busses();
	usb_find_devices();
	
	for (usb_bus = usb_busses; usb_bus; usb_bus = usb_bus->next) {
		for (dev = usb_bus->devices; dev;  dev = dev->next) {
			if (dev->descriptor.idVendor == VEND_AKMOD &&
			    dev->descriptor.idProduct == DEV_COMPULAB) {
				handle = usb_open(dev);
				if (handle) {
					int ret;
					
					ret = usb_claim_interface(handle, 0);
					if (ret) {
						perror("usb_claim_interface()");
						usb_close(handle);
						return NULL;
					}
					return handle;
				} else
					return NULL;
			}
		}
	}
	return NULL;
}

static void akmod_compulab_close(struct usb_dev_handle *handle)
{
	usb_release_interface(handle, 0);
	usb_close(handle);
}

static int akmod_compulab_init(struct usb_dev_handle *handle)
{
	int request = 0x03;
	int requesttype = USB_ENDPOINT_IN | USB_TYPE_VENDOR | USB_RECIP_OTHER;
	int value = 0x34;
	int index = 0x80;
	char iobuf[8];
	int timeout = 1000;

	usb_control_msg(handle, requesttype,
	    request, value, index, 
	    iobuf, sizeof(iobuf), 
	    timeout);

	return 0;
}

static int akmod_compulab_set_get(struct usb_dev_handle *handle,
    unsigned char port, float *a, float *b)
{
	int request = 0x06;
	int requesttype = USB_ENDPOINT_IN | USB_TYPE_VENDOR | USB_RECIP_OTHER;
	int value = ~ port;
	int index = 0x08;
	unsigned char iobuf[8];
	int timeout = 1000;

	usb_control_msg(handle, 
	    requesttype, 
	    request, 
	    value, 
	    index, 
	    (char *)iobuf, 
	    sizeof(iobuf), 
	    timeout);

	*a = iobuf[2] * 8 + (iobuf[4] & 0x0f);
	*b = iobuf[3] * 8 + ((iobuf[4] & 0xf0) / 16);
	*a /= 2048;
	*b /= 2048;

	return 0;
}

struct controller_block_private {
	float *az_in;
	float *el_in;
	float az_out;
	float el_out;

	float az_gain;
	float az_offset;
	float el_gain;
	float el_offset;

	int enabled;

	struct usb_dev_handle *hdl;
};

static void calculate(struct controller_block *wru)
{
	unsigned char port = 0xff;

	if (*wru->private->az_in > 0.0)
		port &= ~0x01;
	else if (*wru->private->az_in < 0.0)
		port &= ~0x02;
	
	if (*wru->private->el_in > 0.0)
		port &= ~0x10;
	else if (*wru->private->el_in < 0.0)
		port &= ~0x20;
	
	if (!wru->private->enabled)
		port = 0xff;

	akmod_compulab_set_get(wru->private->hdl, port,
	    &wru->private->az_out, &wru->private->el_out);

	wru->private->az_out += wru->private->az_offset;
	wru->private->az_out *= wru->private->az_gain;
	wru->private->el_out += wru->private->el_offset;
	wru->private->el_out *= wru->private->el_gain;
}

static struct controller_block_param_list params[] = {
	{ "az_gain"   },
	{ "az_offset" },
	{ "el_gain"   },
	{ "el_offset" },
	{ "az_out"    },
	{ "el_out"    },
	{ "enabled"   },
	{ NULL        },
};

static void param_get(struct controller_block *wru, int param, void *val)
{
	switch (param) {
		case 0:
			*(float*)val = wru->private->az_gain;
			break;
		case 1:
			*(float*)val = wru->private->az_offset;
			break;
		case 2:
			*(float*)val = wru->private->el_gain;
			break;
		case 3:
			*(float*)val = wru->private->el_offset;
			break;
		case 4:
			*(float*)val = wru->private->az_out;
			break;
		case 5:
			*(float*)val = wru->private->el_out;
			break;
		case 6:
			*(int*)val = wru->private->enabled;
			break;
	}
}

static void param_set(struct controller_block *wru, int param, void *val)
{
	switch (param) {
		case 0:
			wru->private->az_gain = *(float*)val;
			break;
		case 1:
			wru->private->az_offset = *(float*)val;
			break;
		case 2:
			wru->private->el_gain = *(float*)val;
			break;
		case 3:
			wru->private->el_offset = *(float*)val;
			break;
		case 4:
			wru->private->az_out = *(float*)val;
			break;
		case 5:
			wru->private->el_out = *(float*)val;
			break;
		case 6:
			wru->private->enabled = *(int*)val;
			break;
	}
}

struct controller_block * block_winrotor_usb_create(char *name)
{
	struct usb_dev_handle *hdl;
	struct controller_block *wru;
	
	hdl = akmod_compulab_open();
	if (!hdl)
		return NULL;
	akmod_compulab_init(hdl);
	
	wru = malloc(sizeof(struct controller_block));
	if (!wru)
		goto err_usb;

	wru->type = "winrotor_usb";
	wru->name = malloc(strlen(name)+1);
	if (!wru->name)
		goto err_wru;
	strcpy(wru->name, name);

	wru->private = malloc(sizeof(struct controller_block_private));
	if (!wru->private)
		goto err_name;
	wru->private->az_gain = 0.0;
	wru->private->el_gain = 0.0;
	wru->private->az_offset = 0.0;
	wru->private->el_offset = 0.0;
	wru->private->enabled = 0;
	wru->private->hdl = hdl;

	wru->inputs = 2;
	wru->input = malloc(sizeof(struct controller_block_interm) * 2);
	if (!wru->input)
		goto err_private;
	wru->input[0].name = "az_in";
	wru->input[0].type = CONTROLLER_BLOCK_TERM_FLOAT;
	wru->input[0].value.f = &wru->private->az_in;
	wru->input[0].ghostof = NULL;
	wru->input[1].name = "el_in";
	wru->input[1].type = CONTROLLER_BLOCK_TERM_FLOAT;
	wru->input[1].value.f = &wru->private->el_in;
	wru->input[1].ghostof = NULL;
	
	wru->outputs = 2;
	wru->output = malloc(sizeof(struct controller_block_outterm) * 2);
	if (!wru->output)
		goto err_input;
	wru->output[0].name = "az_out";
	wru->output[0].type = CONTROLLER_BLOCK_TERM_FLOAT;
	wru->output[0].value.f = &wru->private->az_out;
	wru->output[0].source = wru;
	wru->output[1].name = "el_out";
	wru->output[1].type = CONTROLLER_BLOCK_TERM_FLOAT;
	wru->output[1].value.f = &wru->private->el_out;
	wru->output[1].source = wru;
	
	wru->calculate = calculate;

	if (controller_block_param_list_init(wru, params))
		goto err_output;
	
	wru->param_get = param_get;
	wru->param_set = param_set;

	akmod_compulab_set_get(hdl, 0xff, 
	    &wru->private->az_out, &wru->private->el_out);
	wru->private->az_out = 0.0;
	wru->private->az_out = 0.0;

	controller_block_add(wru);

	return wru;

err_output:
	free(wru->output);
err_input:
	free(wru->input);
err_private:
	free(wru->private);
err_name:
	free(wru->name);
err_wru:
	free(wru);
err_usb:
	akmod_compulab_close(hdl);

	return NULL;
}
