/*
	Copyright Jeroen Vreeken (pe1rxq@amsat.org), 2007
	Copyright Stichting C.A. Muller Radioastronomiestation, 2007

	This program is free software: you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation, either version 3 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program.  If not, see <http://www.gnu.org/licenses/>.

 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <controller/controller_block.h>
#include <block/block_filter_iir.h>

/*
       inputs      outputs
       nr name     nr name

     ----------------------
     |                    |
  ---| 0  in       0  out |----
     |	                  |
     ----------------------
     
 */

struct controller_block_private {
	float *input0;
	float out;
	
	int ind;
	int len;
	float (*transfer)[4];
	
	struct block_filter_iir_param param;
};

static void filter_iir_calculate(struct controller_block *iir)
{
	struct controller_block_private *priv = iir->private;
	float input0 = *priv->input0;
	float (*transfer)[4] = priv->transfer;
	int i;
	int ind, len;
	float out;

	ind = priv->ind;
	transfer[ind][2] = input0;
	len = priv->len;
	ind++;
	ind = ind % len;
	
	priv->ind = ind;

	out = 0;
	for (i = 0; i < len; i++) {
		int j = (i + ind) % len;
		
		out += transfer[i][1] * transfer[j][2];
		out += transfer[i][0] * transfer[j][3];
	}
	transfer[ind][3] = out;

	priv->out = out;
}

static struct controller_block_param_list params[] = {
	{ "transfer", true },
	{ "reset",    true },
	{ NULL },
};

static void param_get(struct controller_block *iir, int param, void *val)
{
	switch (param) {
		case 0:
			memcpy(val, &iir->private->param,
			    sizeof(struct block_filter_iir_param));
			break;
		case 1:
			break;
	}
}

static void param_set(struct controller_block *iir, int param, va_list ap)
{
	int i;
	int new_param = 0;
	
	switch (param) {
		case 0: {
			int len;
			float gain;
			float *poles;
			float *zeros;
			
			len = va_arg(ap, int);
			gain = va_arg(ap, double);
			poles = va_arg(ap, float *);
			zeros = va_arg(ap, float *);
			
			iir->private->param.len = len;
			iir->private->param.gain = gain;
			iir->private->param.poles = poles;
			iir->private->param.zeros = zeros;
			
			new_param = 1;
			break;
		}
		case 1:
			break;
	}

	if (new_param) {
		if (iir->private->transfer)
			free(iir->private->transfer);
		
		iir->private->len = iir->private->param.len + 1;
		iir->private->transfer = malloc(sizeof(float[4]) *
		     iir->private->len);
		if (!iir->private->transfer) {
			iir->private->len = 0;
			return;
		}
		iir->private->transfer[0][0] = 0;
		iir->private->transfer[iir->private->param.len ][1] =
		    1.0 / iir->private->param.gain;
		for (i = 0; i < iir->private->param.len; i++) {
			iir->private->transfer[i + 1][0] =
			    iir->private->param.poles[i];
			iir->private->transfer[i][1] =
			    iir->private->param.zeros[i] / iir->private->param.gain;
		}
		for (i = 0; i < iir->private->len; i++) {
			iir->private->transfer[i][2] = 0;
			iir->private->transfer[i][3] = 0;
		}
		iir->private->ind = 0;
		iir->private->out = 0;
	}
}

static struct controller_block_interm_list interms[] = {
	{ "in", CONTROLLER_BLOCK_TERM_FLOAT, offsetof(struct controller_block_private, input0) },
	{ NULL }
};

static struct controller_block_outterm_list outterms[] = {
	{ "out", CONTROLLER_BLOCK_TERM_FLOAT, offsetof(struct controller_block_private, out) },
	{ NULL }
};

struct controller_block * block_filter_iir_create(char *name)
{
	struct controller_block *iir;
	
	if (!(iir = controller_block_alloc("filter_iir", name, sizeof(struct controller_block_private))))
		return NULL;

	iir->private->out = 0.0;
	iir->private->len = 0;
	iir->private->transfer = NULL;
	iir->private->ind = 0;

	if (controller_block_interm_list_init(iir, interms))
		goto err_block;

	if (controller_block_outterm_list_init(iir, outterms))
		goto err_block;

	iir->calculate = filter_iir_calculate;

	if (controller_block_param_list_init(iir, params))
		goto err_block;

	iir->param_get = param_get;
	iir->param_set = param_set;

	controller_block_add(iir);
	return iir;

err_block:
	controller_block_free(iir);
	return NULL;
}
