/*
	Copyright Jeroen Vreeken (pe1rxq@amsat.org), 2011

	This program is free software: you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation, either version 3 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program.  If not, see <http://www.gnu.org/licenses/>.

 */

#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include <string.h>
#include <sys/ioctl.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>

#include <controller/controller_block.h>

#include <linux/joystick.h>


struct controller_block_private {
	float	*axes;
	bool	*buttons;

	int fd;
};

static void calculate(struct controller_block *block)
{
	struct controller_block_private *priv = block->private;
	struct js_event event;
	int ret;
	
	do {
		ret = read (priv->fd, &event, sizeof(struct js_event));
		if (ret == sizeof(struct js_event)) {
			if (event.type & JS_EVENT_BUTTON) {
				priv->buttons[event.number] = event.value;
			}
			if (event.type & JS_EVENT_AXIS) {
				priv->axes[event.number] = event.value / 32768.0;
			}
		}
	} while (ret == sizeof(struct js_event));
}

struct controller_block * block_joystick_create(char *name, va_list ap)
{
	struct controller_block *block;
	int fd;
	char *device;
	int nr_axes = 0, nr_buttons = 0;
	int i, ret;
	bool sim = false;
		
	device = va_arg(ap, char *);
	
	fd = open(device, O_RDONLY);
	if (fd < 0) {
		printf("Could not open device %s\n", device);
		printf("Going to simulate a joystick\n");
		sim = true;
		device = "simulated joystick";
		nr_axes = 6;
		nr_buttons = 8;
	} else {
		ret = ioctl(fd, FIONBIO, &(int){1});
		if (ret)
			goto err_ioctl;
		ret = ioctl(fd, JSIOCGAXES, &nr_axes);
		if (ret)
			goto err_ioctl;
		ret = ioctl(fd, JSIOCGBUTTONS, &nr_buttons);
		if (ret)
			goto err_ioctl;
	}
	printf("%s: %d axes, %d buttons\n", device, nr_axes, nr_buttons);
	
	block = malloc(sizeof(struct controller_block));
	if (!block)
		goto err_block;

	block->private = malloc(sizeof(struct controller_block_private));
	if (!block->private)
		goto err_private;

	block->private->axes = calloc(sizeof(float), nr_axes);
	if (!block->private->axes)
		goto err_axes;
	block->private->buttons = calloc(sizeof(bool), nr_buttons);
	if (!block->private->buttons)
		goto err_buttons;

	block->type = sim ? "joystick_sim" : "joystick";
	block->name = strdup(name);

	block->inputs = 0;
	block->input = NULL;
	
	/* Inputs on the hardware are outputs on this software block */
	block->outputs = nr_axes + nr_buttons;
	block->output = malloc(sizeof(struct controller_block_outterm) * 
	    (nr_axes + nr_buttons));
	if (!block->output)
		goto err_output;
	for (i = 0; i < nr_axes; i++) {
		block->output[i].type = CONTROLLER_BLOCK_TERM_FLOAT;
		block->output[i].value.f = &block->private->axes[i];
		block->output[i].source = block;
		block->output[i].name = malloc(16);
		sprintf(block->output[i].name, "axis%d", i);
	}
	for (i = 0; i < nr_buttons; i++) {
		int outnr = i + nr_axes;
		block->output[outnr].type = CONTROLLER_BLOCK_TERM_BOOL;
		block->output[outnr].value.b = &block->private->buttons[i];
		block->output[outnr].source = block;
		block->output[outnr].name = malloc(16);
		sprintf(block->output[outnr].name, "button%d", i);
	}
	
	if (sim) {
		block->calculate = NULL;
	} else {
		block->calculate = calculate;
	}
	block->params = 0;
	block->param = NULL;
	block->param_get = NULL;
	block->param_set = NULL;
	
	block->private->fd = fd;

	controller_block_add(block);
	return block;

err_output:
	free(block->private->buttons);
err_buttons:
	free(block->private->axes);
err_axes:
	free(block->private);
err_private:
	free(block);
err_block:
err_ioctl:
	close(fd);
	return NULL;
}
