/*
	Copyright Jeroen Vreeken (pe1rxq@amsat.org), 2007
	Copyright Stichting C.A. Muller Radioastronomiestation, 2007

	This program is free software: you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation, either version 3 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program.  If not, see <http://www.gnu.org/licenses/>.

 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>

#include <controller/controller_block.h>
#include <block/block_matrix_2x2.h>

/*
       inputs         outputs
       nr name        nr name

     --------------------------
  ---| 0  in0         0  out0 |----
     |                        |
  ---| 1  in1         1  out1 |----
     --------------------------
	
	out0 = in0 * c[0][0] + in1 * c[0][1]
	out1 = in0 * c[1][0] + in1 * c[1][1]
      
 */

struct controller_block_private {
	float *input0;
	float *input1;
	float out0;
	float out1;
	
	struct block_matrix_2x2_param param;	
};

static void matrix_2x2_calculate(struct controller_block *matrix)
{
	struct controller_block_private *priv = matrix->private;
	float input0 = *priv->input0;
	float input1 = *priv->input1;
	float out0, out1;

	out0 =
	    input0 * priv->param.c[0][0] +
	    input1 * priv->param.c[0][1];
	out1 =
	    input0 * priv->param.c[1][0] +
	    input1 * priv->param.c[1][1];
	
	priv->out0 = out0;
	priv->out1 = out1;
}

static struct controller_block_param_list params[] = {
	{ "constants", true },
	{ NULL },
};

static void param_get(struct controller_block *matrix, int param, void *val)
{
	struct controller_block_private *priv = matrix->private;

	switch (param) {
		case 0:
			memcpy(val, &priv->param, sizeof(priv->param));
			break;
	}
}

static void param_set(struct controller_block *matrix, int param, va_list val)
{
	struct controller_block_private *priv = matrix->private;
	float *flt1, *flt2;
	
	switch (param) {
		case 0:
			flt1 = va_arg(val, float *);
			flt2 = va_arg(val, float *);
			priv->param.c[0][0] = flt1[0];
			priv->param.c[0][1] = flt1[1];
			priv->param.c[1][0] = flt2[0];
			priv->param.c[1][1] = flt2[1];
			break;
	}
}

static struct controller_block_interm_list interms[] = {
	{ "in0", CONTROLLER_BLOCK_TERM_FLOAT, offsetof(struct controller_block_private, input0) },
	{ "in1", CONTROLLER_BLOCK_TERM_FLOAT, offsetof(struct controller_block_private, input1) },
	{ NULL }
};
static struct controller_block_outterm_list outterms[] = {
	{ "out0", CONTROLLER_BLOCK_TERM_FLOAT, offsetof(struct controller_block_private, out0) },
	{ "out1", CONTROLLER_BLOCK_TERM_FLOAT, offsetof(struct controller_block_private, out1) },
	{ NULL }
};


struct controller_block * block_matrix_2x2_create(char *name)
{
	struct controller_block *matrix;
	struct block_matrix_2x2_param defaults = {{{ 0.0, 0.0 }, { 0.0, 0.0 }}};
	
	if (!(matrix = controller_block_alloc("matrix_2x2", name, sizeof(struct controller_block_private))))
		return NULL;

	matrix->private->out0 = 0.0;
	matrix->private->out1 = 0.0;
	matrix->private->param = defaults;
	

	if (controller_block_interm_list_init(matrix, interms))
		goto err_private;

	if (controller_block_outterm_list_init(matrix, outterms))
		goto err_input;
	
	matrix->calculate = matrix_2x2_calculate;

	if (controller_block_param_list_init(matrix, params))
		goto err_output;

	matrix->param_get = param_get;
	matrix->param_set = param_set;

	controller_block_add(matrix);
	return matrix;

err_output:
	free(matrix->output);
err_input:
	free(matrix->input);
err_private:
	controller_block_free(matrix);
	return NULL;
}
