/*
	Copyright Jeroen Vreeken (pe1rxq@amsat.org), 2007
	Copyright Stichting C.A. Muller Radioastronomiestation, 2007

	This program is free software: you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation, either version 3 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program.  If not, see <http://www.gnu.org/licenses/>.

 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <controller/controller_block.h>

/*
       inputs         outputs
       nr name        nr name

     --------------------------
     |                        |
  ---| 0  in          0  out  |----
     |	                      |
  ---| 1  enable      1  outp |----
     |	                      |
     |	              2  outi |----
     |	                      |
     |	              3  outd |----
     --------------------------
     
     Schematic:
     (Same as PID but with additional anti-windup)
     
                     ------                           outp         
          in --------| kp |------------------------------------    
                 |   ------                                   |    
                 |                         /-----\            |    
	         |   ------    /---\       |   - |    outi  /---\  
                 ----| ki |----| + |-------|  /  |----------| + |--- out
	         |   ------    \---/       | -   |  |       \---/  
	         |               |         \-----/  |         |    
	         |               |  -----           |         |
                 |               ---|z-1|------------         |
                 |                  -----                     |
                 |                                            |
	         |       /---\    ------              outd    |
                 --------| - |----| kd |-----------------------
	         |       \---/    ------
                 |         |
                 |  -----  |
                 ---|z-1|---
	            -----
 */

struct controller_block_private {
	float *input0;
	bool *enable;
	float out;
	float outp;
	float outi;
	float outd;
	float kp;	/* K factors */
	float ki;
	float kd;
	float t;	/* sample period */
	float ci;	/* Factors used in calculate (K * T) */
	float cd;
	float previous;
	float maxi;
	float mini;
};

static void pid_aw_calculate(struct controller_block *pid)
{
	struct controller_block_private *priv = pid->private;
	float outp, outi, outd;
	float input = *priv->input0;
	float previous = priv->previous;
	float maxi = priv->maxi;
	float mini = priv->mini;
	
	outp = input * priv->kp;

	outi = priv->outi + input * priv->ci;
	outi = fmaxf(fminf(outi, maxi), mini);

	outd = (input - previous ) * priv->cd;
	priv->previous = input;

	if (!*priv->enable) {
		outp = 0;
		outi = 0;
		outd = 0;
		priv->previous = 0;
	}
	
	priv->outp = outp;
	priv->outi = outi;
	priv->outd = outd;
	
	priv->out = outp + outi + outd;
}

static struct controller_block_param_list params[] = {
	{ "kp",    true  },
	{ "ki",    true  },
	{ "kd",    true  },
	{ "t",     true  },
	{ "maxw",  false },
	{ "minw",  false },
	{ "reset", true  },
	{ NULL },
};

static void param_get(struct controller_block *pid, int param, void *val)
{
	switch (param) {
		case 0:
			*(float*)val = pid->private->kp;
			break;
		case 1:
			*(float*)val = pid->private->ki;
			break;
		case 2:
			*(float*)val = pid->private->kd;
			break;
		case 3:
			*(float*)val = pid->private->t;
			break;
		case 4:
			*(float*)val = pid->private->maxi;
			break;
		case 5:
			*(float*)val = pid->private->mini;
			break;
		case 6:
			break;
	}
}

static void param_set(struct controller_block *pid, int param, va_list val)
{
	switch (param) {
		case 0:
			pid->private->kp = va_arg(val, double);
			break;
		case 1:
			pid->private->ki = va_arg(val, double);
			break;
		case 2:
			pid->private->kd = va_arg(val, double);
			break;
		case 3:
			pid->private->t = va_arg(val, double);
			break;
		case 4:
			pid->private->maxi = va_arg(val, double);
			break;
		case 5:
			pid->private->mini = va_arg(val, double);
			break;
		case 6:
			pid->private->previous = 0.0;
			pid->private->outi = 0.0;
			pid->private->outp = 0.0;
			pid->private->outd = 0.0;
			pid->private->out = 0.0;
			break;
	}
	pid->private->ci = pid->private->ki * pid->private->t;
	pid->private->cd = pid->private->kd / pid->private->t;
	
}

static struct controller_block_interm_list interms[] = {
	{ "in", CONTROLLER_BLOCK_TERM_FLOAT, offsetof(struct controller_block_private, input0) },
	{ "enable", CONTROLLER_BLOCK_TERM_BOOL, offsetof(struct controller_block_private, enable) },
	{ NULL }
};

static struct controller_block_outterm_list outterms[] = {
	{ "out", CONTROLLER_BLOCK_TERM_FLOAT, offsetof(struct controller_block_private, out) },
	{ "outp", CONTROLLER_BLOCK_TERM_FLOAT, offsetof(struct controller_block_private, outp) },
	{ "outi", CONTROLLER_BLOCK_TERM_FLOAT, offsetof(struct controller_block_private, outi) },
	{ "outd", CONTROLLER_BLOCK_TERM_FLOAT, offsetof(struct controller_block_private, outd) },
	{ NULL }
};

struct controller_block * block_pid_aw_create(char *name)
{
	struct controller_block *pid;
	
	if (!(pid = controller_block_alloc("pid_aw", name, sizeof(struct controller_block_private))))
		return NULL;

	pid->private->out = 0.0;
	pid->private->outp = 0.0;
	pid->private->outi = 0.0;
	pid->private->outd = 0.0;
	pid->private->kp = 0.0;
	pid->private->ki = 0.0;
	pid->private->kd = 0.0;
	pid->private->t = 1.0;
	pid->private->ci = 0.0;
	pid->private->cd = 0.0;
	pid->private->maxi = 0.0;
	pid->private->mini = 0.0;

	pid->private->previous = 0.0;

	if (controller_block_interm_list_init(pid, interms))
		goto err_block;

	if (controller_block_outterm_list_init(pid, outterms))
		goto err_block;

	pid->calculate = pid_aw_calculate;

	if (controller_block_param_list_init(pid, params))
		goto err_block;

	pid->param_get = param_get;
	pid->param_set = param_set;

	controller_block_add(pid);
	return pid;

err_block:
	controller_block_free(pid);
	return NULL;
}
