/*
 * Copyright (c) 2012,
 *      Daan Vreeken <Daan @ Vitsch . nl> - Vitsch Electronics
 *      All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS `AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <controller/controller_block.h>
#include <log/log.h>

/*
                   quadrature_decoder
       input                               output
       name                                  name
     ----------------------------------------------
  ---| i    	                                  |----
  ---| q                                      pos |----
  ---| home                                       |----
     ----------------------------------------------

 Omschrijving:
 Dit blokje kan worden gebruikt om het signaal van twee 90-graden verschoven
 sensoren om te zetten in een encoder positie. De 'home' ingang kan optioneel
 gebruikt worden om de positie op 0 te zetten.
 De waarde van 'pos' wordt bij iedere opgaande flank van 'home' op 0 gezet.
 Iedere verandering van de i- of q-ingang geeft een positie-verandering. Na
 een heel rondje van 360 graden over [q, i] is 'pos' dus 4 hoger (of lager).
 Dit is dus een 'flankenteller' en geen 'rondjes teller'.
*/


struct controller_block_private {
	// ingangen
	bool			*i, *q;
	bool			*home;
	// uitgangen
	int32_t			pos;

	// interne variabelen
	int			last_iq;
	bool			last_home;
};

// een paar definities voor de leesbaarheid van de code
#define OUTPUT(name)		(block->private->name)
#define INPUT(name)		(*block->private->name)
#define VAR(name)		(block->private->name)

#define I_LOW		(0 << 0)
#define I_HIGH		(1 << 0)
#define Q_LOW		(0 << 1)
#define Q_HIGH		(1 << 1)

// truth table to translate current state + next state => increment
static int increment_lut[4][4] = {
	[I_LOW | Q_LOW] = {
		[I_LOW | Q_LOW] = 0,	// no motion
		[I_HIGH | Q_LOW] = +1,	// CW
		[I_LOW | Q_HIGH] = -1,	// CCW
		[I_HIGH | Q_HIGH] = 0,	// 180 degrees..!?
	},
	[I_HIGH | Q_LOW] = {
		[I_LOW | Q_LOW] = -1,
		[I_HIGH | Q_LOW] = 0,
		[I_LOW | Q_HIGH] = 0,
		[I_HIGH | Q_HIGH] = +1,
	},
	[I_LOW | Q_HIGH] = {
		[I_LOW | Q_LOW] = +1,
		[I_HIGH | Q_LOW] = 0,
		[I_LOW | Q_HIGH] = 0,
		[I_HIGH | Q_HIGH] = -1,
	},
	[I_HIGH | Q_HIGH] = {
		[I_LOW | Q_LOW] = 0,
		[I_HIGH | Q_LOW] = -1,
		[I_LOW | Q_HIGH] = +1,
		[I_HIGH | Q_HIGH] = 0,
	},
};


static void calculate(struct controller_block *block)
{
	int		iq;
	int		increment;
	
	
	iq = (INPUT(i) ? I_HIGH : I_LOW) |
	    (INPUT(q) ? Q_HIGH : Q_LOW);

	increment = increment_lut[VAR(last_iq)][iq];
	
	if ((INPUT(home) == true) && (VAR(last_home) == false)) {
		// rising edge on 'home' input
		if (OUTPUT(pos) != 0) {
			log_send(LOG_T_DEBUG, "homing @ %d\n", OUTPUT(pos));
		}
		OUTPUT(pos) = 0;
	} else {
		// add increment to position
		OUTPUT(pos) += increment;
	}
	
//	printf("i=%d	q=%d\n", INPUT(i), INPUT(q));
//	printf("enc: pos=%d	increment=%d	iq=%x  last=%x\n",
//	    OUTPUT(pos), increment, iq, VAR(last_iq));
	
	VAR(last_iq) = iq;
	VAR(last_home) = INPUT(home);
}

static struct controller_block_interm_list interms[] = {
	{ "i",    CONTROLLER_BLOCK_TERM_BOOL, offsetof(struct controller_block_private, i)    },
	{ "q",    CONTROLLER_BLOCK_TERM_BOOL, offsetof(struct controller_block_private, q)    },
	{ "home", CONTROLLER_BLOCK_TERM_BOOL, offsetof(struct controller_block_private, home) },
	{ NULL },
};

static struct controller_block_outterm_list outterms[] = {
	{ "pos", CONTROLLER_BLOCK_TERM_SINT32, offsetof(struct controller_block_private, pos) },
	{ NULL },
};

struct controller_block * block_quadrature_decoder_create(char *name)
{
	struct controller_block		*block;
	
	
	block = controller_block_alloc("quadrature_decoder", name,
	    sizeof(struct controller_block_private));
	if (block == NULL)
		return NULL;

	if (controller_block_interm_list_init(block, interms))
		goto err_block;

	if (controller_block_outterm_list_init(block, outterms))
		goto err_block;

	block->calculate = calculate;

	// this block doesn't have any parameters
	block->param_get = NULL;
	block->param_set = NULL;

	// initial state
	VAR(last_iq) = I_LOW | Q_LOW;
	VAR(last_home) = false;
	OUTPUT(pos) = 0;

	controller_block_add(block);
	return block;

err_block:
	controller_block_free(block);
	return NULL;
}
