/*
	Copyright Jeroen Vreeken (pe1rxq@amsat.org), 2011, 2013
	Copyright Stichting C.A. Muller Radioastronomiestation, 2011

	This program is free software: you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation, either version 3 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program.  If not, see <http://www.gnu.org/licenses/>.

 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <log/log.h>

#include <controller/controller_block.h>
#include <controller/controller_sample.h>

/*
       inputs      outputs
       nr name     nr name

     ---------------------------
     |                         |
  ---| 0  spg_x    0  out_x    |----
     |	                       |
  ---| 1  spg_v    1  out_v    |----
     |	                       |
  ---| 2  spg_a    2  out_a    |----
     |	                       |
  ---| 3  safe     3  spg_rst  |----
     |	                       |
     |             4  spg_en   |----
     |	                       |
     ---------------------------
     
 */

struct controller_block_private {
	float *spg_x;
	float *spg_v;
	float *spg_a;
	bool *safe;
	float out_x;
	float out_v;
	float out_a;
	bool reset;
	bool enable;
	bool emergency;

	float min_x;
	float max_x;
	float max_v;
	float max_a;
	enum {
		SERVO_STATE_DISABLED,
		SERVO_STATE_ENABLED,
		SERVO_STATE_DISABLING
	} state;
	bool enable_param;
	bool override;
};

static void servo_state_calculate(struct controller_block *servo_state)
{
	struct controller_block_private *priv = servo_state->private;
	bool safe = *priv->safe;
	
	if (priv->emergency) {
		priv->emergency = false;
		safe = false;
		log_send(LOG_T_ERROR, "%s: External emergency triggered",
		    servo_state->name);
	}
	if (!safe) {
		priv->enable_param = false;
	}

	if (priv->override) {
		priv->out_x = *priv->spg_x;
		priv->out_v = *priv->spg_v;
		priv->out_a = *priv->spg_a;
		priv->enable = true;
		priv->reset = false;
		return;
	}

	switch (priv->state) {
		case SERVO_STATE_ENABLED:
			priv->out_x = *priv->spg_x;
			priv->out_v = *priv->spg_v;
			priv->out_a = *priv->spg_a;
			     
			if (!safe) {
				priv->enable = false;
				priv->reset = true;
				priv->state = SERVO_STATE_DISABLED;
				log_send(LOG_T_ERROR, 
				    "%s: Unsafe, going to state DISABLED",
				    servo_state->name);
				break;
			}
			
			priv->enable = true;
			priv->reset = false;
			
			if (priv->enable_param) {
				break;
			} else {
				priv->state =
				    SERVO_STATE_DISABLING;
				log_send(LOG_T_INFO, 
				    "%s: Going to state DISABLING",
				    servo_state->name);
			}
			/* Fallthrough if disabling */
		case SERVO_STATE_DISABLING: {
			float out_x = priv->out_x;
			float out_v = priv->out_v;
			float out_a = 0.0;
			float t = controller_sample_period();

			
			if (!safe) {
				priv->enable = false;
				priv->reset = true;
				priv->state =
				    SERVO_STATE_DISABLED;
				log_send(LOG_T_ERROR, 
				    "%s: Unsafe, going to state DISABLED",
				    servo_state->name);
				break;
			}

			if (out_v > 0.0) {
				out_a = -priv->max_a;
				out_v += out_a * t;
				if (out_v < 0.0) {
					out_a += out_v;
					out_v = 0.0;
				}
			} else if (out_v < 0.0) {
				out_a = priv->max_a;
				out_v += out_a * t;
				if (out_v > 0.0) {
					out_a -= out_v;
					out_v = 0.0;
				}
			}
			
			if (out_v == 0.0) {
				priv->state =
				    SERVO_STATE_DISABLED;
				log_send(LOG_T_INFO, 
				    "%s: Going to state DISABLED",
				    servo_state->name);
			}
			
			out_x += out_v * t;
			
			if (out_x > priv->max_x)
				out_x = priv->max_x;
			if (out_x < priv->min_x)
				out_x = priv->min_x;
			
			priv->out_x = out_x;
			priv->out_v = out_v;
			priv->out_a = out_a;
			
			priv->enable = true;
			priv->reset = true;
			
			break;
		}
		case SERVO_STATE_DISABLED:
			priv->out_x = *priv->spg_x;
			priv->out_v = 0.0;
			priv->out_a = 0.0;
			     
			priv->enable = false;
			priv->reset = true;
			
			if (priv->enable_param) {
				priv->state = SERVO_STATE_ENABLED;
				log_send(LOG_T_INFO, 
				    "%s: Going to state ENABLED",
				    servo_state->name);
			}
			break;
	}
}

static struct controller_block_param_list params[] = {
	{ "min_x",    true },
	{ "max_x",    true },
	{ "max_v",    true },
	{ "max_a",    true },
	{ "enabled",  true },
	{ "emergency", true },
	{ "override", true },
	{ NULL },
};

static struct controller_block_interm_list interms[] = {
	{ "spg_x", CONTROLLER_BLOCK_TERM_FLOAT, offsetof(struct controller_block_private, spg_x) },
	{ "spg_v", CONTROLLER_BLOCK_TERM_FLOAT, offsetof(struct controller_block_private, spg_v) },
	{ "spg_a", CONTROLLER_BLOCK_TERM_FLOAT, offsetof(struct controller_block_private, spg_a) },
	{ "safe",  CONTROLLER_BLOCK_TERM_BOOL,  offsetof(struct controller_block_private, safe)  },
	{ NULL },
};

static struct controller_block_outterm_list outterms[] = {
	{ "out_x",  CONTROLLER_BLOCK_TERM_FLOAT, offsetof(struct controller_block_private, out_x)  },
	{ "out_v",  CONTROLLER_BLOCK_TERM_FLOAT, offsetof(struct controller_block_private, out_v)  },
	{ "out_a",  CONTROLLER_BLOCK_TERM_FLOAT, offsetof(struct controller_block_private, out_a)  },
	{ "reset",  CONTROLLER_BLOCK_TERM_BOOL,  offsetof(struct controller_block_private, reset)  },
	{ "enable", CONTROLLER_BLOCK_TERM_BOOL,  offsetof(struct controller_block_private, enable) },
	{ NULL },
};

static void param_get(struct controller_block *servo_state, int param, void *val)
{
	switch (param) {
		case 0:
			*(float*)val = servo_state->private->min_x;
			break;
		case 1:
			*(float*)val = servo_state->private->max_x;

			break;
		case 2:
			*(float*)val = servo_state->private->max_v;
			break;
		case 3:
			*(float*)val = servo_state->private->max_a;
			break;
		case 4:
			*(bool*)val = servo_state->private->enable_param;
			break;
		case 5:
			*(bool*)val = servo_state->private->emergency;
			break;
		case 6:
			*(bool*)val = servo_state->private->override;
			break;
	}
}

static void param_set(struct controller_block *servo_state, int param, va_list val)
{
	switch (param) {
		case 0:
			servo_state->private->min_x = va_arg(val, double);
			break;
		case 1:
			servo_state->private->max_x = va_arg(val, double);
			break;
		case 2:
			servo_state->private->max_v = va_arg(val, double);
			break;
		case 3:
			servo_state->private->max_a = va_arg(val, double);
			break;
		case 4:
			servo_state->private->enable_param = va_arg(val, int);
			break;
		case 5:
			servo_state->private->emergency = va_arg(val, int);
			break;
		case 6: {
			bool override = va_arg(val, int);
			if (override && !servo_state->private->override) {
				log_send(LOG_T_WARNING,
				    "%s: Switching to override mode, beware!",
				    servo_state->name);
			}
			if (!override && servo_state->private->override) {
				log_send(LOG_T_INFO,
				    "%s: Switching back to normal mode",
				    servo_state->name);
			}
			servo_state->private->override = override;
		}
	}
}

struct controller_block * block_servo_state_create(char *name)
{
	struct controller_block *servo_state;
	
	if (!(servo_state = controller_block_alloc("servo_state", name, sizeof(struct controller_block_private))))
		return NULL;

	servo_state->private->out_x = 0.0;
	servo_state->private->out_v = 0.0;
	servo_state->private->out_a = 0.0;
	servo_state->private->reset = true;
	servo_state->private->enable = false;
	servo_state->private->enable_param = false;
	servo_state->private->emergency = false;
	servo_state->private->state = SERVO_STATE_DISABLED;
	servo_state->private->override = false;

	if (controller_block_interm_list_init(servo_state, interms))
		goto err_block;

	if (controller_block_outterm_list_init(servo_state, outterms))
		goto err_block;

	servo_state->calculate = servo_state_calculate;

	if (controller_block_param_list_init(servo_state, params))
		goto err_block;

	servo_state->param_get = param_get;
	servo_state->param_set = param_set;

	controller_block_add(servo_state);
	return servo_state;

err_block:
	controller_block_free(servo_state);
	return NULL;
}
