/*
	Copyright Jeroen Vreeken (pe1rxq@amsat.org), 2011
	Copyright Stichting C.A. Muller Radioastronomiestation, 2011

	This program is free software: you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation, either version 3 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program.  If not, see <http://www.gnu.org/licenses/>.

 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <controller/controller_block.h>
#include <controller/controller_sample.h>
#include <block/block_trajectplayer.h>
#include <log/log.h>

/*
       inputs               outputs
       nr name   type       nr name     type

     ------------------------------------------
     |                                        |
  ---| 0  start  bool       0  position float |----
     |	                                      |
  ---| 1  stop   bool       1  playing  bool  |----
     |                                        |
  ---| 2  select int        2  done     bool  |----
     |	                                      |
     ------------------------------------------
     
 */

struct controller_block_private {
	bool *start;
	bool *stop;
	uint32_t *select;
	
	float position;
	bool playing;
	bool done;
	
	bool reset_on_stop;
	
	int cur;
	float playtime;
	
	int nr;
	struct block_trajectplayer_param **trajectories;
};

static int findtraject(struct controller_block_private *priv, int select)
{
	int i;
	
	for (i = 0; i < priv->nr; i++) {
		if (priv->trajectories[i]->nr == select)
			break;
	}
	return i;
}

static void calculate(struct controller_block *player)
{     
	struct controller_block_private *priv = player->private;
	double playtime = priv->playtime;
	struct block_trajectplayer_param *traject;


	if (*priv->stop) {
		priv->playing = false;
		playtime = 0.0;
		priv->done = false;
		if (priv->reset_on_stop) {
			priv->position = 0.0;
		}
	} else if (priv->playing) {
		traject = priv->trajectories[priv->cur];
		playtime += controller_sample_period();
		if (priv->playtime >= traject->time[traject->len - 1]) {
			priv->playing = false;
			priv->done = true;
			priv->position = traject->position[traject->len - 1];
			playtime = 0.0;
		} else {
			int t1, t2;
			float posdiff;
			float tdiff;
			float tinter;
			
			for (t1 = 0; t1 < traject->len; t1++) {
				if (traject->time[t1] >= priv->playtime) {
					t1--;
					break;
				}
			}
			if (t1 < 0)
				t1 = 0;
			t2 = t1 + 1;
			if (t2 >= traject->len)
				t2 = traject->len;
			tdiff = traject->time[t2] - traject->time[t1];
			tinter = playtime - traject->time[t1];
			posdiff = traject->position[t2] - traject->position[t1];
			priv->position =
			    traject->position[t1] + posdiff * tinter / tdiff;
			priv->playing = true;
			priv->done = false;
		}
	} else {
		if (*priv->start) {
			int cur = findtraject(priv, *priv->select);
			if (cur >= priv->nr) {
				log_send(LOG_T_ERROR,
				     "Illegal select, trajectory not available!");
			} else {
				priv->cur = cur;
				traject = priv->trajectories[cur];
				priv->playing = true;
				priv->done = false;
				playtime = 0.0;
				priv->position = traject->position[0];
			}
		}
	}
	
	priv->playtime = playtime;
}

static struct controller_block_param_list params[] = {
	{ "traject",       true },
	{ "reset_on_stop", true },
	{ NULL },
};

static void param_get(struct controller_block *player, int param, void *val)
{
	//TODO
}

static void param_set(struct controller_block *player, int param, va_list ap)
{
	struct controller_block_private *priv = player->private;
	struct block_trajectplayer_param traject;
	int traject_nr;
	
	switch (param) {
		case 0: {
			traject.nr = va_arg(ap, int);
			traject.len = va_arg(ap, int);
			traject.position = va_arg(ap, float *);
			traject.time = va_arg(ap, float *);
			
			traject_nr = findtraject(priv, traject.nr);
			
			if (traject_nr == priv->nr) {
				priv->trajectories = realloc(priv->trajectories,
				    sizeof(struct block_trajectplayer_param *) *
				    priv->nr);
				priv->trajectories[priv->nr] = calloc(
				    sizeof(struct block_trajectplayer_param), 1);
				priv->nr++;
				priv->trajectories[traject_nr]->nr = traject.nr;
			}
			priv->trajectories[traject_nr]->len = traject.len;
			priv->trajectories[traject_nr]->position = realloc(
			    priv->trajectories[traject_nr]->position,
			    sizeof(float) * traject.len);
			priv->trajectories[traject_nr]->time = realloc(
			    priv->trajectories[traject_nr]->time,
			    sizeof(float) * traject.len);
			memcpy(priv->trajectories[traject_nr]->position,
			    traject.position, sizeof(float) * traject.len);
			memcpy(priv->trajectories[traject_nr]->time,
			    traject.time, sizeof(float) * traject.len);
			
			
			break;
		}
		case 1: {
			priv->reset_on_stop = va_arg(ap, int);
			break;
		}
		default:
			break;
	}
}

struct controller_block * block_trajectplayer_create(char *name)
{
	struct controller_block *player;
	
	player = malloc(sizeof(struct controller_block));
	if (!player)
		return NULL;

	player->type = "trajectplayer";
	player->name = malloc(strlen(name)+1);
	if (!player->name)
		goto err_player;
	strcpy(player->name, name);

	player->private = malloc(sizeof(struct controller_block_private));
	if (!player->private)
		goto err_name;
	player->private->playing = false;
	player->private->done = false;
	player->private->position = 0.0;

	player->private->reset_on_stop = false;
	player->private->nr = 0;
	player->private->cur = 0;
	player->private->trajectories = NULL;

	player->inputs = 3;
	player->input = malloc(sizeof(struct controller_block_interm) * 3);
	if (!player->input)
		goto err_private;
	player->input[0].name = "start";
	player->input[0].type = CONTROLLER_BLOCK_TERM_BOOL;
	player->input[0].value.b = &player->private->start;
	player->input[0].ghostof = NULL;
	player->input[1].name = "stop";
	player->input[1].type = CONTROLLER_BLOCK_TERM_BOOL;
	player->input[1].value.b = &player->private->stop;
	player->input[1].ghostof = NULL;
	player->input[2].name = "select";
	player->input[2].type = CONTROLLER_BLOCK_TERM_UINT32;
	player->input[2].value.u32 = &player->private->select;
	player->input[2].ghostof = NULL;

	player->outputs = 3;
	player->output = malloc(sizeof(struct controller_block_outterm) * 3);
	if (!player->output)
		goto err_input;
	player->output[0].name = "position";
	player->output[0].type = CONTROLLER_BLOCK_TERM_FLOAT;
	player->output[0].value.f = &player->private->position;
	player->output[0].source = player;
	player->output[1].name = "playing";
	player->output[1].type = CONTROLLER_BLOCK_TERM_BOOL;
	player->output[1].value.b = &player->private->playing;
	player->output[1].source = player;
	player->output[2].name = "done";
	player->output[2].type = CONTROLLER_BLOCK_TERM_BOOL;
	player->output[2].value.b = &player->private->done;
	player->output[2].source = player;

	player->calculate = calculate;

	if (controller_block_param_list_init(player, params))
		goto err_output;

	player->param_get = param_get;
	player->param_set = param_set;

	controller_block_add(player);
	return player;

err_output:
	free(player->output);
err_input:
	free(player->input);
err_private:
	free(player->private);
err_name:
	free(player->name);
err_player:
	free(player);
	return NULL;
}
