/*
	Copyright Jeroen Vreeken (pe1rxq@amsat.org), 2009
	Copyright Stichting C.A. Muller Radioastronomiestation, 2009

	This program is free software: you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation, either version 3 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program.  If not, see <http://www.gnu.org/licenses/>.

 */

%option noinput

%{
#include <stdio.h>
#include <inttypes.h>
#include <ctype.h>
#include "il.tab.h"

static void __attribute__ ((unused)) yyunput (int c,char *buf_ptr  );

%}

digit	[0-9]
letter	[a-zA-Z]
id	[a-zA-Z0-9_]+
number	[0-9]+

intnumber	[+-]*{number}

realsimple	[+-]?{number}"."{number} 
realexp		[+-]?{number}"."?{number}?"e"[+-]*{number} 
real		{realsimple}|{realexp}

binarynumber	"2#"[01_]+
octalnumber	"8#"[0-7]+
hexnumber	"16#"[0-9a-fA-F]+

literal		{intnumber}|{real}|{binarynumber}|{octalnumber}|{hexnumber}

timeday		[0-9]+("d"|"D")
timehour	[0-9]+("h"|"H")
timeminute	[0-9]+("m"|"M")
timesecond	[0-9]+("s"|"S")
timemsecond	[0-9]+("ms")
time		[tT]"#"{timeday}?{timehour}?{timeminute}?{timesecond}?{timemsecond}?

vartype		"BOOL"|"BYTE"|"WORD"|"DWORD"|"SINT"|"USINT"|"INT"|"UINT"|"DINT"|"UDINT"|"LINT"|"ULINT"|"REAL"|"LREAL"|"TIME"
varpref		("_TO_"{vartype})*|("#"{literal})*

%%

"PROGRAM"	{ return PROGRAMSYM; }

"SYSTEM_TIME"	{ return SYSTEM_TIMESYM; }


"VAR"		{ return VARSYM; }
"VAR_INPUT"	{ return VAR_INPUTSYM; }
"VAR_OUTPUT"	{ return VAR_OUTPUTSYM; }
"VAR_CONFIG"	{ return VAR_CONFIGSYM; }
"END_VAR"	{ return END_VARSYM; }
"BOOL"/{varpref}	{ return BOOLSYM; }
"BYTE"/{varpref}	{ return UINT8SYM; }
"WORD"/{varpref}	{ return UINT16SYM; }
"DWORD"/{varpref}	{ return UINT32SYM; }
"SINT"/{varpref}	{ return SINT8SYM; }
"USINT"/{varpref}	{ return UINT8SYM; }
"INT"/{varpref}		{ return SINT16SYM; }
"UINT"/{varpref}	{ return UINT16SYM; }
"DINT"/{varpref}	{ return SINT32SYM; }
"UDINT"/{varpref}	{ return UINT32SYM; }
"LINT"/{varpref}	{ return SINT64SYM; }
"ULINT"/{varpref}	{ return UINT64SYM; }
"REAL"/{varpref}	{ return REAL32SYM; }
"LREAL"/{varpref}	{ return REAL64SYM; }
"TIME"/{varpref}	{ return UINT32SYM; }

"LD"		{ return LDSYM; }
"ST"		{ return STSYM; }
"LDN"		{ return LDNSYM; }
"STN"		{ return STNSYM; }
"AND"		{ return ANDSYM; }
"OR"		{ return ORSYM; }
"ANDN"		{ return ANDNSYM; }
"ORN"		{ return ORNSYM; }
"XOR"		{ return XORSYM; }
"JMP"		{ return JMPSYM; }
"JMPC"		{ return JMPCSYM; }
"JMPCN"		{ return JMPCNSYM; }
"ADD"		{ return ADDSYM; }
"SUB"		{ return SUBSYM; }
"MUL"		{ return MULSYM; }
"DIV"		{ return DIVSYM; }
"MOD"		{ return MODSYM; }
"GT"		{ return GTSYM; }
"EQ"		{ return EQSYM; }
"NE"		{ return NESYM; }
"LE"		{ return LESYM; }
"LT"		{ return LTSYM; }
"S"		{ return SSYM; }
"R"		{ return RSYM; }
"N"		{ return NEGATESYM; }
"("		{ return BEGINPARSYM; }
")"		{ return ENDPARSYM; }

"ABS"		{ return ABSSYM; }
"SQRT"		{ return SQRTSYM; }
"LN"		{ return LNSYM; }
"LOG"		{ return LOGSYM; }
"EXP"		{ return EXPSYM; }
"SIN"		{ return SINSYM; }
"COS"		{ return COSSYM; }
"TAN"		{ return TANSYM; }
"ASIN"		{ return ASINSYM; }
"ACOS"		{ return ACOSSYM; }
"ATAN"		{ return ATANSYM; }
"EXPT"		{ return EXPTSYM; }

"SHL"		{ return SHLSYM; }
"SHR"		{ return SHRSYM; }
"ROL"		{ return ROLSYM; }
"ROR"		{ return RORSYM; }
"NOT"		{ return NOTSYM; }

"MIN"		{ return MINSYM; }
"MAX"		{ return MAXSYM; }
"LIMIT"		{ return LIMITSYM; }
"SEL"		{ return SELSYM; }

"TRUNC"		{ return TRUNCSYM; }

":="		{ return LETSYM; }
","		{ return COMMASYM; }

"%"{id}		{ yylval.string = strdup(yytext+1); return VARIABLE; }

{time}		{ 
			uint32_t time = 0;
			int numval = 0;
			char *tmp;
			
			for (tmp = yytext + 2; *tmp; tmp++) {
				if (isdigit(*tmp)) {
					numval = numval * 10 + *tmp - '0';
				} else {
					switch (*tmp) {
						case 'd':
						case 'D':
							time += numval * 
							    (24*3600*1000);
							break;
						case 'h':
						case 'H':
							time += numval * 
							    (3600*1000);
							break;
						case 'm':
						case 'M':
							if (tmp[1] == 's' ||
							    tmp[1] == 'S') {
								tmp++;
								time += numval;
							} else {
								time += numval *
								    (60*1000);
							}
							break;
						case 's':
						case 'S':
							time += numval * 1000;
							break;
					}
					numval = 0;
				}
			}
			yylval.string = malloc(12);
			sprintf(yylval.string, "%d", time);
			return LITERALUINT;
		}

{binarynumber}	{
			uint32_t binnum = 0;
			
			char *tmp;
			
			for (tmp = yytext + 2; *tmp; tmp++) {
				switch (*tmp) {
					case '1':
						binnum = binnum << 1;
						binnum |= 1;
						break;
					case '0':
						binnum = binnum << 1;
						break;
					case '_':
						break;
				}
			}
			yylval.string = malloc(12);
			sprintf(yylval.string, "0x%08x", binnum);
			
			return LITERALUINT; 
		}

{real}		{ yylval.string = strdup(yytext); return LITERALREAL; }

{octalnumber}	{
			yylval.string = strdup(yytext+1);
			yylval.string[0] = '0';
			return LITERALUINT;
		}

{hexnumber}	{
			yylval.string = strdup(yytext+1);
			yylval.string[0] = '0';
			yylval.string[1] = 'x';
			return LITERALUINT;
		}

{intnumber}	{ yylval.string = strdup(yytext); return LITERALSINT; }
"TRUE"		{ yylval.string = strdup("true"); return LITERALUINT; }
"FALSE"		{ yylval.string = strdup("false"); return LITERALUINT; }

{id}[ \t]*":"/[^=]	{ 
			yylval.string = strdup(yytext);
			/* cut off trailing semi-colon and whitespace */
			while (yylval.string[strlen(yylval.string)-1] == ' ' ||
			       yylval.string[strlen(yylval.string)-1] == '\t'||
			       yylval.string[strlen(yylval.string)-1] == ':')
				yylval.string[strlen(yylval.string)-1] = 0;
			return LABEL;
		}

"_TO_"/{vartype}	{
	 		  return TYPECONVERSIONSYM; 
		}

"#"/{literal}	{ return TYPEDLITERALSYM; }

{id}		{ yylval.string = strdup(yytext); return VARIABLE; }

";"		/* skip */

"//".+[\n\r]	/* skip comments */
"/*".+"*/"	/* skip comments */
"(*".+"*)"	/* skip comments */
[ \t\r]		/* skip whitespace */
"\n"		{ yyset_lineno(yyget_lineno()+1);}

%%

int yywrap(void){return 1;}
