/*
	Copyright Jeroen Vreeken (pe1rxq@amsat.org), 2009
	Copyright Stichting C.A. Muller Radioastronomiestation, 2009

	This program is free software: you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation, either version 3 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program.  If not, see <http://www.gnu.org/licenses/>.

 */

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <stdbool.h>
#include "il.tab.h"
#include "il2c.h"

#define IL2C_ACCU_INT		"_il2c_acc_ui32"
#define IL2C_ACCU_INT_TYPE 	UINT32SYM
#define IL2C_ACCU_REAL		"_il2c_acc_double"
#define IL2C_ACCU_REAL_TYPE	REAL64SYM

#undef IL2C_GEN_HEADER

char *il2c_blockname = "il2c";
char *il2c_logname;

int yyparse(void);
extern FILE *yyin;

int yyget_lineno(void);

FILE *il2c_cout;
#ifdef IL2C_GEN_HEADER
FILE *il2c_hout;
#endif


void yyerror(char const *s)
{
	fprintf(stderr, "%s:%d: %s\n", il2c_logname, yyget_lineno(), s);
}

void il2c_program(char *program)
{
	il2c_blockname = strdup(program);
	
	fprintf(il2c_cout, "/* Automaticly generated block '%s' */\n", il2c_blockname);
	fprintf(il2c_cout, "\n");
	fprintf(il2c_cout, "#include <stdio.h>\n");
	fprintf(il2c_cout, "#include <stdlib.h>\n");
	fprintf(il2c_cout, "#include <string.h>\n");
	fprintf(il2c_cout, "#include <math.h>\n");
	fprintf(il2c_cout, "#include <controller/controller_block.h>\n");
	fprintf(il2c_cout, "\n");
}

struct il2c_var {
	struct il2c_var *next;
	int category;
	int type;
	char *name;
	int initsym;
	char *initval;
};

static int il2c_var_cat = IL2C_VAR_NONE;

static struct il2c_var *il2c_vars = NULL;

static int il2c_literal_type;

void il2c_literal_type_set(int type)
{
	il2c_literal_type = type;
}

int il2c_literal_type_get(void)
{
	return il2c_literal_type;
}

void il2c_var_category(int cat)
{
	il2c_var_cat = cat;
}

void il2c_var_add(int type, char *name, int initsym, char *initval)
{
	struct il2c_var *var;
	struct il2c_var **listp;
	
	var = malloc(sizeof(struct il2c_var));
	var->category = il2c_var_cat;
	var->type = type;
	var->name = strdup(name);
	var->initsym = initsym;
	if (initsym) {
		var->initval = strdup(initval);
	}

	for (listp = &il2c_vars; *listp; listp = &(*listp)->next);
	*listp = var;
}

static char *il2c_var_type2c(int type)
{
	switch (type) {
		case SINT8SYM:
			return "int8_t";
		case SINT16SYM:
			return "int16_t";
		case SINT32SYM:
			return "int32_t";
		case SINT64SYM:
			return "int64_t";
		case UINT8SYM:
			return "uint8_t";
		case UINT16SYM:
			return "uint16_t";
		case UINT32SYM:
			return "uint32_t";
		case UINT64SYM:
			return "uint64_t";
		case BOOLSYM:
			return "bool";
		case REAL32SYM:
			return "float";
		case REAL64SYM:
			return "double";
	}
	return NULL;
}

static char *il2c_var_type2cmask(int type)
{
	switch (type) {
		case SINT8SYM:
			return "& 0xff";
		case SINT16SYM:
			return "& 0xffff";
		case SINT32SYM:
			return "";
		case SINT64SYM:
			return "";
		case UINT8SYM:
			return "& 0xff";
		case UINT16SYM:
			return "& 0xffff";
		case UINT32SYM:
			return "";
		case UINT64SYM:
			return "";
		case BOOLSYM:
			return "& 0x1";
		case REAL32SYM:
			return "";
		case REAL64SYM:
			return "";
	}
	return NULL;
}

/* c promoted types */
static char *il2c_var_type2c_promoted(int type)
{
	switch (type) {
		case BOOLSYM:
		case SINT8SYM:
		case SINT16SYM:
			return "int";
		case SINT32SYM:
			return "long";
		case SINT64SYM:
			return "long long";
		case UINT8SYM:
		case UINT16SYM:
			return "unsigned int";
		case UINT32SYM:
			return "unsigned long";
		case UINT64SYM:
			return "unsigned long long";
		case REAL32SYM:
		case REAL64SYM:
			return "double";
	}
	return NULL;
}

static char *il2c_var_type2term(int type)
{
	switch (type) {
		case SINT8SYM:
			return "CONTROLLER_BLOCK_TERM_SINT8";
		case SINT16SYM:
			return "CONTROLLER_BLOCK_TERM_SINT16";
		case SINT32SYM:
			return "CONTROLLER_BLOCK_TERM_SINT32";
		case SINT64SYM:
			return "CONTROLLER_BLOCK_TERM_SINT64";
		case UINT8SYM:
			return "CONTROLLER_BLOCK_TERM_UINT8";
		case UINT16SYM:
			return "CONTROLLER_BLOCK_TERM_UINT16";
		case UINT32SYM:
			return "CONTROLLER_BLOCK_TERM_UINT32";
		case UINT64SYM:
			return "CONTROLLER_BLOCK_TERM_UINT64";
		case BOOLSYM:
			return "CONTROLLER_BLOCK_TERM_BOOL";
		case REAL32SYM:
			return "CONTROLLER_BLOCK_TERM_FLOAT";
		case REAL64SYM:
			return "CONTROLLER_BLOCK_TERM_DOUBLE";
	}
	return NULL;
}

static char *il2c_var_type2accu(int type)
{
	switch (type) {
		case SINT8SYM:
		case SINT16SYM:
		case SINT32SYM:
		case UINT8SYM:
		case UINT16SYM:
		case UINT32SYM:
		case BOOLSYM:
			return IL2C_ACCU_INT;
			break;
		case REAL32SYM:
		case REAL64SYM:
			return IL2C_ACCU_REAL;
	}
	return NULL;
}

static bool il2c_var_declared = false;

void il2c_var_declare(void)
{
	struct il2c_var *var;
	
	if (il2c_var_declared)
		return;
	il2c_var_declared = true;

	fprintf(il2c_cout, "\nstruct controller_block_private {\n");
	
	fprintf(il2c_cout, "\t/* Input variables: */\n");
	for (var = il2c_vars; var; var = var->next)
		if (var->category == IL2C_VAR_INPUT) {
			fprintf(il2c_cout, "\t%s\t*%s;\n",
			    il2c_var_type2c(var->type),
			    var->name);
		}
	fprintf(il2c_cout, "\n\t/* Output variables: */\n");
	for (var = il2c_vars; var; var = var->next)
		if (var->category == IL2C_VAR_OUTPUT)
			fprintf(il2c_cout, "\t%s\t%s;\n",
			    il2c_var_type2c(var->type),
			    var->name);
	fprintf(il2c_cout, "\n\t/* Local variables: */\n");
	for (var = il2c_vars; var; var = var->next)
		if (var->category == IL2C_VAR)
			fprintf(il2c_cout, "\t%s\t%s;\n",
			    il2c_var_type2c(var->type),
			    var->name);
	fprintf(il2c_cout, "\n\t/* Param variables: */\n");
	for (var = il2c_vars; var; var = var->next)
		if (var->category == IL2C_VAR_CONFIG)
			fprintf(il2c_cout, "\t%s\t%s;\n",
			    il2c_var_type2c(var->type),
			    var->name);
	
	fprintf(il2c_cout, "};\n\n");
}

static bool il2c_sample_header_done = false;

void il2c_sample_header(void)
{
	struct il2c_var *var;

	if (il2c_sample_header_done)
		return;
	il2c_sample_header_done = true;

	fprintf(il2c_cout, "\nstatic void %s_calculate(struct controller_block *block)\n",
	    il2c_blockname);
	fprintf(il2c_cout, "{\n");

	for (var = il2c_vars; var; var = var->next) {
		fprintf(il2c_cout, "\t%s\t%s = %sblock->private->%s;\n",
		    il2c_var_type2c(var->type),
		    var->name,
		    var->category == IL2C_VAR_INPUT ? "*" : "",
		    var->name);
	}
	
	fprintf(il2c_cout, 
	    "\t%s\t%s __attribute__ ((__unused__));\n",
	    il2c_var_type2c(IL2C_ACCU_INT_TYPE), IL2C_ACCU_INT);
	fprintf(il2c_cout, 
	    "\t%s\t%s __attribute__ ((__unused__));\n",
	    il2c_var_type2c(IL2C_ACCU_REAL_TYPE), IL2C_ACCU_REAL);

	fprintf(il2c_cout, "\n");
}

void il2c_sample_footer(void)
{
	struct il2c_var *var;

	fprintf(il2c_cout, "\n");

	for (var = il2c_vars; var; var = var->next)
		if (var->category != IL2C_VAR_INPUT)
			fprintf(il2c_cout, "\tblock->private->%s = %s;\n",
			    var->name, var->name);

	fprintf(il2c_cout, "}\n");
}

static int il2c_op_type;
static int il2c_op_literal_type;
static int il2c_op_result_type;

static int il2c_op_stack = 0;

struct il2c_op_tok {
	struct il2c_op_tok *next;
	char *tok;
	int sym;
	int stack;
};

static struct il2c_op_tok *il2c_op_tok;

void il2c_op_start(void)
{
	il2c_op_type = 0;
	il2c_op_literal_type = 0;
	il2c_op_result_type = 0;
	il2c_op_tok = NULL;
}

void il2c_op_token(char *tok)
{
	struct il2c_op_tok *newtok, **list;
	
	newtok = malloc(sizeof(struct il2c_op_tok));
	newtok->tok = strdup(tok);
	newtok->sym = 0;
	newtok->next = NULL;
	newtok->stack = il2c_op_stack;
	
	for (list = &il2c_op_tok; *list; list = &(*list)->next);
	*list = newtok;
}

void il2c_op_token_sym(int sym)
{
	struct il2c_op_tok *newtok, **list;
	
	newtok = malloc(sizeof(struct il2c_op_tok));
	newtok->tok = NULL;
	newtok->sym = sym;
	newtok->next = NULL;
	newtok->stack = il2c_op_stack;
	
	if (!il2c_op_type)
		il2c_op_type = sym;

	for (list = &il2c_op_tok; *list; list = &(*list)->next);
	*list = newtok;
}

void il2c_op_usevar(char *name)
{
	struct il2c_var *var;
	
	for (var = il2c_vars; var; var = var->next)
		if (!strcmp(var->name, name))
			break;
	
	if (!var) {
		char *errtxt = "Variable not declared: ";
		char *errbuf = malloc(strlen(errtxt) + strlen(name)+ 1);
		
		sprintf(errbuf, "%s%s", errtxt, name);
		yyerror(errbuf);
		free(errbuf);
		return;
	}
	
	/* Todo: does this solve all conflicts? */
	if (!il2c_op_type)
		il2c_op_type = var->type;
	il2c_op_token(var->name);
}

void il2c_op_literal(char *val, int type)
{
	il2c_op_literal_type = type;
	il2c_op_token(val);
}

void il2c_op_result(int type)
{
	il2c_op_result_type = type;
}

void il2c_op_end(void)
{
	struct il2c_op_tok *entry, *next;
	char *accu = "?", *result_accu = "?", *cast = "?";
	int stack;

	/*todo: check logic in this if....*/
	if (!il2c_op_type && il2c_op_literal_type)
		il2c_op_type = il2c_op_literal_type;

	if (!il2c_op_result_type)
		il2c_op_result_type = il2c_op_type;

	result_accu = il2c_var_type2accu(il2c_op_result_type);

	accu = il2c_var_type2accu(il2c_op_type);

	cast = il2c_var_type2c(il2c_op_type);
	
	fprintf(il2c_cout, "\t%s = ((%s)%s", result_accu, cast, accu);

	stack = 0;
	for (entry = il2c_op_tok; entry; entry = next) {
		next = entry->next;
		
		if (entry->stack > stack) {
			fprintf(il2c_cout, "(");
			stack++;
		}
		if (entry->stack < stack) {
			fprintf(il2c_cout, ")");
			stack--;
		}
		
		if (entry->tok) {
			fprintf(il2c_cout, "%s", entry->tok);
		} else {
			fprintf(il2c_cout, "(%s)", il2c_var_type2c(entry->sym));
		}
		free(entry->tok);
		free(entry);
	}

	fprintf(il2c_cout, ";\n");
}

void il2c_op_stack_push(void)
{
	il2c_op_stack++;
}

void il2c_op_stack_pop(void)
{
	il2c_op_stack--;
}
void il2c_op_acc_use(void)
{
	struct il2c_op_tok *entry, *save;
	
	il2c_op_token(")");

	if (!il2c_op_stack)
		return;	
	
	for (entry = il2c_op_tok; entry; entry = entry->next)
		if (entry->next->stack == il2c_op_stack)
			break;
	
	if (!entry)
		return;
	
	save = entry->next;
	entry->next = NULL;
	il2c_op_token("(");
	entry->next->next = save;
}

void il2c_store(char *name, bool negate)
{
	struct il2c_var *var;
	char *accu;
	
	for (var = il2c_vars; var; var = var->next)
		if (!strcmp(var->name, name))
			break;

	if (!var) {
		yyerror("Variable unknown");
		return;
	}
	
	accu = il2c_var_type2accu(il2c_op_result_type);
	
	fprintf(il2c_cout, "\t%s = %s(%s)(%s %s);\n", name, negate ? "! " : "", 
	    il2c_var_type2c(var->type), accu,
	    il2c_var_type2cmask(var->type));
}

void il2c_load(char *name, bool negate)
{
	struct il2c_var *var;
	char *accu;
	
	for (var = il2c_vars; var; var = var->next)
		if (!strcmp(var->name, name))
			break;

	if (!var) {
		yyerror("Variable unknown");
		return;
	}
	
	accu = il2c_var_type2accu(var->type);
	il2c_op_result_type = var->type;
	
	fprintf(il2c_cout, "\t%s = %s%s;\n", accu, negate ? "! " : "", name);
}

void il2c_load_literal(char *name, int type, bool negate)
{
	char *accu;
	
	accu = il2c_var_type2accu(type);
	il2c_op_result_type = type;

	fprintf(il2c_cout, "\t%s = %s%s;\n", accu, negate ? "! " : "", name);
}

void il2c_set(char *name, bool setcond)
{
	char *accu = IL2C_ACCU_INT;
	
	fprintf(il2c_cout, "\tif (%s) %s = %s;\n", accu, name, setcond ? "true" : "false");
}

void il2c_label(char *name)
{
	fprintf(il2c_cout, "\n_il2c_label_%s:\n", name);
}

void il2c_jump(char *name)
{
	fprintf(il2c_cout, "\tgoto _il2c_label_%s;\n", name);
}

void il2c_jumpcond(char *name, bool negate)
{
	char *accu = IL2C_ACCU_INT;
	
	fprintf(il2c_cout, "\tif (%s%s) goto _il2c_label_%s;\n", 
	    negate ? "!" : "", accu, name);
}

void il2c_converttype(int from, int to)
{
	fprintf(il2c_cout, "\t%s = (%s)%s;\n",
	    il2c_var_type2accu(to),
	    il2c_var_type2c(to),
	    il2c_var_type2accu(from));
}

/* TODO: these functions need knowledge about the last accu type... */
void il2c_abs(void)
{
	char *iaccu = IL2C_ACCU_INT;
	char *raccu = IL2C_ACCU_REAL;
	
	fprintf(il2c_cout, "\t%s = abs((%s)%s);\n", 
	    iaccu, il2c_var_type2c(SINT32SYM), iaccu);
	fprintf(il2c_cout, "\t%s = fabs(%s);\n",
	    raccu, raccu);
}

void il2c_mathfunc(char *func)
{
	char *raccu = IL2C_ACCU_REAL;

	fprintf(il2c_cout, "\t%s = %s((%s)%s);\n",
	    raccu, func, 
	    il2c_var_type2c(il2c_op_result_type),
	    il2c_var_type2accu(il2c_op_result_type));
	il2c_op_result_type = REAL64SYM;
}

void il2c_expt(char *exp)
{
	fprintf(il2c_cout, "\t%s = pow(%s, %s);\n", 
	    IL2C_ACCU_REAL, il2c_var_type2accu(il2c_op_result_type), exp);
	il2c_op_result_type = IL2C_ACCU_REAL_TYPE;
}

void il2c_shift(char *shift, char *val)
{
	fprintf(il2c_cout, "\t%s = %s %s %s;\n", IL2C_ACCU_INT, IL2C_ACCU_INT, shift, val);
}

void il2c_rotate(char *val, bool left)
{
	int len;

	switch (il2c_op_result_type) {
		case SINT8SYM:
		case UINT8SYM:
			len = 8;
			break;
		case SINT16SYM:
		case UINT16SYM:
			len = 16;
			break;
		default:
			len = 32;
	}

	if (left) {
		fprintf(il2c_cout, "\t%s = (%s)( (%s<<%s) | ( (%s>>(%d-%s)) & ((1<<%s)-1) ) );\n",
		    IL2C_ACCU_INT, il2c_var_type2c(il2c_op_result_type),
		    IL2C_ACCU_INT, val,
		    IL2C_ACCU_INT, len, val,
		    val);
	} else {
		fprintf(il2c_cout, "\t%s = (%s)( (%s>>%s) | ( (%s<<(%d-%s)) ) );\n",
		    IL2C_ACCU_INT, il2c_var_type2c(il2c_op_result_type),
		    IL2C_ACCU_INT, val,
		    IL2C_ACCU_INT, len, val);
	}
}

void il2c_not(void)
{
	fprintf(il2c_cout, "\t%s = ~ %s;\n", IL2C_ACCU_INT, IL2C_ACCU_INT);
}

char *il2c_system_time(void)
{
	return "(uint32_t)(controller_time_nseconds/1000)";
}

void il2c_minmax(bool max, char *val)
{
	fprintf(il2c_cout, "\t%s = %s %s %s ? %s : %s;\n",
	    il2c_var_type2accu(il2c_op_result_type),
	    il2c_var_type2accu(il2c_op_result_type),
	    max ? ">" : "<",
	    val,
	    il2c_var_type2accu(il2c_op_result_type),
	    val);
}

void il2c_sel(char *varfalse, char *vartrue, int type)
{
	fprintf(il2c_cout, "\t%s = %s ? %s : %s;\n",
	    il2c_var_type2accu(type),
	    IL2C_ACCU_INT,
	    vartrue,
	    varfalse);
}

int il2c_system_time_type(void)
{
	return UINT32SYM;
}


void il2c_block_body(void)
{
	int i, inputs = 0, outputs = 0, locals = 0, params = 0;
	struct il2c_var *var;
	
	for (var = il2c_vars; var; var = var->next) {
		if (var->category == IL2C_VAR_INPUT)
			inputs++;
		if (var->category == IL2C_VAR_OUTPUT)
			outputs++;
		if (var->category == IL2C_VAR)
			locals++;
		if (var->category == IL2C_VAR_CONFIG)
			params++;
	}

	if (params)
	{
		fprintf(il2c_cout, "\n");
		fprintf(il2c_cout, "static struct controller_block_param_list params[] = {\n");
		for (var = il2c_vars; var; var = var->next) {
			if (var->category == IL2C_VAR_CONFIG)
				fprintf(il2c_cout, "\t{ \"%s\", false },\n",
				    var->name);
		}
		fprintf(il2c_cout, "\t{ NULL },\n");
		fprintf(il2c_cout, "};\n");
		fprintf(il2c_cout, "\n");
		
		fprintf(il2c_cout, "static void param_get(struct controller_block *block, int param, void *val)\n");
		fprintf(il2c_cout, "{\n");
		fprintf(il2c_cout, "\tswitch (param) {\n");
		i = 0;
		for (var = il2c_vars; var; var = var->next) {
			if (var->category == IL2C_VAR_CONFIG) {
				fprintf(il2c_cout, "\t\tcase %d:\n", i);
				fprintf(il2c_cout, "\t\t\t*(%s*)val = block->private->%s;\n",
				    il2c_var_type2c(var->type),
				    var->name);
				fprintf(il2c_cout, "\t\t\tbreak;\n");
				i++;
			}
		}
		fprintf(il2c_cout, "\t}\n");
		fprintf(il2c_cout, "}\n");
		fprintf(il2c_cout, "\n");

		fprintf(il2c_cout, "static void param_set(struct controller_block *block, int param, va_list val)\n");
		fprintf(il2c_cout, "{\n");
		fprintf(il2c_cout, "\tswitch (param) {\n");
		i = 0;
		for (var = il2c_vars; var; var = var->next) {
			if (var->category == IL2C_VAR_CONFIG) {
				fprintf(il2c_cout, "\t\tcase %d:\n", i);
				fprintf(il2c_cout, "\t\t\tblock->private->%s = va_arg(val, %s);\n",
				    var->name,
				    il2c_var_type2c_promoted(var->type));
				fprintf(il2c_cout, "\t\t\tbreak;\n");
				i++;
			}
		}
		fprintf(il2c_cout, "\t}\n");
		fprintf(il2c_cout, "}\n");
		fprintf(il2c_cout, "\n");
	}

	fprintf(il2c_cout, "\n");

	fprintf(il2c_cout, "static struct controller_block_interm_list interms[] = {\n");
	for (var = il2c_vars; var; var = var->next) {
		if (var->category == IL2C_VAR_INPUT) {
			fprintf(il2c_cout, "\t{ .name = \"%s\", .type = %s,",
			    var->name, il2c_var_type2term(var->type));
			fprintf(il2c_cout, " .priv_offset = offsetof(struct controller_block_private, %s) },\n",
			    var->name);
		}
	}
	fprintf(il2c_cout, "\t{ NULL }\n");
	fprintf(il2c_cout, "};\n");

	fprintf(il2c_cout, "static struct controller_block_outterm_list outterms[] = {\n");
	for (var = il2c_vars; var; var = var->next) {
		if (var->category == IL2C_VAR_OUTPUT) {
			fprintf(il2c_cout, "\t{ .name = \"%s\", .type = %s,",
			    var->name, il2c_var_type2term(var->type));
			fprintf(il2c_cout, " .priv_offset = offsetof(struct controller_block_private, %s) },\n",
			    var->name);
		}
	}
	fprintf(il2c_cout, "\t{ NULL }\n");
	fprintf(il2c_cout, "};\n");
	

	/* create function */

	fprintf(il2c_cout, "\n");

	fprintf(il2c_cout, "struct controller_block * block_%s_create(char *name)\n",
	    il2c_blockname);
	fprintf(il2c_cout, "{\n");
	fprintf(il2c_cout, "\tstruct controller_block *block;\n");
	fprintf(il2c_cout, "\n");
	fprintf(il2c_cout, "\tif (!(block = controller_block_alloc(\"%s\", name, sizeof(struct controller_block_private))))\n",
	    il2c_blockname);
	fprintf(il2c_cout, "\t\treturn NULL;\n");
	fprintf(il2c_cout, "\n");
	
	fprintf(il2c_cout, "\n");

	/* generate input terminals */
	fprintf(il2c_cout, "\tif (controller_block_interm_list_init(block, interms))\n");
	fprintf(il2c_cout, "\t\treturn NULL;\n");

	fprintf(il2c_cout, "\n");

	/* generate output terminals */
	fprintf(il2c_cout, "\tif (controller_block_outterm_list_init(block, outterms))\n");
	fprintf(il2c_cout, "\t\treturn NULL;\n");

	/* function pointers */
	fprintf(il2c_cout, "\n");
	fprintf(il2c_cout, "\tblock->calculate = %s_calculate;\n",
	    il2c_blockname);
	fprintf(il2c_cout, "\n");
	if (params) {
		fprintf(il2c_cout, "\tif (controller_block_param_list_init(block, params))\n");
		fprintf(il2c_cout, "\t\treturn NULL;\n");
		fprintf(il2c_cout, "\tblock->param_get = param_get;\n");
		fprintf(il2c_cout, "\tblock->param_set = param_set;\n");
	} else {
		fprintf(il2c_cout, "\tblock->params = 0;\n");
		fprintf(il2c_cout, "\tblock->param = NULL;\n");
		fprintf(il2c_cout, "\tblock->param_get = NULL;\n");
		fprintf(il2c_cout, "\tblock->param_set = NULL;\n");
	}
	fprintf(il2c_cout, "\n");
	
	for (var = il2c_vars; var; var = var->next) {
		if (var->initsym &&
		    var->category != IL2C_VAR_INPUT) {
			fprintf(il2c_cout, "\tblock->private->%s = %s;\n",
				    var->name,
				    var->initval);

		}
	}
	fprintf(il2c_cout, "\n");
	
	fprintf(il2c_cout, "\tcontroller_block_add(block);\n");
	fprintf(il2c_cout, "\treturn block;\n");
	fprintf(il2c_cout, "}\n");
}

#ifdef IL2C_GEN_HEADER
void il2c_block_header(void)
{
	fprintf(il2c_hout, "/* Automaticly generated block '%s' */\n", il2c_blockname);

	fprintf(il2c_hout, "#ifndef _INCLUDE_%s_H_\n", il2c_blockname);
	fprintf(il2c_hout, "#define _INCLUDE_%s_H_\n", il2c_blockname);

	fprintf(il2c_hout, "\n");
	fprintf(il2c_hout, "struct controller_block * block_%s_create(char *name);\n",
	    il2c_blockname);
	fprintf(il2c_hout, "\n");

	fprintf(il2c_hout, "#endif /* _INCLUDE_%s_H_ */\n", il2c_blockname);
}
#endif

int main(int argc, char **argv)
{
	int ret;
	
	if (argc == 2) {
		char *outname;
		int i;

		yyin = fopen(argv[1], "r");
		if (!yyin) {
			printf("opening %s failed\n", argv[1]);
			return -1;
		}

		outname = malloc(strlen(argv[1]) + 8);
		strcpy(outname, argv[1]);

		for (i = strlen(argv[1]); i > 0; i--) {
			if (argv[1][i] == '.')
				break;
		}
		if (!i)
			i = strlen(argv[1]);
		
		outname[i + 0] = '.';
		outname[i + 1] = 'i';
		outname[i + 2] = 'l';
		outname[i + 3] = '2';
		outname[i + 4] = 'c';
		outname[i + 5] = '.';
		outname[i + 6] = 'c';
		outname[i + 7] = 0;

		il2c_cout = fopen(outname, "w");
		
		outname[strlen(outname) - 1] = 'h';

#ifdef IL2C_GEN_HEADER
		il2c_hout = fopen(outname, "w");
#endif
		
		il2c_logname = argv[1];
	} else {
		yyin = stdin;
		il2c_cout = stdout;
#ifdef IL2C_GEN_HEADER
		il2c_hout = stdout;
#endif
		il2c_logname = "stdin";
	}
	
	ret = yyparse();
	il2c_sample_footer();

	il2c_block_body();

#ifdef IL2C_GEN_HEADER
	il2c_block_header();
#endif

	return ret;
}
