/*
	Copyright Jeroen Vreeken (pe1rxq@amsat.org), 2007, 2008, 2009
	Copyright Stichting C.A. Muller Radioastronomiestation, 2007, 2008, 2009

	This program is free software: you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation, either version 3 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program.  If not, see <http://www.gnu.org/licenses/>.

 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <pthread.h>
#include <dlfcn.h>

#include <controller/controller_block.h>
#include <controller/controller_sample.h>
#include <log/log.h>

int nr_blocks = 0;
static struct controller_block **blocks = NULL;

struct calculate_func {
	void (*func)(struct controller_block *);
	struct controller_block *block;
};

static int nr_calculates = 0;
static struct calculate_func *calculates = NULL;

static char *controller_block_context = "unknown";

#define BLOCK_PREFIX "block_"
#define BLOCK_POSTFIX "_create"

int controller_block_create(char *type, char *name, va_list ap)
{
	void *handle;
	struct controller_block *(*create_func)(char *name, va_list ap);
	char *prefix = BLOCK_PREFIX;
	char *postfix = BLOCK_POSTFIX;
	char *symbol;
	int ret = 0;
	
	handle = dlopen(NULL, RTLD_NOW);
	if (!handle) {
		log_send(LOG_T_ERROR, 
		    "Could not get handle from dlopen(): %s", dlerror());
		return -1;
	}
	
	symbol = malloc(strlen(prefix)+strlen(type)+strlen(postfix)+1);
	sprintf(symbol, "%s%s%s", prefix, type, postfix);
	
	create_func = dlsym(handle, symbol);
	
	if (create_func) {
		if(!create_func(name, ap)) {
			log_send(LOG_T_ERROR, "%s() failed", symbol);
			ret = -1;
		}
	} else {
		log_send(LOG_T_ERROR, 
		    "Could not find function %s for block %s", symbol, type);
		ret = -1;
	}
	
	free(symbol);
	
	dlclose(handle);
	
	return ret;
}

static char * controller_block_term_str(enum controller_block_term_type et)
{
	char *type;

	switch (et) {
		case CONTROLLER_BLOCK_TERM_FLOAT:
			type = "float";
			break;
		case CONTROLLER_BLOCK_TERM_BOOL:
			type = "bool";
			break;
		case CONTROLLER_BLOCK_TERM_UINT32:
			type = "uint32";
			break;
		default:
			type = "unknown";
			break;
	};
	return type;
}

void controller_block_add(struct controller_block *newblock)
{
	void *tmp;
	
	newblock->context = strdup(controller_block_context);

	log_send(LOG_T_DEBUG, "Adding a new block: %s (%s) in context %s",
	    newblock->name, newblock->type, newblock->context);
	
	tmp = realloc(blocks, sizeof(struct controller_block *) * (nr_blocks+1));
	if (tmp) {
		int i;
	
		blocks = tmp;
		blocks[nr_blocks] = newblock;
		
		log_send(LOG_T_DEBUG, "\tInputs:");
		for (i = 0; i < newblock->inputs; i++) {
			char *type;
			
			type = controller_block_term_str(newblock->input[i].type);
			log_send(LOG_T_DEBUG,
			    "\t\t%d\t%s (%s)", i, newblock->input[i].name,
			    type);
			*newblock->input[i].value.v = NULL;
			newblock->input[i].otherside = NULL;
			newblock->input[i].block = newblock;
		}
		log_send(LOG_T_DEBUG, "\tOutputs:");
		for (i = 0; i < newblock->outputs; i++) {
			char *type;
			
			type = controller_block_term_str(newblock->output[i].type);
			log_send(LOG_T_DEBUG, "\t\t%d\t%s (%s)",
			    i, newblock->output[i].name, type);
		}
		log_send(LOG_T_DEBUG, "\tParams:");
		for (i = 0; i < newblock->params; i++)
			log_send(LOG_T_DEBUG, "\t\t%d\t%s%s", i, 
			    newblock->param[i].name,
			    newblock->param[i].sample ? " (SAMPLE)" : "");
		
		nr_blocks++;
	}
}

struct controller_block *controller_block_find(char *name)
{
	int i;
	
	for (i = 0; i < nr_blocks; i++) {
		if (!strcmp(blocks[i]->name, name))
			return blocks[i];
	}
	return NULL;
}

int controller_block_nr(void)
{
	return nr_blocks;
}

struct controller_block *controller_block_get(int nr)
{
	return blocks[nr];
}

struct controller_block_outterm *controller_block_find_outterm(
    struct controller_block *block, char *output)
{
	int i;
	
	for (i = 0; i < block->outputs; i++) {
		if (!strcmp(block->output[i].name, output))
			return &block->output[i];
	}
	
	return NULL;
}


char *controller_block_context_get(void)
{
	return controller_block_context;
}

void controller_block_context_set(char *new_context)
{
	controller_block_context = new_context;
}

/*
	Sample processing
 */


int controller_block_sample_init(void)
{
	int i, j, k;
	void *tmp;
	bool check = true;

	/* Sanity check */
	for (i = 0; i < nr_blocks; i++) {
		int j;
		
		for (j = 0; j < blocks[i]->inputs; j++) {
			if (*blocks[i]->input[j].value.v == NULL) {
				log_send(LOG_T_ERROR,
				    "%s.%s is not connected!",
				    blocks[i]->name,
				    blocks[i]->input[j].name);
				check = false;
			}
		}
	}
	if (!check) {
		log_send(LOG_T_ERROR, 
		     "Errors found while checking block connections");
		return -1;
	}

	while (1) {
		k = -1;
		for (i = 0; i < nr_blocks; i++) {
			for (j = 0; j < blocks[i]->inputs; j++) {
				if (blocks[i]->input[j].otherside) {
					for (k = 0; k < nr_blocks; k++) {
						if (blocks[i]->input[j].otherside ==
						    blocks[k])
							break;
					}
					if (i < k) {
						break;
					}
				}
				if (i < k)
					break;
			}
			if (i < k)
				break;
		}
		if (i < k) {
			struct controller_block *dep = blocks[k];
			
			log_send(LOG_T_DEBUG, "Moving %s before %s in chain",
			    blocks[k]->name,
			    blocks[i]->name);
			memmove(&blocks[i + 1], &blocks[i],
			    sizeof(struct controller_block *) * (k - i));
			blocks[i] = dep;
		} else {
			if (i == nr_blocks) {
				log_send(LOG_T_DEBUG, "Sorting done");
				break;
			}
		}
	}
	

	for (i = 0; i < nr_blocks; i++) {
		int first = 1;
		for (j = 0; j < blocks[i]->inputs; j++) {
			if (blocks[i]->input[j].otherside) {
				if (first)
					log_send(LOG_T_DEBUG, "%s chained to:",
					    blocks[i]->name);
				first = 0;
				log_send(LOG_T_DEBUG, "\t%s",
				    blocks[i]->input[j].otherside->name);
			}
		}
	}

	for (i = 0; i < nr_blocks; i++) {
		if (blocks[i]->calculate) {
			log_send(LOG_T_DEBUG,
			    "Block %s has a calculate function",
			    blocks[i]->name);
			tmp = realloc(calculates,
			     sizeof(struct calculate_func) * (nr_calculates+1));
			if (tmp) {
				calculates = tmp;
				calculates[nr_calculates].func =
				    blocks[i]->calculate;
				calculates[nr_calculates].block = blocks[i];
				nr_calculates++;
			} else {
				log_send(LOG_T_ERROR,
				    "Out of memory allocating memory");
			}
		} else
			log_send(LOG_T_DEBUG, 
			    "Block %s has no calculate function",
			    blocks[i]->name);
	}
	log_send(LOG_T_DEBUG, "Found %d calculate functions in %d blocks",
	    nr_calculates, nr_blocks);

	return 0;
}

int controller_block_output_get_float(char *block, char *param, float *value)
{
	int i, j, ret = -1;
	
	for (i = 0; i < nr_blocks; i++) {
		if (strcmp(blocks[i]->name, block))
			continue;
		
		for (j = 0; j < blocks[i]->outputs; j++) {
			if (!strcmp(blocks[i]->output[j].name, param) &&
			    blocks[i]->output[j].type == CONTROLLER_BLOCK_TERM_FLOAT) {
				*value = *blocks[i]->output[j].value.f;
				ret = 0;
				break;
			}
		}
	}
	return ret;
}


static struct controller_block_link *links = NULL;
static int link_nr = 0;

struct controller_block_link *controller_block_link_get(int nr)
{
	return &links[nr];
}

int controller_block_link_nr(void)
{
	return link_nr;
}

int controller_block_connect(char *outblock, char *outterm, char *inblock, char *interm, int chain)
{
	struct controller_block_link *tmp;

	link_nr++;
	tmp = realloc(
	    links, 
	    sizeof(struct controller_block_link)*link_nr);
	if (tmp) {
		links = tmp;
		links[link_nr-1].outblock = strdup(outblock);
		links[link_nr-1].outterm = strdup(outterm);
		links[link_nr-1].inblock = strdup(inblock);
		links[link_nr-1].interm = strdup(interm);
		links[link_nr-1].chain = chain;
		links[link_nr-1].type = NULL;
		
		return 0;
	} else {
		return -1;
	}
}

static int controller_block_resolve_link(struct controller_block_link *link)
{
	struct controller_block *out = NULL;
	struct controller_block *in = NULL;
	float *outptr = NULL;
	struct controller_block *outsource = NULL;
	struct controller_block_interm *realin = NULL;
	int i;
	enum controller_block_term_type outtype = CONTROLLER_BLOCK_TERM_VOID;
	
	for (i = 0; i < nr_blocks; i++) {
		if (!strcmp(link->outblock, blocks[i]->name))
			out = blocks[i];
		if (!strcmp(link->inblock, blocks[i]->name))
			in = blocks[i];
	}
	
	if (in && out) {
		for (i = 0; i < out->outputs; i++) {
			if (!strcmp(out->output[i].name, link->outterm)) {
				outptr = out->output[i].value.v;
				outsource = out->output[i].source;
				outtype = out->output[i].type;
			}
		}
		if (outptr) {
			for (i = 0; i < in->inputs; i++) {
				if (strcmp(in->input[i].name, link->interm) ||
				    in->input[i].type != outtype)
					continue;

			    	link->type = controller_block_term_str(in->input[i].type);
				
				log_send(LOG_T_DEBUG,
				    "Linking %s.%s to %s.%s%s",
				    link->outblock, link->outterm, 
				    link->inblock, link->interm,
				    link->chain ? " (chained)" : "");

				*in->input[i].value.v = outptr;
				if (link->chain)
					in->input[i].otherside = outsource;
				else
					in->input[i].otherside = NULL;
				for (realin = &in->input[i];
				    realin->ghostof;
				    realin = realin->ghostof) {
					if (link->chain)
						realin->ghostof->otherside = outsource;
					else
						realin->ghostof->otherside = NULL;
				}
				return 0;
			}
			log_send(LOG_T_ERROR, 
			    "Input terminal %s.%s not found",
			    link->inblock, link->interm);
		} else {
			log_send(LOG_T_ERROR,
			    "Output terminal %s.%s not found",
			    link->outblock, link->outterm);
		}
	} else {
		if (!in) {
			log_send(LOG_T_DEBUG,
			    "Input block %s does not (yet) exist, will try to link later",
			    link->inblock);
		}
		if (!out) {
			log_send(LOG_T_DEBUG,
			    "Output block %s does not (yet) exist, will try to link later",
			    link->outblock);
		}
	}
	
	return -1;
}


int controller_block_link(void)
{
	int i, r;
	int ret = 0;
	
	for (i = 0; i < controller_block_link_nr(); i++) {
		struct controller_block_link *link;
		
		link = controller_block_link_get(i);
		if (!link->type) {
			r = controller_block_resolve_link(link);
			if (r) {
				ret = r;
			}
		}
	}
	
	return ret;
}



uint64_t controller_time_nseconds = 0;
uint32_t controller_time_seconds = 0;
uint32_t controller_time_samplenr = 0;
uint32_t controller_samplenr = 0;

void controller_block_calculate(void)
{
	int i;
	
	for (i = 0; i < nr_calculates; i++)
		calculates[i].func(calculates[i].block);
}




int controller_block_interm_list_init(struct controller_block *block,
    struct controller_block_interm_list *list)
{
	int inputs, i;
	
	for (inputs = 0; list[inputs].name; inputs++);
	
	block->inputs = inputs;
	block->input = calloc(sizeof(struct controller_block_interm), inputs);
	if (!block->input)
		return -1;
		
	for (i = 0; i < inputs; i++) {
		block->input[i].name = list[i].name;
		block->input[i].type = list[i].type;
		block->input[i].value.v = (void **)(
		    (char *)block->private + list[i].priv_offset);
		block->input[i].ghostof = NULL;
	}

	return 0;
}

int controller_block_outterm_list_init(struct controller_block *block,
    struct controller_block_outterm_list *list)
{
	int outputs, i;
	
	for (outputs = 0; list[outputs].name; outputs++);
	
	block->outputs = outputs;
	block->output = calloc(sizeof(struct controller_block_outterm), outputs);
	if (!block->output)
		return -1;
		
	for (i = 0; i < outputs; i++) {
		block->output[i].name = list[i].name;
		block->output[i].type = list[i].type;
		block->output[i].value.v = 
		    (char *)block->private + list[i].priv_offset;
		block->output[i].source = block;
	}

	return 0;
}


/* Helper function to create a new block */
struct controller_block *controller_block_alloc(char *type, char *name,
    size_t private_size)
{
	struct controller_block	*blk;
	
	blk = calloc(1, sizeof(struct controller_block));
	if (!blk)
		goto err_calloc;
	
	blk->type = type;
	blk->name = strdup(name);
	if (!blk->name)
		goto err_name;
	
	if (private_size > 0) {
		blk->private = calloc(1, private_size);
		if (!blk->private)
			goto err_private;
	}

	return blk;

err_private:
	free(blk->name);
err_name:
	free(blk);
err_calloc:
	return NULL;
}

void controller_block_free(struct controller_block *blk)
{
	if (blk->name)
		free(blk->name);
	if (blk->private)
		free(blk->private);
	
	if (blk->input)
		free(blk->input);
	if (blk->output)
		free(blk->output);
	
	free(blk);
}
