/*
	Copyright Jeroen Vreeken (pe1rxq@amsat.org), 2013
	Copyright Stichting C.A. Muller Radioastronomiestation, 2013

	This program is free software: you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation, either version 3 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program.  If not, see <http://www.gnu.org/licenses/>.

 */
/*
	controller_block_param.c
	
	This module contains functions for handling block parameters
	-Wrappers for safe set/get while a sample thread might be running.
	-Helper which can be used when creating a new block
 */

#include <string.h>
#include <stdlib.h>

#include <controller/controller_block.h>
#include <controller/controller_sample.h>
#include <log/log.h>

static pthread_mutex_t param_lock;
static sem_t param_sync_sem;
static void (*param_set_func)(struct controller_block *, int, va_list) = NULL;
static void (*param_get_func)(struct controller_block *, int, void *) = NULL;
static int param_nr;
static void *param_value;
static va_list param_set_value;
static struct controller_block *param_block;

void controller_block_param_init(void)
{
	pthread_mutex_init(&param_lock, NULL);
	sem_init(&param_sync_sem, 0, 0);
}

int controller_block_param_set(char *blockname, char *param, va_list value)
{
	int i, j;
	int nr_blocks = controller_block_nr();
	struct controller_block *block;
	
	for (i = 0; i < nr_blocks; i++) {
		block = controller_block_get(i);
	
		if (strcmp(block->name, blockname))
			continue;
			
		for (j = 0; j < block->params; j++) {
			if (strcmp(block->param[j].name, param))
				continue;
			
			if (!controller_sample_running() ||
			    controller_sample_context() ||
			    !block->param[j].sample) {
				block->param_set(
				    block, j, value);
				return 0;
			}
				
			pthread_mutex_lock(&param_lock);

			param_nr = j;
			va_copy(param_set_value, value);
			param_block = block;
			param_set_func = block->param_set;
				
			sem_wait(&param_sync_sem);

			pthread_mutex_unlock(&param_lock);
			return 0;
		}
	}
	return -1;
}

int controller_block_param_get(char *blockname, char *param, void *value)
{
	int i, j;
	int nr_blocks = controller_block_nr();
	struct controller_block *block;
	
	for (i = 0; i < nr_blocks; i++) {
		block = controller_block_get(i);

		if (strcmp(block->name, blockname))
			continue;

		for (j = 0; j < block->params; j++) {
			if (strcmp(block->param[j].name, param))
				continue;
					
			if (!controller_sample_running() ||
			    controller_sample_context() ||
			    !block->param[j].sample) {
				block->param_get(
				    block, j, value);
				return 0;
			}

			pthread_mutex_lock(&param_lock);
					
			param_nr = j;
			param_value = value;
			param_block = block;
			param_get_func = block->param_get;
				
			sem_wait(&param_sync_sem);
					
			pthread_mutex_unlock(&param_lock);
			return 0;
		}
	}
	return -1;
}

void controller_block_param_handle(void)
{
	if (param_set_func) {
		param_set_func(param_block, param_nr, param_set_value);
		param_set_func = NULL;
		sem_post(&param_sync_sem);
	}
	if (param_get_func) {
		param_get_func(param_block, param_nr, param_value);
		param_get_func = NULL;
		sem_post(&param_sync_sem);
	}
}    

/* helper function to initialize param lists */
int controller_block_param_list_init(struct controller_block *block,
    struct controller_block_param_list *list)
{
	int i;
	
	for (i = 0; list[i].name; i++);
	block->params = i;
	block->param = malloc(sizeof(struct controller_block_param_list) * i);
	if (!block->param)
		return 1;
	
	for (i = 0; list[i].name; i++){
		block->param[i].name = malloc(strlen(list[i].name) + 1);
		if (!block->param[i].name)
			goto err_undo;
		strcpy(block->param[i].name, list[i].name);
		block->param[i].sample = list[i].sample;
	}
	
	return 0;

err_undo:
	for (; i > 0; i--) {
		free(block->param[i - 1].name);
	}
	free(block->param);
	
	return 1;
}
