/*
	Copyright Jeroen Vreeken (jeroen@vreeken.net), 2014

	This program is free software: you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation, either version 3 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program.  If not, see <http://www.gnu.org/licenses/>.

 */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <pthread.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/ioctl.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <signal.h>
#include <errno.h>
#include <limits.h>

#include <log/log.h>
#include <dt_port_numbers.h>

#include <controller/controller_command.h>
#include <controller/controller_sample.h>

#include <command/command.h>

static struct controller_command **command_list = NULL;
static int command_list_nr = 0;

static bool sample_hook_installed = false;

static void controller_command_sample_hook(void *arg)
{
	controller_command_server_start(CTRL_COMMAND_PORT, command_list_nr * 2);
}


struct controller_command *controller_command_create(
    struct controller_block *block, char *name, char *unit)
{
	struct controller_command *command;
	
	command = calloc(1, sizeof(struct controller_command));
	if (!command)
		goto err;

	command->block = block;
	command->name = strdup(name);
	command->unit = strdup(unit);

	command_list = realloc(command_list, 
	    sizeof(struct controller_command *) * (command_list_nr + 1));
	command_list[command_list_nr] = command;
	
	command_list_nr++;

	if (!sample_hook_installed) {
		controller_sample_start_hook(
		    controller_command_sample_hook, NULL);
		sample_hook_installed = true;
	}

err:
	return command;
}

void controller_command_destroy(struct controller_command *command)
{
	int i;
	
	for (i = 0; i < command_list_nr; i++) {
		if (command_list[i] == command) {
			memmove(&command_list[i], &command_list[i+1],
			    (command_list_nr - i - 1) * 
			    sizeof(struct controller_command *));
		}
	}
	
	free(command->name);
	free(command->unit);
	free(command);
}


struct command_hdl {
	bool free;
	
	struct command command;
	
	struct controller_command *ccommand;
};

static int nr_command_hdl;
static struct command_hdl *command_hdl;

static int controller_command_server_port;

/* Thes are created when starting the server and sent to each new client */
struct command_pkt *controller_command_pkt_list;

static void handler_close(struct command *command)
{
	struct command_hdl *hdl = command->private;
	
	log_send(LOG_T_DEBUG, "Command connection closed");

	hdl->free = true;
}

static void handler_name(struct command *command, char *name)
{
	struct command_hdl *hdl = command->private;
	int i;
	bool found = false;

	for (i = 0; i < command_list_nr; i++) {
		if (!strcmp(name, command_list[i]->name)) {
			found = true;
			hdl->ccommand = command_list[i];
			command->type = command_list[i]->value_type;
			break;
		}
	}
	
	if (!found) {
		log_send(LOG_T_DEBUG, "Command server %s does not exist", name);
		command_close(command);
		hdl->free = true;
	} else {
		log_send(LOG_T_DEBUG, "Handling commands for %s", name);
	}
}

static void handler_entry(struct command *command, struct command_entry *entry)
{
	struct command_hdl *hdl = command->private;
	int qw = hdl->ccommand->queue_w;
	int qwn = (qw + 1) % CONTROLLER_COMMAND_QUEUE_DEPTH;
	int qr = hdl->ccommand->queue_r;
	
	if (qwn == qr) {
		log_send(LOG_T_WARNING, 
		    "%s: Command queue full, dropping command",
		    hdl->ccommand->name);
	} else {
		memcpy(&hdl->ccommand->queue[qw], entry, 
		    sizeof(struct command_entry));
		hdl->ccommand->queue_w = qwn;
	}
}

static void *controller_command_handle(void *arg)
{
	int i;
	fd_set fdrx;
	int high;
	struct timeval tv;

	while (1) {
		FD_ZERO(&fdrx);
		high = 0;
		
		for (i = 0; i < nr_command_hdl; i++) {
			if (command_hdl[i].free)
				continue;
			
			command_fd_set(&command_hdl[i].command, &fdrx, &high);
		}
		
		tv.tv_sec = 0;
		tv.tv_usec = 100000;
	
		select(high + 1, &fdrx, NULL, NULL, &tv);
		
		for (i = 0; i < nr_command_hdl; i++) {
			if (command_hdl[i].free)
				continue;
			
			command_handle(&command_hdl[i].command, &fdrx);
		}
	}

	return NULL;
}

static void *controller_command_server(void *arg)
{
	int listen_sock;
	struct sockaddr_in sockaddr;
	int val;
	
	signal(SIGPIPE, SIG_IGN);
	
	listen_sock = socket(AF_INET, SOCK_STREAM, 0);
	if (listen_sock < 0) {
		perror("socket() failed");
		return NULL;
	}
	
	sockaddr.sin_family = AF_INET;
	sockaddr.sin_addr.s_addr = htonl(INADDR_ANY);
	sockaddr.sin_port = htons(controller_command_server_port);
	
	val = 1;
	setsockopt(listen_sock, SOL_SOCKET, SO_REUSEADDR, &val, sizeof(int));

	if (bind(listen_sock,
	    (struct sockaddr *)&sockaddr, sizeof(sockaddr)) < 0) {
		perror("bind() failed");
		close(listen_sock);
		return NULL;
	}
	
	if (listen(listen_sock, 8) < 0) {
		perror("listen() failed");
		close(listen_sock);
		return NULL;
	}
	
	while (1) {
		int client;
		struct sockaddr_in sockaddr_client;
		socklen_t len;
		
		len = sizeof(sockaddr_client);
		client = accept(listen_sock,
		    (struct sockaddr *)&sockaddr_client, &len);
		
		if (client >= 0) {
			int i;
			
			for (i = 0; i < nr_command_hdl; i++) {
				if (!command_hdl[i].free)
					continue;
				
				command_initialize_fd(&command_hdl[i].command, client);

				/* Send info to client */
				command_packet_write(&command_hdl[i].command,
				    controller_command_pkt_list);
			
				command_hdl[i].free = false;
				log_send(LOG_T_DEBUG, 
				    "Command connection opened, using handle %d", i);
				break;
			}
			if (i == nr_command_hdl)
				close(client);
		} else {
			log_send(LOG_T_ERROR, "accept() failed: %s",
			    strerror(errno));
		}
	}

	return NULL;
}

void controller_command_server_start(int portnr, int max)
{
	pthread_t thread_id;
	pthread_attr_t attr;
	int i, j;

	log_send(LOG_T_DEBUG, "Starting command server on port %d", portnr);

	command_hdl = calloc(max, sizeof(struct command_hdl));
	if (!command_hdl)
		return;
	nr_command_hdl = max;
	for (i = 0; i < max; i++) {
		command_hdl[i].free = true;
		command_initialize(&command_hdl[i].command, 8192);

		command_hdl[i].command.private = &command_hdl[i];
		command_hdl[i].command.handler_close = handler_close;
		command_hdl[i].command.handler_name = handler_name;
		command_hdl[i].command.handler_entry = handler_entry;
	}

	controller_command_pkt_list = command_packet_new();
	for (i = 0; i < command_list_nr; i++) {
		log_send(LOG_T_DEBUG, "Adding command server '%s' for block %s",
		    command_list[i]->name,
		    command_list[i]->block->name);
		log_send(LOG_T_DEBUG, "\tValue type: %s", 
		    enum_command_value_type2str(command_list[i]->value_type));
		log_send(LOG_T_DEBUG, "\tCommand types:");
		for (j = 0; j < COMMAND_PTYPE_MAX; j++) {
			if (command_list[i]->command_types[j] != COMMAND_PTYPE_NULL) {
				log_send(LOG_T_DEBUG, "\t\t%s",
				     enum_command_ptype2str(
				     command_list[i]->command_types[j]));
			} else {
				break;
			}
		}
		command_packet_list_add(controller_command_pkt_list,
		    command_list[i]->name,
		    command_list[i]->value_type,
		    command_list[i]->unit,
		    j, command_list[i]->command_types);
	}
	
	controller_command_server_port = portnr;

	pthread_attr_init(&attr);
	pthread_attr_setstacksize(&attr, PTHREAD_STACK_MIN * 2);
	pthread_create(&thread_id, &attr,
	    controller_command_server, NULL);
	pthread_create(&thread_id, &attr,
	    controller_command_handle, NULL);	
}
