/*
	Copyright Jeroen Vreeken (pe1rxq@amsat.org), 2009, 2013
	Copyright Stichting C.A. Muller Radioastronomiestation, 2009, 2013

	This program is free software: you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation, either version 3 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program.  If not, see <http://www.gnu.org/licenses/>.

 */

%code requires {
typedef void*                 yyscan_t;
}

%{

#include <stdio.h>
#include <controller/controller_block.h>
#include <controller/controller_load_int.h>
#include <controller/controller_trace.h>
#include <controller/controller_sample.h>

#define YYERROR_VERBOSE

#include <controller/controller_load_parser.tab.h>
#include <controller/controller_load_parser.yy.h>

void yyerror(yyscan_t *scanner, char const *s);

%}

%parse-param {yyscan_t scanner}
%lex-param {yyscan_t scanner}

%pure-parser

%start ctrllist

%token BRACEOPENSYM
%token BRACECLOSESYM
%token COMMASYM
%token <string> STRINGSYM
%token <string> VARIABLENAMESYM
%token <string> DOUBLEVARIABLESYM
%token <string> STRINGVARIABLESYM
%token <integer> BOOLSYM
%token <integer> INTSYM
%token <dbl> DOUBLESYM
%token <ul> UNSIGNEDLONGSYM
%token FREQUENCYSYM
%token BLOCKSSYM
%token LINKSSYM
%token TRACESSYM
%token PARAMSSYM
%token SETSYM
%token PARENTHESESOPENSYM
%token PARENTHESESCLOSESYM
%token INCLUDESYM

%token FUNC_DEG2RAD_SYM
%token FUNC_RPM2RADS_SYM

%token <dbl> CONST_PI_SYM

%left DUMMY_DOUBLEVAR
%left DUMMY_INTVAR
%left FLOATCASTSYM DOUBLECASTSYM INTCASTSYM BOOLCASTSYM

%left MINSYM
%left PLUSSYM
%left SLASHSYM
%left STARSYM
%left UMINUS

%union
{
	char *string;
	int integer;
	double dbl;
	unsigned long ul;
}

%type <string> stringvar
%type <dbl> doublevar
%type <integer> intvar
%type <ul> unsignedlongvar

%type <dbl> deg2rad
%type <dbl> rpm2rads

%%

ctrllist: ctrl
	| ctrl ctrllist
	;

ctrl	: frequency
	| links
	| blocks
	| traces
	| params
	| include
	| set
	;

frequency: FREQUENCYSYM DOUBLESYM
		{ controller_load_frequency($2); }
	| FREQUENCYSYM INTSYM
		{ controller_load_frequency($2); }
	;

include: INCLUDESYM STRINGSYM
		{ 
			if (controller_load_include($2)) {
				yyerror(scanner, "Failed to parse include");
				YYERROR;
			}
		}
	;

blocks	: BLOCKSSYM BRACEOPENSYM blocklist BRACECLOSESYM

blocklist: block
	| block blocklist
	;

block	: BRACEOPENSYM
		STRINGSYM COMMASYM 
		STRINGSYM varlist
		BRACECLOSESYM
		{
			if(controller_load_block_create($2, $4, scanner)) {
				yyerror(scanner, "Error creating block");
				YYERROR;
			}
			free($2);
			free($4);
			controller_load_var_clear(scanner);
		}
	;

params	: PARAMSSYM BRACEOPENSYM paramlist BRACECLOSESYM

paramlist: param
	| param paramlist
	;

param	: BRACEOPENSYM
		STRINGSYM COMMASYM
		STRINGSYM
		varlist
		BRACECLOSESYM
		{
			if (controller_load_block_param_set($2, $4, scanner)) {
				yyerror(scanner, "Error setting parameter");
				YYERROR;
			}
			controller_load_var_clear(scanner);
		}
	;

varlist	:	/* varlist may be emtpy */
	| COMMASYM var varlist
	;

var	: stringvar
		{ controller_load_var_add_str($1, scanner); }
	| doublevar %prec DUMMY_DOUBLEVAR
		{ controller_load_var_add_dbl($1, scanner); }
	| intvar %prec DUMMY_INTVAR
		{ controller_load_var_add_int($1, scanner); }
	| unsignedlongvar
		{ controller_load_var_add_ul($1, scanner); }
	| FLOATCASTSYM 
		{ controller_load_var_add_float_array_start(scanner); }
		BRACEOPENSYM floatlist BRACECLOSESYM
		{ controller_load_var_add_float_array_end(scanner); }
	;

stringvar: STRINGSYM
	| STRINGVARIABLESYM
		{ controller_load_variable_string_get($1); }
	;

floatlist: floatlistvar
	| floatlistvar COMMASYM floatlist
	;

floatlistvar: doublevar
		{ controller_load_var_add_flt($1, scanner); }
	;

doublevar: DOUBLEVARIABLESYM
		{ $$ = controller_load_variable_double_get($1); }
	| DOUBLESYM
		{ $$ = $1; }
	| FLOATCASTSYM doublevar
		{ $$ = $2; }
	| DOUBLECASTSYM doublevar
		{ $$ = $2; }
	| FLOATCASTSYM intvar
		{ $$ = $2; }
	| DOUBLECASTSYM intvar
		{ $$ = $2; }
	| FLOATCASTSYM unsignedlongvar
		{ $$ = $2; }
	| DOUBLECASTSYM unsignedlongvar
		{ $$ = $2; }
	| doublevar MINSYM doublevar
		{ $$ = ($1) - ($3); }
	| doublevar PLUSSYM doublevar
		{ $$ = ($1) + ($3); }
	| doublevar SLASHSYM doublevar
		{ $$ = ($1) / ($3); }
	| doublevar STARSYM doublevar
		{ $$ = ($1) * ($3); }
	| deg2rad
		{ $$ = $1; }
	| rpm2rads
		{ $$ = $1; }
	| CONST_PI_SYM
		{ $$ = M_PI; }
	| MINSYM doublevar %prec UMINUS
		{ $$ = - $2; }
	;


deg2rad: FUNC_DEG2RAD_SYM PARENTHESESOPENSYM doublevar PARENTHESESCLOSESYM
		{ $$ = ($3) * M_PI / 180.0; }
	;

rpm2rads: FUNC_RPM2RADS_SYM PARENTHESESOPENSYM doublevar PARENTHESESCLOSESYM
		{ $$ = ($3) * 2.0 * M_PI / 60.0; }
	;

intvar: INTSYM
		{ $$ = $1; }
	| INTCASTSYM intvar
		{ $$ = $2; }
	| INTCASTSYM doublevar
		{ $$ = $2; }
	| BOOLSYM
		{ $$ = $1; }
	| BOOLCASTSYM intvar
		{ $$ = $2; }
	| MINSYM intvar %prec UMINUS
		{ $$ = - $2; }
	;

unsignedlongvar: UNSIGNEDLONGSYM
		{ $$ = $1; }
	;

set:	SETSYM VARIABLENAMESYM doublevar
		{ controller_load_variable_double_set($2, $3); }
	| SETSYM VARIABLENAMESYM stringvar
		{ controller_load_variable_string_set($2, $3); }
	;

links	: LINKSSYM BRACEOPENSYM linklist BRACECLOSESYM

linklist: link
	| link linklist
	;

link	: BRACEOPENSYM
		STRINGSYM COMMASYM 
		STRINGSYM COMMASYM 
		STRINGSYM COMMASYM 
		STRINGSYM COMMASYM
		BOOLSYM
		BRACECLOSESYM
		{
			if (controller_block_connect($2, $4, $6, $8, $10)) {
				yyerror(scanner, "Error connecting blocks");
				YYERROR;
			}
			free($2);
			free($4);
			free($6);
			free($8);
		}
	;

traces	: TRACESSYM BRACEOPENSYM tracelist BRACECLOSESYM

tracelist: trace
	| trace tracelist
	;

trace	: BRACEOPENSYM
		STRINGSYM COMMASYM
		STRINGSYM COMMASYM
		STRINGSYM COMMASYM
		STRINGSYM
		BRACECLOSESYM
		{
			struct controller_trace_name trace;
			
			trace.name = $2;
			trace.unit = $4;
			trace.blockname = $6;
			trace.termname = $8;

			controller_trace_list_add(&trace);
		}
	;
