/*
	Copyright Jeroen Vreeken (pe1rxq@amsat.org), 2007
	Copyright Stichting C.A. Muller Radioastronomiestation, 2007

	This program is free software: you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation, either version 3 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
/*
	This module starts and maintains the sample thread.
	It also provides access to sample timing information.
	
 */
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <pthread.h>
#include <limits.h>
#include <time.h>
#include <string.h>
#include <sys/mman.h>
#include <errno.h>
#include <signal.h>

#include <controller/controller_block.h>
#include <controller/controller_sample.h>
#include <log/log.h>
#include <shell/shell.h>


static int nsec_interval;
static int controller_sampleoverruns = 0;
static double controller_frequency = 1.0;

static pthread_t controller_sample_thread;
static bool controller_sample_thread_running = false;

struct sample_timing {
	double min;
	double max;
	double avg;
	long long n;
	char *name;
};

static struct sample_timing st_start, st_io, st_end, st_sample;

struct controller_sample_start_hook {
	void (*func)(void *arg);
	void *arg;
};

static struct controller_sample_start_hook *controller_sample_start_hooks = NULL;
static int controller_sample_start_hooks_nr = 0;



static inline void tsnorm(struct timespec *ts)
{
	while (ts->tv_nsec >= 1000000000) {
		ts->tv_nsec -= 1000000000;
		ts->tv_sec++;
	}
}


static void sample_timing_init(struct sample_timing *st, char *name)
{
	memset(st, 0, sizeof(struct sample_timing));
	st->name = name;
}

static void sample_timing_add(struct sample_timing *st, 
    struct timespec *t1, struct timespec *t2)
{
	double t;
	
	t = t2->tv_nsec * 1.0 - t1->tv_nsec * 1.0;
	t += (t2->tv_sec - t1->tv_sec) * 1000000000;

	if (st->n == 0) {
		st->min = t;
		st->max = t;
		st->avg = t;
		st->n = 1;
		return;
	}
	
	if (t > st->max)
		st->max = t;
	if (t < st->min)
		st->min = t;
	st->avg = st->avg * st->n + t;
	st->n++;
	st->avg /= st->n * 1.0;

	return;
}

static int sample_timing_diff_nsec(struct timespec *t1, struct timespec *t2)
{
	int diff;
	
	diff = t2->tv_nsec - t1->tv_nsec;
	diff += (t2->tv_sec - t1->tv_sec) * 1000000000;
	
	return diff;
}


static ssize_t sample_timing_snprintf(struct sample_timing *st,
    char *dest, size_t n)
{
	return snprintf(dest, n, "%s\t%.3f\t%.3f\t%.3f",
	    st->name,
	    st->min / 1000.0, st->avg / 1000.0, st->max / 1000.0);
}


static int sample_timing_sane(struct timespec *t_exp, struct timespec *t_act)
{
	long s_diff;
	long n_diff;
	
	s_diff = t_act->tv_sec - t_exp->tv_sec;
	
	if (s_diff < 0) {
		return -1;
	}
	if (s_diff > 1) {
		return 1;
	}

	n_diff = s_diff * 1000 * 1000 * 1000;
	n_diff += t_act->tv_nsec - t_exp->tv_nsec;
	if (n_diff > nsec_interval * 10) {
		return 1;
	}

	return 0;
}

static void alarm_handler(int d)
{
	log_send(LOG_T_ERROR, "Sample has been inactive to long!");
}

/*
	This thread should be realtime....
 */
static void *sample_thread(void *arg)
{
	struct timespec t;
	struct sched_param param;
	struct timespec t_start;
	struct timespec t_io;
	struct timespec t_end;
	param.sched_priority = 99;
	clockid_t clock;
#ifdef __FreeBSD__
	pthread_condattr_t attr;
	pthread_mutex_t lock;
	pthread_cond_t cv;
#endif
	sigset_t sigset;
	struct sigaction sigact;

	sigemptyset (&sigset);
	sigaddset(&sigset, SIGALRM);
	sigprocmask(SIG_UNBLOCK, &sigset, NULL);
	sigact.sa_handler = alarm_handler;
	sigemptyset (&sigact.sa_mask);
	sigact.sa_flags = 0;
	sigaction(SIGALRM, &sigact, NULL);
	
	log_send(LOG_T_DEBUG, "Starting sample thread");
	controller_sample_thread_running = true;

	if (sched_setscheduler(0, SCHED_FIFO, &param) == -1) {
		log_send(LOG_T_WARNING, "sched_setscheduler() failed: %s",
		    strerror(errno));
	}

	if (mlockall(MCL_CURRENT | MCL_FUTURE )) {
		log_send(LOG_T_WARNING, "mlockall failed: %s",
		    strerror(errno));
	}

#ifndef __FreeBSD__
	clock = CLOCK_REALTIME;
#else
	clock = CLOCK_MONOTONIC_PRECISE;
	pthread_condattr_init(&attr);
	pthread_condattr_setclock(&attr, clock);
	pthread_cond_init(&cv, &attr);
	pthread_condattr_destroy(&attr);

	pthread_mutex_init(&lock, NULL);
	pthread_mutex_lock(&lock);
#endif

	clock_gettime(clock, &t);
	
	t.tv_nsec = 0;
	t.tv_sec++;

	while (1) {
		/* Pre sample stuff: */
		controller_samplenr++;
		
		t.tv_nsec += nsec_interval;
		tsnorm(&t);
		
		controller_time_nseconds = t.tv_sec * 1000000000;
		controller_time_nseconds += t.tv_nsec;
		controller_time_seconds = t.tv_sec;
		controller_time_samplenr = t.tv_nsec / nsec_interval;

		/* Wait for the right moment... 
		   (Internal timer or external interrupt) */
#ifndef __FreeBSD__
		clock_nanosleep(clock, TIMER_ABSTIME, &t, NULL);
#else
		pthread_cond_timedwait(&cv, &lock, &t);
#endif

		clock_gettime(clock, &t_start);

		/* Do sample stuff */
		controller_block_calculate();
		
		clock_gettime(clock, &t_io);
		
		controller_block_trace();
		controller_block_param_handle();
		
		clock_gettime(clock, &t_end);
		
		if (sample_timing_diff_nsec(&t, &t_end) > nsec_interval) {
			controller_sampleoverruns++;
		}
		if (sample_timing_sane(&t, &t_start)) {
			log_send(LOG_T_ERROR, 
			    "Time sanity check triggered: %lld.%09ld %lld.%09ld",
			    (long long)t.tv_sec, t.tv_nsec, 
			    (long long)t_start.tv_sec, t_start.tv_nsec);
			t.tv_sec = t_start.tv_sec;
			t.tv_nsec = (t_start.tv_nsec / nsec_interval) * nsec_interval;
			log_send(LOG_T_WARNING,
			    "Using current time for next sampe: %lld.%09ld",
			    (long long)t.tv_sec, t.tv_nsec);
		} else {
			sample_timing_add(&st_start, &t, &t_start);
			sample_timing_add(&st_io, &t_start, &t_io);
			sample_timing_add(&st_end, &t, &t_end);
			sample_timing_add(&st_sample, &t_start, &t_end);
		}
		alarm(1);
	}
	return NULL;
}

bool controller_sample_context(void)
{
	return pthread_equal(controller_sample_thread, pthread_self());
}

bool controller_sample_running(void)
{
	return false;
}


static void *sample_monitor(void *arg)
{
	char linebuf[201];
	
	linebuf[200] = 0;
	while(1) {
		log_send(LOG_T_DEBUG, "Samplenr: %d\tOverruns: %d",
		    controller_samplenr, controller_sampleoverruns);
		sample_timing_snprintf(&st_start, linebuf, 200);
		log_send(LOG_T_DEBUG, linebuf);
		sample_timing_snprintf(&st_end, linebuf, 200);
		log_send(LOG_T_DEBUG, linebuf);
		sample_timing_snprintf(&st_io, linebuf, 200);
		log_send(LOG_T_DEBUG, linebuf);
		sample_timing_snprintf(&st_sample, linebuf, 200);
		log_send(LOG_T_DEBUG, linebuf);
		sleep(1);
	}
	return NULL;
}

/********************************************
 *	Shell interface:
 */

static void controller_sample_timing_reset(void)
{
	sample_timing_init(&st_start, "start");
	sample_timing_init(&st_end, "end");
	sample_timing_init(&st_io, "io");
	sample_timing_init(&st_sample, "sample");
	controller_sampleoverruns = 0;
}

static ssize_t controller_sample_timing_snprintf(char *dest, size_t n)
{
	ssize_t used = 0;
	
	used += sample_timing_snprintf(&st_start, dest, n);
	used += sprintf(dest + used, "\n");
	used += sample_timing_snprintf(&st_end, dest + used, n - used);
	used += sprintf(dest + used, "\n");
	used += sample_timing_snprintf(&st_io, dest + used, n - used);
	used += sprintf(dest + used, "\n");
	used += sample_timing_snprintf(&st_sample, dest + used, n  - used);
	used += sprintf(dest + used, "\n");

	return used;
}

static int controller_sample_timing(char *args, char *out, int *outlen)
{
	if (args && !strcmp(args, "print")) {
		*outlen = sprintf(out, "Sample timing:\n");
		*outlen += controller_sample_timing_snprintf(out + *outlen, 2000);
	} else if (args && !strcmp(args, "reset")) {
		controller_sample_timing_reset();
		*outlen = sprintf(out, "Sample timing has been cleared.\n");
	} else {
		*outlen = sprintf(out,
		     "Valid arguments are: 'reset' and 'print'\n");
	}

	return *outlen;
}


static struct shell_cmd sample_shell_cmd = {
	"timing", "reset/print sample timing", controller_sample_timing
};

int controller_sample_shell_add(void)
{
	return shell_cmd_add(&sample_shell_cmd);
}



double controller_sample_period(void)
{
	return 1.0 / controller_frequency;
}

void controller_sample_frequency_set(double frequency)
{
	controller_frequency = frequency;
}


void controller_sample_start_hook(void (*func)(void *arg), void *arg)
{
	controller_sample_start_hooks = realloc(
	    controller_sample_start_hooks,
	    sizeof(struct controller_sample_start_hook)*(controller_sample_start_hooks_nr+1));
	controller_sample_start_hooks[controller_sample_start_hooks_nr].func = func;
	controller_sample_start_hooks[controller_sample_start_hooks_nr].arg = arg;

	controller_sample_start_hooks_nr++;
}

int controller_sample_start(void)
{
	pthread_t thread_id;
	pthread_attr_t attr;
	int i;

	nsec_interval = 1000000000.0 / controller_frequency;

	log_send(LOG_T_DEBUG, "nsec_interval: %d %f", 
	    nsec_interval, controller_frequency);

	sample_timing_init(&st_start, "start");
	sample_timing_init(&st_end, "end");
	sample_timing_init(&st_io, "io");
	sample_timing_init(&st_sample, "sample");

	for (i = 0; i < controller_sample_start_hooks_nr; i++)
		controller_sample_start_hooks[i].func(
		    controller_sample_start_hooks[i].arg);

	pthread_attr_init(&attr);
	pthread_attr_setstacksize(&attr, PTHREAD_STACK_MIN * 2);
	pthread_create(&controller_sample_thread, &attr, sample_thread, NULL);
	pthread_create(&thread_id, &attr, sample_monitor, NULL);

	return 0;
}
