/*
	Copyright Jeroen Vreeken (pe1rxq@amsat.org), 2007 - 2013
	Copyright Stichting C.A. Muller Radioastronomiestation, 2007 - 2013

	This program is free software: you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation, either version 3 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program.  If not, see <http://www.gnu.org/licenses/>.

 */
/*
	dt_ctr.c
	
	Main loop for the DT controller.
	Here the controller network is initialized.
	(By loading a .ctrl file supplied on the command line).
	After initialization the sample thread is started and this function's
	work is done.
 */

#define _GNU_SOURCE

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <math.h>
#include <signal.h>

#include "dynarg.h"

#include <controller/controller_block.h>
#include <controller/controller_trace.h>
#include <controller/controller_sample.h>
#include <controller/controller_dumpdot.h>
#include <controller/controller_load.h>
#include <shell/shell.h>
#include <ec/ec.h>
#include <log/log.h>

#include <dt_port_numbers.h>



int main(int argc, char **argv)
{
	char *ctrl_filename;
	char *dot_filename;
	sigset_t sigset;

	sigemptyset (&sigset);
	sigaddset(&sigset, SIGALRM);
	sigprocmask(SIG_BLOCK, &sigset, NULL);

	log_server_start(CTRL_LOG_PORT, LOG_T_DEBUG, LOG_T_INFO);

	controller_block_param_init();
	
	if (argc < 2) {
		printf("Usage: %s <controllerfile>\n", argv[0]);
		printf("\n");
		printf("E.g.: %s dt_ctrl.ctrl\n", argv[0]);
		goto err_init;
	}
	
	ctrl_filename = argv[1];
	
	/* Create and link blocks */
	log_send(LOG_T_DEBUG, "Going to load controller file '%s'", 
	    ctrl_filename);
	if (controller_load(ctrl_filename)) {
		log_send(LOG_T_ERROR, "Could not load controller file");
		goto err_init;
	}

	controller_trace_server_start(CTRL_TRACE_PORT, 100);

	/* Start command shell */
	
	controller_sample_shell_add();
	controller_load_shell_add();

	shell_intro = "Dwingeloo Telescope controller\n";
	shell_prompt = "DT>";
	
	shell_server_start(CTRL_SHELL_PORT, 10);

	asprintf(&dot_filename, "%s.dot", ctrl_filename);
	controller_dumpdot(dot_filename);
	free(dot_filename);

	/* Start 'sample' */

	if (controller_block_sample_init()) {
		goto err_init;
	}

	controller_sample_start();

	while (1) {
		float az_pos;
		float el_pos;
		float el_tor;

		if (controller_block_output_get_float("elevation_input_matrix", "out0", &el_pos))
			log_send(LOG_T_ERROR, "Getting elevation_input_matrix.out0 failed\n");
		if (controller_block_output_get_float("elevation_input_matrix", "out1", &el_tor))
			log_send(LOG_T_ERROR, "Getting elevation_input_matrix.out1 failed\n");
		if (controller_block_output_get_float("dt_az", "position", &az_pos))
			log_send(LOG_T_ERROR, "Getting dt_az.position failed\n");

		log_send(LOG_T_INFO, "Elevation: %e (Torsion: %e) Azimuth: %e",
		    el_pos, el_tor, az_pos);
		
		/* Log once every 15 minutes */
		sleep(15 * 60);
	}

	return 0;

err_init:
	log_server_flush();
	return 1;
}
