/*
	Copyright Jeroen Vreeken (pe1rxq@amsat.org), 2007 - 2013
	Copyright Stichting C.A. Muller Radioastronomiestation, 2007 - 2012

	This program is free software: you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation, either version 3 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program.  If not, see <http://www.gnu.org/licenses/>.

 */

/*
	ec.c
	
	Low level ethercat bus handling.
	Higher level functions like ESC or application can use these.
	- Send/Receive PDOs
	- Send/Receive datagrams
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <time.h>
#include <pthread.h>
#include <semaphore.h>

#include <controller/controller_sample.h>

#include <ec/ec.h>
#include <ec/ec_int.h>
#include <ec/eth.h>
#include <log/log.h>

static int ec_sock;

struct unaligned_uint64_t {
	uint64_t u64;
} __packed;

struct unaligned_uint16_t {
	uint16_t u16;
} __packed;

struct ec_pdo_frame {
	unsigned char tx_frame[1500];
	unsigned char rx_frame[1500];
	
	struct ec_frame_hdr *frame_hdr;
	struct unaligned_uint64_t *time;
	struct unaligned_uint16_t *time_wkc;
	size_t flen;
	uint32_t nr;
	int tx_users;
	int rx_users;
	int ready;
};

static struct ec_pdo_frame tx_pdo_real = {
	frame_hdr: (struct ec_frame_hdr *)&tx_pdo_real.tx_frame,
	flen: sizeof(struct ec_frame_hdr),
	nr: -1,
	tx_users: 0,
	rx_users: 0,
};
static struct ec_pdo_frame *tx_pdo = &tx_pdo_real;

static struct ec_pdo_frame rx_pdo = {
	frame_hdr: (struct ec_frame_hdr *)&rx_pdo.tx_frame,
	flen: sizeof(struct ec_frame_hdr),
	nr: -1,
	tx_users: 0,
	rx_users: 0,
};

static int rx_pdo_enabled = 1;
static int tx_pdo_enabled = 1;
static int single_pdo = 0;

static int ec_rx_pdo_add_time(void)
{
	struct ec_dgram_hdr *dgram;
	struct ec_dgram_hdr *prev;
	struct unaligned_uint16_t *wkc;
	
	rx_pdo.rx_users = ec_slave_dc_count();

	dgram = (struct ec_dgram_hdr *)(rx_pdo.tx_frame + rx_pdo.flen);
	rx_pdo.time = (struct unaligned_uint64_t *)
	    (rx_pdo.tx_frame + rx_pdo.flen + sizeof(struct ec_dgram_hdr));
	
	rx_pdo.flen += sizeof(struct ec_dgram_hdr);
	rx_pdo.flen += sizeof(uint64_t) + 2;

	log_send(LOG_T_DEBUG,
	    "Adding %zu bytes for clock distribution to rx_pdo, total pdo is now: %zu bytes",
	    sizeof(uint64_t), rx_pdo.flen);

	wkc = (struct unaligned_uint16_t *)
	    (rx_pdo.tx_frame + rx_pdo.flen - 2);
	wkc->u16 = 0;
	rx_pdo.time_wkc = (struct unaligned_uint16_t *)
	    (rx_pdo.rx_frame + rx_pdo.flen - 2);

	EC_FRAME_HDR_SET(rx_pdo.frame_hdr,
	    rx_pdo.flen - sizeof(struct ec_frame_hdr),
	    EC_FRAME_HDR_TYPE_EC);

	dgram->idx = 0;

	dgram->cmd = EC_DGRAM_HDR_CMD_BWR;
	dgram->addr.position.pos = 0x0000;
	dgram->addr.position.off = ESC_ADDR_MAP_DC_SYSTEM_TIME;

	EC_DGRAM_HDR_SET(dgram, sizeof(uint64_t), 0, 0);

	for (prev = (struct ec_dgram_hdr *)(rx_pdo.tx_frame + sizeof(struct ec_frame_hdr));
	    prev != dgram;
	    prev = ((void*)prev) + EC_DGRAM_HDR_LEN(prev) + sizeof(struct ec_dgram_hdr) + 2) {
		EC_DGRAM_HDR_SET(prev, EC_DGRAM_HDR_LEN(prev), EC_DGRAM_HDR_C(prev), 1);
	}
	
	return 0;
}

int ec_rx_pdo_add(struct ec_dgram_addr *addr, unsigned char **data, size_t len)
{
	struct ec_dgram_hdr *dgram;
	struct ec_dgram_hdr *prev;
	struct unaligned_uint16_t *wkc;
	
	for (prev = (struct ec_dgram_hdr *)(rx_pdo.tx_frame + sizeof(struct ec_frame_hdr));
	    (size_t)prev < (size_t)rx_pdo.tx_frame + rx_pdo.flen;
	    prev = ((void*)prev) + EC_DGRAM_HDR_LEN(prev) + sizeof(struct ec_dgram_hdr) + 2) {
		if (addr->type == EC_DGRAM_ADDR_AUTO_INC &&
		    prev->cmd == EC_DGRAM_HDR_CMD_APRD &&
		    prev->addr.position.pos == addr->addr.position.pos &&
		    prev->addr.position.off == addr->addr.position.off) {
			printf("Double registration for rx PDO");
			*data = (unsigned char *)prev + sizeof(struct ec_dgram_hdr);
			return 0;
		}
	}

	dgram = (struct ec_dgram_hdr *)(rx_pdo.tx_frame + rx_pdo.flen);
	*data = rx_pdo.rx_frame + rx_pdo.flen + sizeof(struct ec_dgram_hdr);
	
	rx_pdo.flen += sizeof(struct ec_dgram_hdr);
	rx_pdo.flen += len + 2;

	log_send(LOG_T_DEBUG,
	    "Adding %zu bytes data to rx_pdo, total pdo is now: %zu bytes",
	    len, rx_pdo.flen);

	wkc = (struct unaligned_uint16_t*)(rx_pdo.tx_frame + rx_pdo.flen - 2);
	wkc->u16 = 0;

	EC_FRAME_HDR_SET(rx_pdo.frame_hdr,
	    rx_pdo.flen - sizeof(struct ec_frame_hdr),
	    EC_FRAME_HDR_TYPE_EC);

	dgram->idx = 0;

	switch (addr->type) {
		case EC_DGRAM_ADDR_AUTO_INC:
			dgram->cmd = EC_DGRAM_HDR_CMD_APRD;
			dgram->addr.position.pos = addr->addr.position.pos;
			dgram->addr.position.off = addr->addr.position.off;
			log_send(LOG_T_DEBUG, "Address: %x.%x", 
			    addr->addr.position.pos, addr->addr.position.off);
			break;
		default:
			printf("Addressing mode not implemented!\n");
			return 1;
	}

	EC_DGRAM_HDR_SET(dgram, len, 0, 0);

	for (prev = (struct ec_dgram_hdr *)(rx_pdo.tx_frame + sizeof(struct ec_frame_hdr));
	    prev != dgram;
	    prev = ((void*)prev) + EC_DGRAM_HDR_LEN(prev) + sizeof(struct ec_dgram_hdr) + 2) {
		EC_DGRAM_HDR_SET(prev, EC_DGRAM_HDR_LEN(prev), EC_DGRAM_HDR_C(prev), 1);
	}
	
	return 0;
}

int ec_tx_pdo_add(struct ec_dgram_addr *addr, unsigned char **data, size_t len)
{
	struct ec_dgram_hdr *dgram;
	struct ec_dgram_hdr *prev;
	struct unaligned_uint16_t *wkc;

	for (prev = (struct ec_dgram_hdr *)(tx_pdo->tx_frame + sizeof(struct ec_frame_hdr));
	    (size_t)prev < (size_t)tx_pdo->tx_frame + tx_pdo->flen;
	    prev = ((void*)prev) + EC_DGRAM_HDR_LEN(prev) + sizeof(struct ec_dgram_hdr) + 2) {
		if (addr->type == EC_DGRAM_ADDR_AUTO_INC &&
		    prev->cmd == EC_DGRAM_HDR_CMD_APWR &&
		    prev->addr.position.pos == addr->addr.position.pos &&
		    prev->addr.position.off == addr->addr.position.off) {
			printf("Double registration for tx PDO");
			*data = (unsigned char *)prev + sizeof(struct ec_dgram_hdr);
			return 0;
		}
	}
	
	dgram = (struct ec_dgram_hdr *)(tx_pdo->tx_frame + tx_pdo->flen);
	*data = tx_pdo->tx_frame + tx_pdo->flen + sizeof(struct ec_dgram_hdr);
	
	tx_pdo->flen += sizeof(struct ec_dgram_hdr);
	tx_pdo->flen += len + 2;
	
	tx_pdo->tx_users++;

	log_send(LOG_T_DEBUG,
	    "Adding %zu bytes data to tx_pdo, total pdo is now: %zu bytes",
	    len, tx_pdo->flen);
	log_send(LOG_T_DEBUG, "tx_pdo has %d users", tx_pdo->tx_users);

	wkc = (struct unaligned_uint16_t*)(tx_pdo->tx_frame + tx_pdo->flen - 2);
	wkc->u16 = 0;

	EC_FRAME_HDR_SET(tx_pdo->frame_hdr,
	    tx_pdo->flen - sizeof(struct ec_frame_hdr),
	    EC_FRAME_HDR_TYPE_EC);

	dgram->idx = 0;

	switch (addr->type) {
		case EC_DGRAM_ADDR_AUTO_INC:
			dgram->cmd = EC_DGRAM_HDR_CMD_APWR;
			dgram->addr.position.pos = addr->addr.position.pos;
			dgram->addr.position.off = addr->addr.position.off;
			break;
		default:
			printf("Addressing mode not implemented!\n");
			return 1;
	}

	EC_DGRAM_HDR_SET(dgram, len, 0, 0);

	for (prev = (struct ec_dgram_hdr *)(tx_pdo->tx_frame + sizeof(struct ec_frame_hdr));
	    prev != dgram;
	    prev = ((void*)prev) + EC_DGRAM_HDR_LEN(prev) + sizeof(struct ec_dgram_hdr) + 2) {
		EC_DGRAM_HDR_SET(prev, EC_DGRAM_HDR_LEN(prev), EC_DGRAM_HDR_C(prev), 1);
	}
	
	return 0;
}

static ssize_t ec_dgram(void *tx_frame, void *rx_frame, size_t len)
{
	ssize_t ret;
	struct ec_dgram_hdr *rx_dgram1, *tx_dgram1;
	static uint8_t dgram_idx = 0;
	ssize_t minlen = sizeof(struct ec_frame_hdr)+sizeof(struct ec_dgram_hdr);
	dgram_idx++;

	tx_dgram1 = (struct ec_dgram_hdr *)
	    ((char *)tx_frame +  sizeof(struct ec_frame_hdr));
	rx_dgram1 = (struct ec_dgram_hdr *)
	    ((char *)rx_frame +  sizeof(struct ec_frame_hdr));
	

	tx_dgram1->idx = dgram_idx;

	ret = eth_send(ec_sock, tx_frame, len);
	if (ret != len)
		return 1;

	do {
		ret = eth_recv(ec_sock, rx_frame, len);

		if (ret < minlen)
			return 1;
	} while (rx_dgram1->idx != dgram_idx);

	return 0;

}

int ec_tx_pdo(uint32_t nr)
{
	ssize_t ret;

	if (single_pdo)
		return 0;

	if (!tx_pdo_enabled)
		return 1;

	if (tx_pdo->nr != nr) {
		tx_pdo->ready = 1;
		tx_pdo->nr = nr;
	} else {
		tx_pdo->ready++;
	}
	
	if(tx_pdo->ready < tx_pdo->tx_users) {
			return 0;
	}
	
	ret = ec_dgram(tx_pdo->tx_frame, tx_pdo->rx_frame, tx_pdo->flen);

	return ret;
}

bool ec_rx_pdo_err_state = false;
bool ec_rx_pdo_wkc_err_state = false;

int ec_rx_pdo(uint32_t nr)
{
	ssize_t ret;
	struct timespec t;
	int r = 0;

	if (!rx_pdo_enabled)
		return 1;

	if (rx_pdo.nr == nr)
		return 0;

	clock_gettime(0, &t);
	rx_pdo.time->u64 = htole64(
	    (ESC_DC_EPOCH + (uint64_t)t.tv_sec) * 1000000000 +
	    (uint64_t)t.tv_nsec
	);

	ret = ec_dgram(rx_pdo.tx_frame, rx_pdo.rx_frame, rx_pdo.flen);

	if (ret) {
		if (!ec_rx_pdo_err_state)
			log_send(LOG_T_ERROR, 
			    "Did not send/receive ethercat frame %d!", nr);
		ec_rx_pdo_err_state = true;
		r = ret;
		goto rx_err;
	} else {
		if (ec_rx_pdo_err_state) {
			log_send(LOG_T_WARNING,
			    "Ethercat frame send/received again. (%d)", nr);
		}
		ec_rx_pdo_err_state = false;
	}

	if (le16toh(rx_pdo.time_wkc->u16) != rx_pdo.rx_users) {
		if (!ec_rx_pdo_wkc_err_state) {
			log_send(LOG_T_ERROR,
			    "Only received clock ack from %d ethercat slaves! (%d expected)",
			    le16toh(rx_pdo.time_wkc->u16), rx_pdo.rx_users);
		}
		ec_rx_pdo_wkc_err_state = true;
		r = 1;
	} else {
		if (ec_rx_pdo_wkc_err_state) {
			log_send(LOG_T_WARNING,
			    "Received clock ack from %d ethercat slaves again",
			    rx_pdo.rx_users);
			ec_rx_pdo_wkc_err_state = false;
		}
	}	
rx_err:
	rx_pdo.nr = nr;
	return r;
}

struct ec_pdo_hook {
	struct ec_pdo_hook *next;
	
	int (*hook)(void *arg);
	void *arg;
};

static struct ec_pdo_hook *ec_rx_pdo_enable_hooks = NULL;
static struct ec_pdo_hook *ec_tx_pdo_enable_hooks = NULL;
static struct ec_pdo_hook *ec_rx_pdo_disable_hooks = NULL;
static struct ec_pdo_hook *ec_tx_pdo_disable_hooks = NULL;

int ec_rx_pdo_enable(int enable)
{
	struct ec_pdo_hook *hook;
	bool oldstate = rx_pdo_enabled;

	if (enable && !oldstate) {
		log_send(LOG_T_INFO, "Enabling ethercat rx PDOs");
		for (hook = ec_rx_pdo_enable_hooks; hook; hook = hook->next)
			hook->hook(hook->arg);
	}

	rx_pdo_enabled = enable;
	
	if (!enable && oldstate) {
		log_send(LOG_T_INFO, "Disabling ethercat rx PDOs");
		for (hook = ec_rx_pdo_disable_hooks; hook; hook = hook->next)
			hook->hook(hook->arg);
	}
	
	return 0;
}

int ec_tx_pdo_enable(int enable)
{
	struct ec_pdo_hook *hook;
	bool oldstate = tx_pdo_enabled;
	
	if (enable && !oldstate) {
		log_send(LOG_T_INFO, "Enabling ethercat tx PDOs");
		for (hook = ec_tx_pdo_enable_hooks; hook; hook = hook->next)
			hook->hook(hook->arg);
	}

	tx_pdo_enabled = enable;
	
	if (!enable && oldstate) {
		log_send(LOG_T_INFO, "Disabling ethercat tx PDOs");
		for (hook = ec_tx_pdo_disable_hooks; hook; hook = hook->next)
			hook->hook(hook->arg);
	}
	
	return 0;
}

int ec_rx_pdo_enable_hook_attach(int (*hook)(void *arg), void *arg)
{
	struct ec_pdo_hook *newhook;
	
	newhook = malloc(sizeof(struct ec_pdo_hook));
	if (!newhook)
		return -1;
	
	newhook->hook = hook;
	newhook->arg = arg;
	newhook->next = ec_rx_pdo_enable_hooks;
	ec_rx_pdo_enable_hooks = newhook;

	return 0;
}

int ec_tx_pdo_enable_hook_attach(int (*hook)(void *arg), void *arg)
{
	struct ec_pdo_hook *newhook;
	
	newhook = malloc(sizeof(struct ec_pdo_hook));
	if (!newhook)
		return -1;
	
	newhook->hook = hook;
	newhook->arg = arg;
	newhook->next = ec_tx_pdo_enable_hooks;
	ec_tx_pdo_enable_hooks = newhook;

	return 0;
}

int ec_rx_pdo_disable_hook_attach(int (*hook)(void *arg), void *arg)
{
	struct ec_pdo_hook *newhook;
	
	newhook = malloc(sizeof(struct ec_pdo_hook));
	if (!newhook)
		return -1;
	
	newhook->hook = hook;
	newhook->arg = arg;
	newhook->next = ec_rx_pdo_disable_hooks;
	ec_rx_pdo_disable_hooks = newhook;

	return 0;
}

int ec_tx_pdo_disable_hook_attach(int (*hook)(void *arg), void *arg)
{
	struct ec_pdo_hook *newhook;
	
	newhook = malloc(sizeof(struct ec_pdo_hook));
	if (!newhook)
		return -1;
	
	newhook->hook = hook;
	newhook->arg = arg;
	newhook->next = ec_tx_pdo_disable_hooks;
	ec_tx_pdo_disable_hooks = newhook;

	return 0;
}



int ec_slave_count(void)
{
	char frame[1500];
	struct ec_dgram_hdr *dgram;
	struct ec_frame_hdr *frame_hdr = (struct ec_frame_hdr *)frame;
	char *payload;
	ssize_t flen;
	struct unaligned_uint16_t *wkc;
	int ret;
	
	flen = sizeof(struct ec_frame_hdr);
	dgram = (struct ec_dgram_hdr *)(frame + flen);
	payload = (frame + flen + sizeof(struct ec_dgram_hdr));	
	flen += sizeof(struct ec_dgram_hdr);
	flen += sizeof(char) + 2;

	log_send(LOG_T_DEBUG, "Reading byte 0 of all ethercat slaves");
	wkc = (struct unaligned_uint16_t *)(frame + flen - 2);
	wkc->u16 = 0;
	*payload = 0x00;

	EC_FRAME_HDR_SET(frame_hdr,
	    flen - sizeof(struct ec_frame_hdr),
	    EC_FRAME_HDR_TYPE_EC);

	dgram->idx = 0;

	dgram->cmd = EC_DGRAM_HDR_CMD_BRD;
	dgram->addr.position.pos = 0x0000;
	dgram->addr.position.off = ESC_ADDR_MAP_TYPE;

	EC_DGRAM_HDR_SET(dgram, sizeof(char), 0, 0);

	ret = ec_dgram(frame, frame, flen);

	if (ret)
	{
		log_send(LOG_T_ERROR, "Did not send/receive ethercat frame!\n");
		return -1;
	}
	
	return le16toh(wkc->u16);
}

/* Count the number of slaves which can receive a distributed clock */
int ec_slave_dc_count(void)
{
	int slaves = ec_slave_count();
	int dc_slaves = 0;
	int i;
	ssize_t r;
	
	for (i = 0; i < slaves; i++) {
		struct ec_dgram_addr devaddr;
		uint16_t val16;
		
		ec_addr_set_auto_inc_nr(&devaddr, i);

		devaddr.addr.position.off = ESC_ADDR_MAP_FEATURES;
		r = ec_datagram_read(&devaddr, &val16, 2);
	
		if (r != 2)
			continue;
		
		val16 = le16toh(val16);

		if (val16 & ESC_FEATURE_DISTRIBUTED_CLOCKS)
			dc_slaves++;
	}
	return dc_slaves;
}

static struct ec_dgram_addr *datagram_addr;
static void *datagram_buffer;
static size_t datagram_len;
static ssize_t datagram_ret;
static bool datagram_read = false;
static bool datagram_write = false;

static bool sample_running = false;
static pthread_mutex_t datagram_lock;
static sem_t datagram_sync_sem;

static void ec_datagram_sample_init(void *arg)
{
	struct timeval timeout;
	double timeout_sec;

	timeout_sec = controller_sample_period() / 2;
	
	timeout.tv_sec = (int)(timeout_sec);
	timeout.tv_usec = (int)(timeout_sec * 1000000) % 1000000;
	log_send(LOG_T_INFO,
	    "Setting ethercat rx/tx timeout to %d.%06d sec based on %f sec period", 
	    (int)timeout.tv_sec, (int)timeout.tv_usec, controller_sample_period());

	eth_timeout_set(ec_sock, &timeout, &timeout);

	log_send(LOG_T_INFO, "Switching ethercat datagrams to sample-synchronised mode");
	pthread_mutex_init(&datagram_lock, NULL);
	sem_init(&datagram_sync_sem, 0, 0);
	
	sample_running = true;
}

static ssize_t ec_datagram_read_real(struct ec_dgram_addr *addr,
    void *buffer, size_t len)
{
	unsigned char frame[1500];
	struct ec_frame_hdr *frame_hdr;
	struct ec_dgram_hdr *dgram_hdr;
	void *data;
	struct unaligned_uint16_t *wkc;
	size_t flen =
	    sizeof(struct ec_frame_hdr) +
	    sizeof(struct ec_dgram_hdr) +
	    len +
	    sizeof(uint16_t);
	ssize_t ret;

	frame_hdr = (struct ec_frame_hdr*)frame;
	dgram_hdr = (struct ec_dgram_hdr*)(frame + sizeof(struct ec_frame_hdr));
	data = (unsigned char *)dgram_hdr + sizeof(struct ec_dgram_hdr);

	memset(frame, 0, flen);
	wkc = (struct unaligned_uint16_t*)(frame + flen - sizeof(uint16_t));
	wkc->u16 = 0;
	
	EC_FRAME_HDR_SET(frame_hdr, flen - sizeof(struct ec_frame_hdr),
	    EC_FRAME_HDR_TYPE_EC);
	
	dgram_hdr->idx = 0;

	switch (addr->type) {
		case EC_DGRAM_ADDR_AUTO_INC:
			dgram_hdr->cmd = EC_DGRAM_HDR_CMD_APRD;
			dgram_hdr->addr.position.pos = addr->addr.position.pos;
			dgram_hdr->addr.position.off = addr->addr.position.off;
			break;
		default:
			goto err_add;
	}

	EC_DGRAM_HDR_SET(dgram_hdr, len, 0, 0);
	
	ret = ec_dgram(frame, frame, flen);

	if (ret)
		goto err_sendrecv;

	/* wkc should be updated by slave */
	if (le16toh(wkc->u16) == 1) {
		memcpy(buffer, data, len);
		return len;
	} else
		return 0;

err_add:
err_sendrecv:
	return -1;
}

ssize_t ec_datagram_read(struct ec_dgram_addr *addr,
    void *buffer, size_t len)
{
	ssize_t ret;
	
	if (controller_sample_context())
		return ec_datagram_read_real(addr, buffer, len);

	pthread_mutex_lock(&datagram_lock);

	if (!sample_running) {
		ret = ec_datagram_read_real(addr, buffer, len);
	} else {
		datagram_addr = addr;
		datagram_buffer = buffer;
		datagram_len = len;
		datagram_read = true;
	
		sem_wait(&datagram_sync_sem);
	
		ret = datagram_ret;
	}
	
	pthread_mutex_unlock(&datagram_lock);

	return ret;
}

static ssize_t ec_datagram_write_real(struct ec_dgram_addr *addr,
    void *buffer, size_t len)
{
	unsigned char frame[1500];
	struct ec_frame_hdr *frame_hdr;
	struct ec_dgram_hdr *dgram_hdr;
	void *data;
	struct unaligned_uint16_t *wkc;
	size_t flen =
	    sizeof(struct ec_frame_hdr) +
	    sizeof(struct ec_dgram_hdr) +
	    len +
	    sizeof(uint16_t);
	ssize_t ret;
//	struct timeval timeout;
//	fd_set fdr;

	frame_hdr = (struct ec_frame_hdr*)frame;
	dgram_hdr = (struct ec_dgram_hdr*)(frame + sizeof(struct ec_frame_hdr));
	data = (unsigned char *)dgram_hdr + sizeof(struct ec_dgram_hdr);

	memset(frame, 0, flen);
	wkc = (struct unaligned_uint16_t*)(frame + flen - sizeof(uint16_t));
	wkc->u16 = 0;
	
	EC_FRAME_HDR_SET(frame_hdr, flen - sizeof(struct ec_frame_hdr),
	    EC_FRAME_HDR_TYPE_EC);
	
	dgram_hdr->idx = 0;

	switch (addr->type) {
		case EC_DGRAM_ADDR_AUTO_INC:
			dgram_hdr->cmd = EC_DGRAM_HDR_CMD_APWR;
			dgram_hdr->addr.position.pos = addr->addr.position.pos;
			dgram_hdr->addr.position.off = addr->addr.position.off;
			break;
		default:
			goto err_add;
	}

	EC_DGRAM_HDR_SET(dgram_hdr, len, 0, 0);
	
	memcpy(data, buffer, len);
	
	ret = ec_dgram(frame, frame, flen);

	if (ret)
		goto err_sendrecv;

	/* wkc should be updated by slave */
	if (le16toh(wkc->u16) == 1) {
		return len;
	} else
		return 0;

err_add:
err_sendrecv:
	return -1;
}

ssize_t ec_datagram_write(struct ec_dgram_addr *addr,
    void *buffer, size_t len)
{
	ssize_t ret;
	
	if (controller_sample_context())
		return ec_datagram_write_real(addr, buffer, len);
	
	pthread_mutex_lock(&datagram_lock);

	if (!sample_running || controller_sample_context()) {
		ret = ec_datagram_write_real(addr, buffer, len);
	} else {
		datagram_addr = addr;
		datagram_buffer = buffer;
		datagram_len = len;
		datagram_write = true;
	
		sem_wait(&datagram_sync_sem);
	
		ret = datagram_ret;
	}
	
	pthread_mutex_unlock(&datagram_lock);

	return ret;
}

void ec_datagram_sample(void)
{
	if (datagram_write) {
		datagram_ret = ec_datagram_write_real(
		    datagram_addr,
		    datagram_buffer,
		    datagram_len);
		datagram_write = false;
		sem_post(&datagram_sync_sem);
	} else if (datagram_read) {
		datagram_ret = ec_datagram_read_real(
		    datagram_addr,
		    datagram_buffer,
		    datagram_len);
		datagram_read = false;
		sem_post(&datagram_sync_sem);
	}
}



static bool init_done = false;

bool ec_initialized(void)
{
	return init_done;
}

int ec_init(char *ifname, bool single_cycle)
{
	struct timeval timeout;
	double timeout_sec;

	if (init_done)
		return 0;

	single_pdo = single_cycle;
	if (single_pdo) {
		tx_pdo = &rx_pdo;
		log_send(LOG_T_DEBUG, "Using a single PDO for rx and tx.");
	}
	
	log_send(LOG_T_DEBUG, 
	    "Going to initialize interface %s for control data", ifname);

	ec_sock = eth_open(ifname);
	if (ec_sock < 0)
		goto err_nosocket;
	
	timeout_sec = 0.1;
	
	timeout.tv_sec = (int)(timeout_sec);
	timeout.tv_usec = (int)(timeout_sec * 1000000) % 1000000;
	log_send(LOG_T_DEBUG,
	    "Setting rx/tx timeout to %d.%06d sec.",
	    (int)timeout.tv_sec, (int)timeout.tv_usec);

	eth_timeout_set(ec_sock, &timeout, &timeout);

	if (ec_slave_count() >= 0) {
		log_send(LOG_T_DEBUG, "Found ethercat slaves");
	} else {
		log_send(LOG_T_WARNING, "No ethercat slaves found");
	}
	log_send(LOG_T_DEBUG, 
	    "Successfull init, %d slaves found.", ec_slave_count());

	ec_rx_pdo_add_time();

	controller_sample_start_hook(ec_datagram_sample_init, NULL);

	init_done = true;

	return 0;

err_nosocket:
	log_send(LOG_T_ERROR, "Failed to initialize Ethercat subsystem");
	return -1;
}
