/*
	Copyright Jeroen Vreeken (pe1rxq@amsat.org), 2008, 2009, 2013
	Copyright Stichting C.A. Muller Radioastronomiestation, 2008, 2009, 2013

	This program is free software: you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation, either version 3 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program.  If not, see <http://www.gnu.org/licenses/>.

 */

#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#include <ec/esc.h>
#include <ec/esc_esi.h>
#include <ec/esc_coe.h>
#include <ec/esc_mailbox.h>
#include <ec/ec_stoeber.h>
#include <ec/esc_dc.h>
#include <log/log.h>

static void ec_stoeber_emergency_handler(struct canopen_dev *dev, 
    uint16_t code, uint8_t reg, uint8_t vendor[5])
{
	char *msg = "unknown";
	
	if (code == 0x0000) {
		msg = "no_error";
	} else if (code == 0x2110) {
		msg = "short circuit earth";
	} else if (code == 0x2230) {
		msg = "intern short circuit earth";
	} else if (code == 0x2310) {
		msg = "continuous overcurrent";
	} else if (code == 0x3110) {
		msg = "mains overvoltage";
	} else if (code == 0x3120) {
		msg = "mains undervoltage";
	} else if (code == 0x4210) {
		msg = "temperature device";
	} else if (code == 0x4280) {
		msg = "temperature device i2t";
	} else if (code == 0x4310) {
		msg = "temperature drive";
	} else if (code == 0x4310) {
		msg = "temperature drive i2t";
	} else if (code == 0x5000) {
		msg = "device hardware";
	} else if (code == 0x5200) {
		msg = "device hardware control";
	} else if (code == 0x6010) {
		msg = "software reset";
	} else if (code == 0x6100) {
		msg = "internal software";
	} else if (code == 0x6200) {
		msg = "user software";
	} else if (code == 0x6300) {
		msg = "data record";
	} else if (code == 0x6310) {
		msg = "loss of parameters";
	} else if (code == 0x7110) {
		msg = "brake chopper";
	} else if (code == 0x7120) {
		msg = "motor";
	} else if (code == 0x7303) {
		msg = "resolver 1 fault";
	} else if (code == 0x8100) {
		msg = "communication";
	} else if (code == 0x8400) {
		msg = "velocity speed control";
	} else if (code == 0x8311) {
		msg = "excess torque";
	} else if (code == 0x9000) {
		msg = "external error";
	} else if (code == 0xa000) {
		msg = "Transition from Pre-Operational to Safe-Operational was not successful";
	} else if (code == 0xa001) {
		msg = "Transition from Safe-Operational to Pre-Operational was not successful";
	}
	
	log_send(LOG_T_WARNING, "%s: %s",
	    dev->logname, msg);
}

int ec_stoeber_init(struct ec_stoeber *stbr)
{
	uint32_t leval32;
	uint16_t leval16;
	uint8_t val8;
	struct esc_mailbox *mb;
	struct esc_syncmanager_info mb_rd, mb_wr, pdo_tx, pdo_rx;
	struct canopen_dev *canopen_dev;
	uint32_t vendorid, productcode, serialno;
	struct timespec timeout = { 1, 0 };
	struct esc_dc_sync dc_sync;
	
	vendorid = esc_esi_vendorid_get(&stbr->addr);
	productcode = esc_esi_productcode_get(&stbr->addr);
	serialno = esc_esi_serialno_get(&stbr->addr);
	log_send(LOG_T_DEBUG, 
	    "VendorId: 0x%04x, ProductCode: 0x%04x, SerialNo: 0x%04x",
	    vendorid, productcode, serialno);
	/* TODO: check values */
	
	if (!stbr->esc_dev)
		stbr->esc_dev = esc_device_create(&stbr->addr, "stoeber");
	if (!stbr->esc_dev)
		return -1;
	
	if (esc_al_state_set(&stbr->addr, ESC_AL_STATE_INIT, &timeout) < 0)
		return -1;
	
	log_send(LOG_T_DEBUG, "Initializing mailbox");
	
	
	dc_sync.sync0_cycle = 4000 * 1000;
	dc_sync.sync0_shift = 0;
	dc_sync.sync_out_active = true;
	dc_sync.sync0_active = true;
	dc_sync.sync1_active = false;
	esc_dc_init(&stbr->addr, &dc_sync);

	stbr->addr.addr.position.off = 0x0000;
	
	
	/* mailbox read (slave->master) @0x18f6 */
	mb_rd.start = 0x18f6;
	mb_rd.len = 0x0040;
	mb_rd.sm = 1;
	
	/* mailbox write (master->slave) @0x1800 */
	mb_wr.start = 0x1800;
	mb_wr.len = 0x0040;
	mb_wr.sm = 0;
	
	if (stbr->canopen_dev) {
		esc_mailbox_reinit(&stbr->addr, &mb_rd, &mb_wr);
		
		canopen_dev = stbr->canopen_dev;
	} else {
		mb = esc_mailbox_create(&stbr->addr, &mb_rd, &mb_wr);
		canopen_dev = esc_coe_create(mb);
	
		canopen_emergency_handler_attach(canopen_dev, ec_stoeber_emergency_handler);

		stbr->canopen_dev = canopen_dev;
	}
	
	log_send(LOG_T_DEBUG, "Going to state pre-operational");
	
	if (esc_al_state_set(&stbr->addr, ESC_AL_STATE_PRE_OPERATIONAL, &timeout) < 0)
		return -1;

	/* A180 device control, disable it for now */
	val8 = 0x00;
	canopen_write_param(canopen_dev, STOEBER_PARAM2INDEX('A', 180), 0x0, &val8, 1);

	/* A61 fault reset source to parameter */
	val8 = 2;
	canopen_write_param(canopen_dev, STOEBER_PARAM2INDEX('A', 61), 0x0, &val8, 1);
	usleep(3000);
	/* A180 device control, reset faults */
	val8 = 0x02;
	canopen_write_param(canopen_dev, STOEBER_PARAM2INDEX('A', 180), 0x0, &val8, 1);
	usleep(50000);
	/* A180 device control, reset faults */
	val8 = 0x00;
	canopen_write_param(canopen_dev, STOEBER_PARAM2INDEX('A', 180), 0x0, &val8, 1);

	canopen_read_param(canopen_dev, STOEBER_PARAM2INDEX('E', 52), 0x0, &leval32, 4);
	stbr->serial_number = le32toh(leval32);

	/* A21 brake resistor */
	leval16 = htole16(stbr->brake_resistor * 10.0);
	canopen_write_param(canopen_dev, STOEBER_PARAM2INDEX('A', 21), 0x0, &leval16, 2);
	/* A22 brake resistor power */
	leval16 = htole16(stbr->brake_power);
	canopen_write_param(canopen_dev, STOEBER_PARAM2INDEX('A', 22), 0x0, &leval16, 2);

	/* A29 fault quickstop */
	val8 = 1;
	canopen_write_param(canopen_dev, STOEBER_PARAM2INDEX('A', 29), 0x0, &val8, 1);

	/* A34 auto start */
	val8 = 1;
	canopen_write_param(canopen_dev, STOEBER_PARAM2INDEX('A', 34), 0x0, &val8, 1);
	
	/* A45*/
	val8 = 1;
	canopen_write_param(canopen_dev, STOEBER_PARAM2INDEX('A', 45), 0x0, &val8, 1);
	
	/* A60*/
	val8 = 1;
	canopen_write_param(canopen_dev, STOEBER_PARAM2INDEX('A', 60), 0x0, &val8, 1);
	
	/* A150 cycle time */
	val8 = 0x06;
	canopen_write_param(canopen_dev, STOEBER_PARAM2INDEX('A', 150), 0x0, &val8, 1);

	/* A258 set PDO timeout */
	leval16 = htole16(1000);
	canopen_write_param(canopen_dev, STOEBER_PARAM2INDEX('A', 258), 0x0, &leval16, 2);

	/* G90 activate PLL */
	val8 = 1;
	canopen_write_param(canopen_dev, STOEBER_PARAM2INDEX('G', 90), 0x0, &val8, 1);

	/* G91 pll phase offset */
	leval32 = htole32(0);
	canopen_write_param(canopen_dev, STOEBER_PARAM2INDEX('G', 91), 0x0, &leval32, 4);

	/* G92 pll gain pk */
	leval32 = htole32(10);
	canopen_write_param(canopen_dev, STOEBER_PARAM2INDEX('G', 92), 0x0, &leval32, 4);

	/* G93 pll low pass */
	leval32 = htole32(999);
	canopen_write_param(canopen_dev, STOEBER_PARAM2INDEX('G', 93), 0x0, &leval32, 4);

	/* G98 cycle to 4ms */
	leval16 = htole16(4000);
	canopen_write_param(canopen_dev, STOEBER_PARAM2INDEX('G', 98), 0x0, &leval16, 2);

	/* Set I06 to 2 decimal digits */
	val8 = 2;
	canopen_write_param(canopen_dev, STOEBER_PARAM2INDEX('I', 6), 0x0, &val8, 1);

	/* C32 (motor integrator) to 3000 ms */
	leval16 = htole16(32000);
	canopen_write_param(canopen_dev, STOEBER_PARAM2INDEX('C', 32), 0x0, &leval16, 2);

	/* C33 low pass reference speed 1 ms */
	leval32 = htole32(10);
	canopen_write_param(canopen_dev, STOEBER_PARAM2INDEX('C', 33), 0x0, &leval32, 4);

	/* C130 torque limit source to parameter */
	val8 = 4;
	canopen_write_param(canopen_dev, STOEBER_PARAM2INDEX('C', 130), 0x0, &val8, 1);

	/* Set I10 to maximum speed */
	leval32 = htole32(stbr->max_speed / (2*M_PI) * 360 * 100);
//	leval32 = htole32(3000 * 360 / 60 * 100);
	canopen_write_param(canopen_dev, STOEBER_PARAM2INDEX('I', 10), 0x0, &leval32, 4);

	/* Set I11 to maximum acceleration */
//	leval32 = htole32(2000000 * 100);
	leval32 = htole32(stbr->max_accel * 100);
	canopen_write_param(canopen_dev, STOEBER_PARAM2INDEX('I', 11), 0x0, &leval32, 4);
		
	/* Set I25 to 100% speed ff */
	leval16 = htole16(100);
	canopen_write_param(canopen_dev, STOEBER_PARAM2INDEX('I', 25), 0x0, &leval16, 2);

	/* Set I30 to 1 (encoder signal 0) */
	val8 = 1;
	canopen_write_param(canopen_dev, STOEBER_PARAM2INDEX('I', 30), 0x0, &val8, 1);

	/* Set I37 to 1 (active) */
	val8 = 1;
	canopen_write_param(canopen_dev, STOEBER_PARAM2INDEX('I', 37), 0x0, &val8, 1);

	/* Set I100 to parameter (execute source) */
	val8 = 2;
	canopen_write_param(canopen_dev, STOEBER_PARAM2INDEX('I', 100), 0x0, &val8, 1);
		
	/* B17 Torque M0 */
	canopen_read_param(canopen_dev, STOEBER_PARAM2INDEX('B', 17), 0x0, &leval32, 4);
	log_send(LOG_T_DEBUG, "B17: %08x %d", le32toh(leval32), le32toh(leval32));
	stbr->standstill_torque = (float)(le32toh(leval32)) / 1000.0;
	log_send(LOG_T_DEBUG, "Standstill torque: %f", stbr->standstill_torque);

	/* C03 max pos torque 150% of standstill */
	leval32 = htole32(150);
	canopen_write_param(canopen_dev, STOEBER_PARAM2INDEX('C', 3), 0x0, &leval32, 4);

	/* C05 max neg torque -150% of standstill */
	leval32 = htole32(-150);
	canopen_write_param(canopen_dev, STOEBER_PARAM2INDEX('C', 5), 0x0, &leval32, 4);

//	/* A180 device control, enable it to make sure we stay
//	   at the same position */
//	val8 = 0x01;
//	canopen_write_param(canopen_dev, 0x20b4, 0x0, &val8, 1);

	/* I211 command byte to 4 (move_velocity) */
	val8 = 0x04;
	canopen_write_param(canopen_dev, STOEBER_PARAM2INDEX('I', 211), 0x0, &val8, 1);
	/* I215 v to 0 */
	leval16 = htole16(0);
	canopen_write_param(canopen_dev, STOEBER_PARAM2INDEX('I', 215), 0x0, &leval16, 2);
	/* I210 latch command */
	val8 = 0x00;
	canopen_write_param(canopen_dev, STOEBER_PARAM2INDEX('I', 210), 0x0, &val8, 1);
	val8 = 0x01;
	canopen_write_param(canopen_dev, STOEBER_PARAM2INDEX('I', 210), 0x0, &val8, 1);

	/* Transmit pdo: */

	/* I80 (4bytes) */
	leval32 = STOEBER_PDO_MAP('I', 80, 0, 32);
	canopen_write_param(canopen_dev, STOEBER_PARAM2INDEX('A', 233), 0x0, &leval32, 4);
		
	/* E91 (4bytes) */
	leval32 = STOEBER_PDO_MAP('E', 91, 0, 32);
	canopen_write_param(canopen_dev, STOEBER_PARAM2INDEX('A', 233), 0x1, &leval32, 4);

	/* E90 (2bytes) */
	leval32 = STOEBER_PDO_MAP('E', 90, 0, 16);
	canopen_write_param(canopen_dev, STOEBER_PARAM2INDEX('A', 233), 0x2, &leval32, 4);
		
	/* E19 (2bytes) */
	leval32 = STOEBER_PDO_MAP('E', 19, 0, 16);
	canopen_write_param(canopen_dev, STOEBER_PARAM2INDEX('A', 233), 0x3, &leval32, 4);

	/* E10 (2bytes) */
	leval32 = STOEBER_PDO_MAP('E', 10, 0, 16);
	canopen_write_param(canopen_dev, STOEBER_PARAM2INDEX('A', 233), 0x4, &leval32, 4);

	leval32 = 0;
	canopen_write_param(canopen_dev, STOEBER_PARAM2INDEX('A', 233), 0x5, &leval32, 4);

	/* Receive pdo: */

	/* A180 (1byte) */
	leval32 = STOEBER_PDO_MAP('A', 180, 0, 32);
	canopen_write_param(canopen_dev, STOEBER_PARAM2INDEX('A', 225), 0x0, &leval32, 4);
		
	/* F210 (1byte) */
	leval32 = STOEBER_PDO_MAP('F', 210, 0, 32);
	canopen_write_param(canopen_dev, STOEBER_PARAM2INDEX('A', 225), 0x1, &leval32, 4);

	/* I210 (2bytes) */
	leval32 = STOEBER_PDO_MAP('I', 210, 0, 16);
	canopen_write_param(canopen_dev, STOEBER_PARAM2INDEX('A', 225), 0x2, &leval32, 4);
		
	/* I213 (4bytes) */
	leval32 = STOEBER_PDO_MAP('I', 213, 0, 16);
	canopen_write_param(canopen_dev, STOEBER_PARAM2INDEX('A', 225), 0x3, &leval32, 4);

	/* I215 (2bytes) */
	leval32 = STOEBER_PDO_MAP('I', 215, 0, 16);
	canopen_write_param(canopen_dev, STOEBER_PARAM2INDEX('A', 225), 0x4, &leval32, 4);

	/* C230 (2bytes) */
	leval32 = STOEBER_PDO_MAP('C', 230, 0, 16);
	canopen_write_param(canopen_dev, STOEBER_PARAM2INDEX('A', 225), 0x5, &leval32, 4);


	log_send(LOG_T_DEBUG, "Setup PDOs");

	/* sm2 tx pdo (master->slave) @0x1600 */

	pdo_tx.start = 0x1600;
	pdo_tx.len = 0x000c;
	pdo_tx.sm = 2;

	/* sm3 rx pdo (slave->master) @0x1a00 */

	pdo_rx.start = 0x1a00;
	pdo_rx.len = 0x000e;
	pdo_rx.sm = 3;
		
	esc_pdo_tx_set(&stbr->addr, &pdo_tx, &stbr->tx_buffer, false);
	esc_pdo_rx_set(&stbr->addr, &pdo_rx, &stbr->rx_buffer);
	
	log_send(LOG_T_DEBUG, "going to safe operational");
	
	if (esc_al_state_set(&stbr->addr, ESC_AL_STATE_SAFE_OPERATIONAL, &timeout) < 0)
		return -1;

	log_send(LOG_T_DEBUG, "ec_stoeber: Done");
	return 0;

}
