/*
	Copyright Jeroen Vreeken (pe1rxq@amsat.org), 2013
	Copyright Stichting C.A. Muller Radioastronomiestation, 2013

	This program is free software: you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation, either version 3 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program.  If not, see <http://www.gnu.org/licenses/>.

 */
/*
	esc_coe.c
	
	These functions implement a canopen device (which can be used with
	the canopen functions) based on a ethercat mailbox.
 */

#include <string.h>
#include <pthread.h>
#include <limits.h>
#include <unistd.h>

#include <ec/ec.h>
#include <ec/esc_coe.h>
#include <log/log.h>



struct canopen_dev_priv {
	struct esc_mailbox *mbox;
	int refcount;
};

static ssize_t esc_coe_send(struct canopen_dev *dev, void *req, size_t len,
    struct timespec *timeout)
{
	int ret;
	unsigned char coereq[10];
	int coe_type = 0x2;
	
	memcpy(coereq + 2, req, len);
	coereq[0] = 0x00;
	coereq[1] = coe_type << 4;
	
	ret = esc_mailbox_write(dev->priv->mbox, ESC_MAILBOX_TYPE_COE, coereq, 
	    len+2, timeout);

	if (ret > 2)
		return ret - 2;
	return ret;
}

static ssize_t esc_coe_recv(struct canopen_dev *dev, void *res, size_t len,
    struct timespec *timeout)
{
	int type;
	int ret;
	unsigned char coeres[10];
	int coe_type;
	
	ret = esc_mailbox_read(dev->priv->mbox, &type, coeres, len + 2, timeout);
	if (ret == 0)
		return 0;
	if (type != ESC_MAILBOX_TYPE_COE) {
		return -1;
	}
	
	coe_type = (coeres[1] >> 4) & 0x0f;

	/* check for emergencies */
	if (coe_type == 1) {
		canopen_emergency_handler(dev, coeres + 2, ret - 2);
		return 0;
	}
	
	if (ret > 2) {
		memcpy(res, coeres + 2, ret - 2);
		return ret - 2;
	}
	return ret;
}

static void *esc_coe_emergency_poll(void *arg)
{
	struct canopen_dev *dev = arg;
	unsigned char cmd[8];
	struct timespec timeout = { 0, 0 };
	
	log_send(LOG_T_DEBUG, "Emergency poll thread started for CoE device %p",
	    dev);
	while (1) {
		sleep(1);

		pthread_mutex_lock(&dev->lock);
	
		dev->recv_response(dev, cmd, 8, &timeout);
	
		pthread_mutex_unlock(&dev->lock);
	}
	return NULL;
}

struct canopen_dev *esc_coe_create(struct esc_mailbox *mailbox)
{
	struct canopen_dev *dev;
	pthread_t thread_id;
	pthread_attr_t attr;
	
	dev = malloc(sizeof(struct canopen_dev));
	if (!dev)
		goto err_malloc;
	dev->priv = malloc(sizeof(struct canopen_dev_priv));
	if (!dev->priv)
		goto err_priv;
	
	dev->priv->mbox = mailbox;
	dev->send_request = &esc_coe_send;
	dev->recv_response = &esc_coe_recv;
	dev->logname = "esc";
	dev->priv->refcount = 2;

	canopen_init(dev);

	pthread_attr_init(&attr);
	pthread_attr_setstacksize(&attr, PTHREAD_STACK_MIN * 2);
	pthread_create(&thread_id, &attr,
	    esc_coe_emergency_poll, dev);

	return dev;

err_priv:
	free(dev);
err_malloc:
	return NULL;
}

void esc_coe_destroy(struct canopen_dev *coe)
{
	coe->priv->refcount--;

	if (!coe->priv->refcount) {
		free(coe->priv);
		free(coe);
	}
}
