/*
	Copyright Jeroen Vreeken (pe1rxq@amsat.org), 2013
	Copyright Stichting C.A. Muller Radioastronomiestation, 2013

	This program is free software: you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation, either version 3 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program.  If not, see <http://www.gnu.org/licenses/>.

 */
/*
	esc_device.c
	
	Code for using and maintaining the esc_device.	
 */

#include <string.h>
#include <pthread.h>
#include <limits.h>

#include <ec/ec.h>
#include <ec/esc.h>
#include <ec/esc_device.h>
#include <ec/esc_esi.h>
#include <ec/esc_mailbox.h>
#include <ec/esc_watchdog.h>
#include <log/log.h>

struct esc_device *esc_device_create(struct ec_dgram_addr *ec_addr, char *name)
{
	struct esc_device *dev;
	
	log_send(LOG_T_DEBUG, "%s: Creating device", name);
	
	dev = calloc(1, sizeof(struct esc_device));
	if (dev)
		dev->name = strdup(name);

	memcpy(&dev->addr, ec_addr, sizeof(struct ec_dgram_addr));

	dev->mbox_rd_sm = -1;
	dev->mbox_wr_sm = -1;

	esc_init(&dev->addr);

	return dev;
}

void esc_device_destroy(struct esc_device *dev)
{
	/* todo: kill poll thread */

	if (dev->categories)
		esc_esi_categories_destroy(dev->categories);
	if (dev->name)
		free(dev->name);
	
	free(dev);
}

static void *esc_device_poll(void *arg)
{
	struct esc_device *dev = arg;
	enum esc_al_status status;
	enum esc_al_state state;
	bool err_status = false;
	bool err_state = false;
	
	log_send(LOG_T_DEBUG, "Poll thread started for esc device %s", 
	    dev->name);

	while (1) {
		char *str;
		bool recover = false;
		int recover_state = ESC_AL_STATE_SAFE_OPERATIONAL;
		
		sleep(1);

		status = esc_al_status_code_get(&dev->addr);
		state = esc_al_state_get(&dev->addr);
		
		if (state == ESC_AL_STATE_ERROR) {
			if (!err_state) {
				log_send(LOG_T_ERROR, 
				    "%s: Could not read ESC state",
				    dev->name);
			}
			err_state = true;
		} else {
			if (err_state) {
				log_send(LOG_T_WARNING, 
				    "%s: ESC state readable again",
				    dev->name);
			}
			err_state = false;
		}
		if (status < 0) {
			if (!err_status) {
				log_send(LOG_T_ERROR, 
				    "%s: Could not read ESC status",
				    dev->name);
			}
			err_status = true;
		} else {
			if (err_status) {
				log_send(LOG_T_WARNING, 
				    "%s: ESC status readable again",
				    dev->name);
			}
			err_status = false;
		}
		
		if (!err_state) {
			if (state != dev->state) {
				log_send(LOG_T_WARNING,
				    "%s: ESC state changed: %d %s -> %d %s",
				    dev->name, 
				    dev->state, esc_al_state2str(dev->state),
				    state, esc_al_state2str(state));
			}
			
			if (dev->auto_recover && state < dev->state) {
				recover = true;
				if (state == ESC_AL_STATE_INIT) {
					recover_state =
					    ESC_AL_STATE_PRE_OPERATIONAL;
				} else if (state == ESC_AL_STATE_PRE_OPERATIONAL) {
					recover_state =
					    ESC_AL_STATE_SAFE_OPERATIONAL;
				} else {
					recover_state = dev->state;
				}
			} else {
				dev->state = state;
			}
		}
		
		if (status == dev->status || err_status)
			if (!recover)
				continue;
		
		dev->status = status;
		
		str = esc_al_status2str(status);

		if (!status) {
			log_send(LOG_T_WARNING,
			    "%s: ESC status changed: No error", dev->name);
			if (!recover)
				continue;
		} else {
			switch(status) {
			case ESC_AL_STATUS_SM_WATCHDOG:
				str = "Sync manager watchdog";
				recover = true;
				recover_state = ESC_AL_STATE_OPERATIONAL;
				break;
			case ESC_AL_STATUS_NO_ERROR:
			case ESC_AL_STATUS_UNSPECIFIED:
			case ESC_AL_STATUS_INVALID_REQ:
			case ESC_AL_STATUS_UNKNOWN_REQ:
			case ESC_AL_STATUS_BOOTSTRAP_UNSUP:
			case ESC_AL_STATUS_NO_VALID_FMW:
			case ESC_AL_STATUS_INVALID_MBOX1:
			case ESC_AL_STATUS_INVALID_MBOX2:
			case ESC_AL_STATUS_INVALID_SM:
			case ESC_AL_STATUS_NO_VALID_INPUT:
			case ESC_AL_STATUS_NO_VALID_OUTPUT:
			case ESC_AL_STATUS_SYNC_ERROR:
			case ESC_AL_STATUS_INVALID_SM_TYPES:
			case ESC_AL_STATUS_INVALID_OUT_CONF:
			case ESC_AL_STATUS_INVALID_IN_CONF:
			case ESC_AL_STATUS_INVALID_WD_CONF:
			case ESC_AL_STATUS_SLAVE_COLD:
			case ESC_AL_STATUS_SLAVE_INIT:
			case ESC_AL_STATUS_SLAVE_PREOP:
			case ESC_AL_STATUS_SLAVE_SAFEOP:
			case ESC_AL_STATUS_INVALID_DC_S_CONF:
			case ESC_AL_STATUS_INVALID_DC_L_CONF:
			case ESC_AL_STATUS_PLL_ERROR:
			case ESC_AL_STATUS_INVALID_DC_IO:
			case ESC_AL_STATUS_INVALID_DC_TO:
			case ESC_AL_STATUS_MBOX_EOE:
			case ESC_AL_STATUS_MBOX_COE:
			case ESC_AL_STATUS_MBOX_FOE:
			case ESC_AL_STATUS_MBOX_SOE:
			case ESC_AL_STATUS_MBOX_VOE:
			default:
				break;
			}
			log_send(LOG_T_ERROR, "%s: ESC status code changed: 0x%04x: %s",
			    dev->name, status, str);
		}

		if (recover) {
			struct timespec to = { 1, 0 };
			log_send(LOG_T_WARNING, 
			    "%s: Trying to recover to state %d %s",
			    dev->name, recover_state,
			    esc_al_state2str(recover_state));
			esc_al_error_ack(&dev->addr);
			esc_al_state_set(&dev->addr, recover_state, &to);
			dev->state = recover_state;
		}
	}

	return NULL;
}

int esc_device_poll_start(struct esc_device *dev)
{
	pthread_t thread_id;
	pthread_attr_t attr;

	pthread_attr_init(&attr);
	pthread_attr_setstacksize(&attr, PTHREAD_STACK_MIN * 2);
	pthread_create(&thread_id, &attr, esc_device_poll, dev);

	return 0;
}

/* initialize up to pre operational */
int esc_device_initialize_pre_operational(struct esc_device *dev)
{
	int r = 0;
	struct timespec	timeout = { 1, 0 };

	if (dev->tx_pdo)
		esc_pdo_tx_set(&dev->addr, &dev->sm[dev->tx_pdo_sm],
		    &dev->tx_data, dev->tx_pdo_watchdog);
	if (dev->rx_pdo)
		esc_pdo_rx_set(&dev->addr, &dev->sm[dev->rx_pdo_sm], 
		    &dev->rx_data);

	if (esc_al_state_set(&dev->addr, ESC_AL_STATE_INIT, &timeout) < 0)
		log_send(LOG_T_ERROR, "%s: Could not go to state init",
		    dev->name);


	if (dev->mbox_wr_sm >= 0 && dev->mbox_rd_sm >= 0) {
		dev->mailbox = esc_mailbox_create(&dev->addr, 
		    &dev->sm[dev->mbox_rd_sm], &dev->sm[dev->mbox_wr_sm]);
	}

	if (esc_al_state_set(&dev->addr, ESC_AL_STATE_PRE_OPERATIONAL, &timeout)< 0) {
		log_send(LOG_T_ERROR, 
		    "%s: Could not go to state pre operational", dev->name);
		r = -1;
	}

	esc_dc_init(&dev->addr, &dev->dc_sync);
	esc_watchdog_init(dev);

	r = esc_device_poll_start(dev);

	dev->auto_recover = true;

	return r;
}

