/*
	Copyright Jeroen Vreeken (pe1rxq@amsat.org), 2007, 2008
	Copyright Stichting C.A. Muller Radioastronomiestation, 2007, 2008

	This program is free software: you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation, either version 3 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program.  If not, see <http://www.gnu.org/licenses/>.

 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <math.h>

#include <controller/controller_block.h>
#include <controller/controller_trace.h>
#include <controller/controller_sample.h>
#include <block/block_filter_iir.h>
#include <block/block_random.h>
#include <block/block_limit_2nd.h>
#include <block/block_gain.h>

#define FREQ 250.0

struct block {
	char *name;
	struct controller_block *(*create)(char *);
};

struct param_misc {
	char *blockname;
	char *paramname;
	void *value;
};

struct block_link {
	char *outblock;
	char *outterm;
	char *inblock;
	char *interm;
};

static struct block controller_blocks[] = {
	{ "noise",  block_random_create                },
	{ "iir",    block_filter_iir_create            },
	{ "iir2",    block_filter_iir_create            },
	{ "iir3",    block_filter_iir_create            },
	{ "gain",  block_gain_create			},
	{ "limit_2nd", block_limit_2nd_create		},
	
	{ NULL, NULL }
};


static struct param_misc controller_param[] = {
	{ "iir",           "transfer",       &(struct block_filter_iir_param) {
						4,
						1.426452631e+00,
						(float[]){ -0.4935483739, 2.2962251554 ,-4.1111529408,  3.3084749925 },
						(float[]){ 1 ,           -3.9974196347,  5.9948409340, -3.9974196347 },
	                                     } },
	{ "iir2",           "transfer",       &(struct block_filter_iir_param) {
						4,
						1.426452631e+00,
						(float[]){ -0.8431538806, 3.5149541322, -5.5003478020 ,3.8285471832},
						(float[]){ 1 ,         -3.9987352229,  5.9974708457 , -3.9987352229 },
	                                     } },
	{ "iir3",    "transfer",            
						&(struct block_filter_iir_param) {
	                                                      2,
							      2.555570536e+04,
						              (float[]) {
                                 				-0.9823854506,
								1.9822289298
							      },
							      (float[]) {
							           1,
								   2
							      },
	                                                  } },
	{ "limit_2nd",     "min",                 &(float){-0.001           }},
	{ "limit_2nd",     "max",                 &(float){ 0.001           }},
	{ "gain",	"gain",			&(float){ 0.1	}},
		{ NULL, NULL, NULL }
};

static struct block_link controller_links[] = {
	{ "noise",                         "value",   "iir",            "in"           },
	{ "noise",                         "value",   "iir2",            "in"           },
	{ "noise",                         "value",   "iir3",            "in"           },
	{ "noise",			"value", 		"gain", 	"in"	},
	{ "gain", 			"out",			"limit_2nd",	"in"	},
	
	{ NULL, NULL, NULL, NULL }
};

static struct controller_trace_name dt_trace_list[] = {
	{ "iir",                     "none",       "iir",                      "out" },
	{ "iir2",                     "none",       "iir2",                      "out" },
	{ "iir3",                     "none",       "iir3",                      "out" },
	{ "limit", 		"none",		"limit_2nd",		"out" },
	{ "noise",                     "none",       "noise",                      "value" },
	
	{ NULL, NULL, NULL, NULL }
};



int main(int argc, char **argv)
{
	struct controller_block *block;
	int i;
	
	controller_block_param_init();
	
	/* Create blocks */

	for (i = 0; controller_blocks[i].name; i++) {
		block = controller_blocks[i].create(controller_blocks[i].name);
		if (!block) {
			printf("Could not create block %s\n",
			    controller_blocks[i].name);
			return 1;
		}
	}

	/* Set params */
	
	for (i = 0; controller_param[i].blockname; i++) {
		if (controller_block_param_set(
		    controller_param[i].blockname,
		    controller_param[i].paramname,
		    controller_param[i].value) < 0) {
			printf("Could not set %s.%s\n",
			    controller_param[i].blockname,
			    controller_param[i].paramname);
			return 1;
		}
	}


	/* Link blocks */

	for (i = 0; controller_links[i].outblock; i++) {
		if (controller_block_connect(
		    controller_links[i].outblock,
		    controller_links[i].outterm,
		    controller_links[i].inblock,
		    controller_links[i].interm,
		    1) < 0) {
			printf("Could not link %s.%s to %s.%s\n",
			    controller_links[i].outblock,
			    controller_links[i].outterm,
			    controller_links[i].inblock,
			    controller_links[i].interm);
			return 1;
		}
	}

	
	/* Add traces to the trace interface */

	for (i = 0; dt_trace_list[i].name; i++)
		controller_trace_list_add(&dt_trace_list[i]);

	controller_trace_server_start(10000, 100);

	/* Start 'sample' */

	controller_block_sample_init();

	controller_sample_frequency_set(FREQ);

	printf("Starting sample thread\n");
	controller_sample_start();

	while (1) {
		/* Log once every 15 minutes */
		sleep(15 * 60);
	}

	return 0;
}
